#ifndef NETWORK_H_
#define NETWORK_H_

#include "generic_types.h"
#include "net_types.h"
#include "buffer_management.h"
#include "device.h"

#ifndef MAX_NETWORK_IF_CONFIGS  
    #ifdef ARCH_PATMOS
        // Patmos has two Ethernet MACs at the DE2-115 Board
        #define MAX_NETWORK_IF_CONFIGS 2
    #else
        #define MAX_NETWORK_IF_CONFIGS 4
    #endif
#endif

#ifndef MAX_NETWORK_IF_NAME_LENGTH  
    #define MAX_NETWORK_IF_NAME_LENGTH 16
#endif

#ifndef WCET_MAXIMUM_QUEUE_LENGTH  
    #define WCET_MAXIMUM_QUEUE_LENGTH 255
#endif


typedef enum 
{
    NETWORK_INTERFACE_NONE                  =   0x0000, // Interfaces has no type => undefined, error value
    NETWORK_INTERFACE_PATMOS_ETHERNET       =   0x0001, // Ethernet Interface of Patmos
    NETWORK_INTERFACE_PATMOS_ETHERNET_TSN   =   0x0002, // Ethernet Interface with TSN support of Patmos
    NETWORK_INTERFACE_PATMOS_NOC            =   0x0003, // Network on Chip interface of Patmos
} network_interface_types_t;

typedef enum 
{
    NETWORK_ADDRESS_IP   =   0x0001, // Ethernet Interface of Patmos
} network_address_types_t;

typedef enum 
{
    NETWORK_L2_ETHERNET         =   0x0001, // Ethernet Interface
    NETWORK_L2_ETHERNET_VLAN    =   0x0002, // Ethernet with VLAN 
} network_layer2_types_t;

typedef enum
{
    NETWORK_OPTION_LOGGING                  =   0x0001,
    NETWORK_OPTION_PRINT_DRIVER_DBG_INFO    =   0x0002,
    NETWORK_OPTION_TSN_CFG                  =   0x0010,
    NETWORK_OPTION_GET_PTP_CLOCK            =   0x0020,
    NETWORK_OPTION_SET_PTP_CLOCK            =   0x0021,
    NETWORK_OPTION_GET_PTP_LAST_RX_TIMESTAMP=   0x0022,
    NETWORK_OPTION_GET_PTP_LAST_TX_TIMESTAMP=   0x0023,
    NETWORK_OPTION_SET_PTP_CLOCK_OFFSET     =   0x0024,
} network_options_t;


typedef struct 
{
    network_address_types_t type;
    union 
    {
        struct 
        {
            ip_address_t  ip_address;
            ip_address_t     gateway;
            uint16_t         netmask;   // how many bits belong to network
        } ip;
        
    };
} network_address_t;

typedef struct 
{
    network_layer2_types_t l2_type;
    union  
    {
        vlan_header_t vlan;
    };
} network_layer2_config_t;


typedef struct 
{
    char *interface_name;                       // user readable name of the interface
    network_interface_types_t interface_type;   // what kind of interface is connected
    device_config_t device_config;              // Device configuration 
    network_layer2_config_t layer2_config;      // e.g. vlan
    network_address_t address;                  // sets layer 3 addresses
} network_interface_config_t;

// initialise the network stack
void network_init(buffer_management_config_t *buffer_mgmt_cfg);

// register a new interface
uint16_t network_register_interface(network_interface_config_t* config);
// duplicate an already existing network interface
uint16_t network_duplicate_interface(uint16_t interface_id, network_interface_config_t* config);

// get/set options for a network interface or the global network interface
network_error_t network_set_option(uint16_t interface_id, network_options_t option, void* value, uint32_t value_size);
network_error_t network_get_option(uint16_t interface_id, network_options_t option, void* value, uint32_t *value_size);

// request/free memory space
void *network_get_buffer(uint16_t size);
void network_return_buffer(void *buffer); 

// check whether an interface exists
network_error_t network_interface_exists(uint16_t interface_id);

// update sending and receiving routings of the single network interfaces
void network_update(void);

// handover message to network interface 
network_error_t network_send(uint16_t interface_id, network_frame_t *frame, uint8_t blocking );

// read all received frames from all network interfaces
void network_read_interfaces(void);
// read all received frames from a single interface
network_error_t network_read_interface(uint16_t interface_id);
// read a single message from a single interface
network_error_t network_interface_get_message(uint16_t interface_id);

// retrieve interface id by name, via network address and network mask
uint16_t network_get_interface_by_name(char *name);
uint16_t network_get_interface_by_ip_address(network_address_t *address);
uint16_t network_get_interface_by_ip_netmask(ip_address_t *address);

// reconfigure network address of network interface
network_error_t network_set_interface_address(uint16_t interface_id, network_address_t *address);

// retrieve information of a given network interface
network_error_t network_get_interface_address(uint16_t interface_id, network_address_t* addr);
network_error_t network_get_interface_ip_address(uint16_t interface_id, ip_address_t* addr);
uint16_t network_get_interface_ip_netmask(uint16_t interface_id);
network_error_t network_get_mac_address(uint16_t interface_id, mac_address_t* mac_addr);
network_error_t network_get_l2_config(uint16_t interface_id, network_layer2_config_t* config);



#endif
