#ifndef SOCKET_H_
#define SOCKET_H_

#include "net_types.h"
#include "generic_types.h"

#ifndef MAX_SOCKETS
    #define MAX_SOCKETS 64
#endif

#ifndef MAX_CONNECTIONS_PER_SOCKET
    #define MAX_CONNECTIONS_PER_SOCKET 10
#endif


#define GLOBAL_SOCKET 0xFFFF

#if MAX_SOCKETS >= GLOBAL_SOCKET
    #error The number of Sockets is not allowed to exceed GLOBAL_SOCKET
#endif

#if MIN_PORT_ID < 1024
    #undef MIN_PORT_ID
#endif
#ifndef MIN_PORT_ID
    #define MIN_PORT_ID 1024
#endif

#if MAX_PORT_ID > 65535
    #undef MAX_PORT_ID
#endif
#ifndef MAX_PORT_ID
    #define MAX_PORT_ID 65535
#endif

#ifndef SOCKET_MAX_FRAME_SIZE
    #define SOCKET_MAX_FRAME_SIZE 1500
#endif

#ifndef SOCKET_RECV_QUEUE_LENGTH
    #define SOCKET_RECV_QUEUE_LENGTH 15
#endif

typedef enum 
{
    SOCKET_DOMAIN_NONE      =   0x0000, // No Domain, socket not in use
    SOCKET_DOMAIN_CONFIG    =   0x0001, // Socket is beeing configured
    SOCKET_DOMAIN_IP        =   0x0002, // Domain for IPv4 and IPv6
    SOCKET_DOMAIN_ETHERNET  =   0x0003, // Domain for using the Ethernet MAC Layer
} socket_domain_t;

typedef enum 
{
    SOCKET_TYPE_RAW         =   0x0001, // Socket for raw network access
    SOCKET_TYPE_DATAGRAM    =   0x0002, // UDP Socket
    SOCKET_TYPE_STREAM      =   0x0003, // TCP Socket
} socket_type_t;

typedef struct
{
    mac_address_t hw_address;
    ip_address_t ip_address;
    uint16_t port;
    uint16_t if_id;
} socket_address_t;

typedef struct 
{
    mac_address_t hw_address;
    ethertype_t protocol;
} socket_ethernet_t;

typedef struct 
{
    ip_protocol_t protocol;
    uint16_t src_port;
} socket_ip_t;

typedef enum
{
    SOCKET_OPTION_PRIORITY                      =   0x0001, // set priority of socket
    SOCKET_OPTION_BLOCKING                      =   0x0002, // set socket to block when receiving/transmitting data
    SOCKET_OPTION_NON_BLOCK                     =   0x0003, // set socket to non-block when receiving/transmitting data
    SOCKET_OPTION_DEBUG_LOG                     =   0x0004, // Enable/disable debug output to stdout
    SOCKET_OPTION_TCP_NO_CONCAT                 =   0x0005, // Disable Concatination of TCP messages when a message is already in transmission
// This options which are only applicable for GLOBAL_SOCKET
    SOCKET_OPTION_ACCEPT_L2_BRODCASTS           =   0x8001, // Do not discard if a frame/packet uses a Layer 2 broadcast address, this flag is only applicable for GLOBAL_SOCKET
    SOCKET_OPTION_ACCEPT_IP_BRODCASTS           =   0x8002, // Do not discard if a frame/packet uses a IPv4 broadcast address, this flag is only applicable for GLOBAL_SOCKET
    SOCKET_OPTION_MISSING_L2_ADDRESS_NO_FAIL    =   0x8003, // The transmission does not fail if the L2 address is not known, a broadcast address will be used instead, this flag is only applicable for GLOBAL_SOCKET
//    SOCKET_OPTION_SEND_ICMP_RESPONSES           =   0x8004, // Send answers to icmp echo request, send destination unreachable, this flag is only applicable for GLOBAL_SOCKET

}socket_options_t;

typedef struct
{
    uint8_t priority_set:1;
    uint8_t blocking:1;
    uint8_t debug_output:1;
    uint8_t priority;
} socket_aux_options_t;

typedef struct
{
    void *buffer;
    void *msg;
    socket_address_t addr;
    time_stamp_t ts;
    uint16_t frame_length;
} socket_msg_queue_entry_t;

typedef struct
{
    socket_msg_queue_entry_t *queue;
    uint8_t read_ptr;
    uint8_t write_ptr;
    uint8_t length;
} socket_msg_queue_t;

typedef struct 
{
    struct 
    {
        uint32_t sequence_number;
        uint32_t acknowledgement_number;
    } rx;
    
    struct 
    {
        uint32_t first_seq_num;
        uint32_t sequence_number;
        uint32_t acknowledgement_number;
    } tx;
    
    socket_address_t peer;
    tcp_state_t state;
    uint8_t* response;
    uint16_t response_len;
} socket_connection_t;

typedef struct 
{
    socket_domain_t domain;
    socket_type_t type;
    uint16_t bound_interface;
    union 
    {
        socket_ethernet_t eth;
        socket_ip_t ip;
    };
    socket_aux_options_t options;
    socket_msg_queue_t recv_msg_queue;
    socket_connection_t connections[MAX_CONNECTIONS_PER_SOCKET];
} socket_t;

typedef struct 
{
    uint8_t *space;
    uint16_t buffer_size;
    uint16_t entry_size;
    uint16_t wr_pointer;
    uint16_t rd_pointer;
} socket_buffer_t;

typedef struct 
{
    uint8_t *buffer_space;
    uint16_t buffer_entry_size;
    uint16_t buffer_size;
} socket_buffer_cfg_t;

// request a socket
unsigned short socket(socket_domain_t domain, socket_type_t type, unsigned short protocol);

// bind the socket to send and receive i.e. via TCP/UDP Port
// or bind socket to network interface for L2 communication
network_error_t socket_bind(unsigned short socket_id, socket_address_t * sock_addr);
network_error_t socket_bind_if(unsigned short socket_id, unsigned short interface_id);

// set a configuration option for a single interface or the global socket
network_error_t socket_set_option(unsigned short socket_id, socket_options_t option, void *value, uint32_t value_size);

// send a message to a given address
network_error_t socket_send(unsigned short socket_id, socket_address_t * dest_addr, unsigned char* data, unsigned short len);

// check if a message was received for the socket and return that message
network_error_t socket_receive(uint16_t socket_id, socket_address_t *src_addr, uint8_t* data, uint16_t *data_length, uint16_t buffer_length);

// check if a message was received for the socket and return that message with a timestamp when the message was received
network_error_t socket_receive_w_ts(uint16_t socket_id, socket_address_t *src_addr, uint8_t* data, uint16_t *data_length, uint16_t buffer_length, time_stamp_t *rx_ts);

// handover message from network interface to socket interface
network_error_t socket_handle_frame(uint16_t net_id, uint8_t* data, uint16_t len, time_stamp_t *rx_ts);

// return pointer to a specific socket
socket_t* socket_get(uint16_t socket_id);

// copy all the contenent of addr1 in addr2 (deep copy)
void socket_deep_copy_addr(socket_address_t* addr1, socket_address_t* addr2);

#endif
