#include <stdlib.h>
#include <stdio.h>

#include "../../inc/buffer_management.h"
#include "mmio/uniform_buffer.h"
#include "mmio/tlsf.h"
#include "sysram/uniform_buffer.h"
#include "sysram/tlsf.h"

static void * config_ptr[MAX_BUFFER_CONFIGS];
static buffer_mgmt_type_t buffer_mgmt_type_array[MAX_BUFFER_CONFIGS];
static unsigned short config_wr_ptr=0;

unsigned short buffer_management_register(buffer_management_config_t* config )
{
    // find next free entry
    char found =0;
    uint16_t buffer_mgmt_id;
    for(unsigned short i=config_wr_ptr; i < MAX_BUFFER_CONFIGS; i++)
    {
        if(buffer_mgmt_type_array[i] == BUFFER_MGMT_TYPE_NONE)
        {
            config_wr_ptr = i;
            buffer_mgmt_id=i;
            found =1;
            break;
        }
    }
    // check if buffer was found
    if(found == 0 )
    {
        for(unsigned short i=0; i < config_wr_ptr; i++)
        {
            if(buffer_mgmt_type_array[i] == BUFFER_MGMT_TYPE_NONE)
            {
                config_wr_ptr = i;
                buffer_mgmt_id=i;
                found =1;
                break;
            }
        }   
        // check if buffer was found
        if(found == 0)
        {
            return 0xFFFF;
        }
    }

    // Init DSA
    if(config->type == BUFFER_MGMT_TYPE_UNIFORM)
    {
        config_ptr[buffer_mgmt_id] = uniform_buffer_init( config);
    }
    else if (config->type == BUFFER_MGMT_TYPE_TLSF)
    {
        config_ptr[buffer_mgmt_id] = tlsf_init( config);
    }
    else if(config->type == BUFFER_MGMT_TYPE_MMIO_UNIFORM)
    {
        config_ptr[buffer_mgmt_id] = mmio_uniform_buffer_init( config);
    }
    else if (config->type == BUFFER_MGMT_TYPE_MMIO_TLSF)
    {
        config_ptr[buffer_mgmt_id] = mmio_tlsf_init( config);
    }
    else
    {
        return 0xFFFF;
    }
    buffer_mgmt_type_array[buffer_mgmt_id] = config->type;
    return buffer_mgmt_id;
}

void buffer_management_derigister(unsigned short buf_mgmt_id)
{
    // check if id is valid
    if(buf_mgmt_id < MAX_BUFFER_CONFIGS)
    {
        config_ptr[buf_mgmt_id] = NULL;
    }
}

void * buffer_management_malloc(unsigned short size, unsigned short buf_mgmt_id)
{
    void * buffer = NULL;
    // check if id is valid
    if(buf_mgmt_id >= MAX_BUFFER_CONFIGS)
    {
        return NULL;
    }
    
    // select which DSA shall be used
    switch (buffer_mgmt_type_array[buf_mgmt_id])
    {
        case BUFFER_MGMT_TYPE_UNIFORM:
            buffer = uniform_buffer_malloc(config_ptr[buf_mgmt_id]);
            break;
        case BUFFER_MGMT_TYPE_TLSF:
            buffer = tlsf_malloc(size, config_ptr[buf_mgmt_id]);
            break;
        case BUFFER_MGMT_TYPE_MMIO_UNIFORM:
            buffer = mmio_uniform_buffer_malloc(config_ptr[buf_mgmt_id]);
            break;
        case BUFFER_MGMT_TYPE_MMIO_TLSF:
            buffer = mmio_tlsf_malloc(size, config_ptr[buf_mgmt_id]);
            break;
        default:
            break;
    }

    return buffer;
}

void * buffer_management_realloc(void *  buffer_in, unsigned short size, unsigned short buf_mgmt_id)
{
    void * buffer_out = NULL;
    // check if id is valid
    if(buf_mgmt_id >= MAX_BUFFER_CONFIGS)
    {
        return NULL;
    }

    // select which DSA shall be used
    switch (buffer_mgmt_type_array[buf_mgmt_id])
    {
        case BUFFER_MGMT_TYPE_UNIFORM:
            buffer_out = uniform_buffer_realloc(buffer_in, size, config_ptr[buf_mgmt_id]);
            break;
        case BUFFER_MGMT_TYPE_TLSF:
            buffer_out = tlsf_realloc(buffer_in, size, config_ptr[buf_mgmt_id]);
            break;
        case BUFFER_MGMT_TYPE_MMIO_UNIFORM:
            buffer_out = mmio_uniform_buffer_realloc(buffer_in, size, config_ptr[buf_mgmt_id]);
            break;
        case BUFFER_MGMT_TYPE_MMIO_TLSF:
            buffer_out = mmio_tlsf_realloc(buffer_in, size, config_ptr[buf_mgmt_id]);
            break;
        default:
            break;
    }

    return buffer_out;
}

void buffer_management_free(void * buffer, unsigned short buf_mgmt_id)
{
    // check if id is valid
    if(buf_mgmt_id >= MAX_BUFFER_CONFIGS)
    {
        return;
    }
    
    // select which DSA shall be used
    switch (buffer_mgmt_type_array[buf_mgmt_id])
    {
        case BUFFER_MGMT_TYPE_UNIFORM:
            uniform_buffer_free(buffer, config_ptr[buf_mgmt_id]);
            break;
        case BUFFER_MGMT_TYPE_TLSF:
            tlsf_free(buffer, config_ptr[buf_mgmt_id]);            
            break;
        case BUFFER_MGMT_TYPE_MMIO_UNIFORM:
            mmio_uniform_buffer_free(buffer, config_ptr[buf_mgmt_id]);
            break;
        case BUFFER_MGMT_TYPE_MMIO_TLSF:
            mmio_tlsf_free(buffer, config_ptr[buf_mgmt_id]);            
            break;
        default:
            break;
    }
}


void buffer_management_memory_dump(unsigned short buf_mgmt_id)
{
    // check if id is valid
    if(buf_mgmt_id >= MAX_BUFFER_CONFIGS)
    {
        return;
    }

    // select which DSA shall be used
    switch (buffer_mgmt_type_array[buf_mgmt_id])
    {
        case BUFFER_MGMT_TYPE_UNIFORM:
        case BUFFER_MGMT_TYPE_MMIO_UNIFORM:
            break;
        case BUFFER_MGMT_TYPE_TLSF:
            tlsf_memory_dump(config_ptr[buf_mgmt_id]);            
            break;
        case BUFFER_MGMT_TYPE_MMIO_TLSF:
            mmio_tlsf_memory_dump(config_ptr[buf_mgmt_id]);            
            break;
        default:
            break;
    }
}
