/*
 * Buffer Management for libnetwork
 * 
 * Authors: Christoph Lehr
 */

#include <stdio.h>

#include "uniform_buffer.h"
#include "../../inc/generic_types.h"

#ifdef ARCH_PATMOS
    #include <machine/exceptions.h>
#endif

void * mmio_uniform_buffer_init(buffer_management_config_t* config)
{
    // copy config values
    mmio_uniformm_runtime_config_ptr_t in_mem_cfg = (mmio_uniformm_runtime_config_ptr_t) config->start_address;
    in_mem_cfg->start_address = (mmio_ptr_t) (config->start_address + sizeof(mmio_uniformm_runtime_config_t));
    in_mem_cfg->mgmt_address = in_mem_cfg->start_address;
    in_mem_cfg->end_address =  (mmio_ptr_t) config->end_address;
    in_mem_cfg->buffer_size = config->uniform_cfg.buffer_size;

    // create initial buffer
    mmio_ptr_t buffer_prev = in_mem_cfg->start_address;
    in_mem_cfg->free_list_first = (mmio_uniform_block_header_ptr_t)buffer_prev;

    // add offset for alignment
    in_mem_cfg->start_address -= ((uint64_t)in_mem_cfg->start_address) % sizeof(mmio_ptr_t);
    in_mem_cfg->start_address += sizeof(mmio_ptr_t);

    mmio_ptr_t buffer = in_mem_cfg->start_address + in_mem_cfg->buffer_size;
    // iterate until no full buffer has place
    mmio_ptr_t end_of_memory_space = (mmio_ptr_t) (in_mem_cfg->end_address - in_mem_cfg->buffer_size);  
    while(buffer_prev < end_of_memory_space)
    {
        mmio_uniform_block_header_ptr_t bh_prev = (mmio_uniform_block_header_ptr_t) buffer_prev;
        bh_prev->next = (mmio_uniform_block_header_ptr_t) buffer;
        in_mem_cfg->free_list_last = bh_prev->next;
        buffer_prev = buffer;
        buffer += in_mem_cfg->buffer_size;
    }
    mmio_uniform_block_header_ptr_t bh_prev = (mmio_uniform_block_header_ptr_t) buffer_prev;
    bh_prev->next = NULL;
    
    return (void *) in_mem_cfg;
}

void * mmio_uniform_buffer_malloc(void * config_ptr ) 
{
    mmio_uniformm_runtime_config_ptr_t config = (mmio_uniformm_runtime_config_ptr_t) config_ptr;
    
    if(config->free_list_first == NULL )
    {
        return NULL;
    }

    #ifdef ARCH_PATMOS
    // check whether interrupts are enabled
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }
    #endif

    mmio_uniform_block_header_ptr_t  buffer = config->free_list_first;
    config->free_list_first = buffer->next;
    if(buffer->next == NULL)
    {
        config->free_list_last = NULL;
    }

    #ifdef ARCH_PATMOS
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }
    #endif

    return (void *) buffer;
}


void mmio_uniform_buffer_free(void * buffer, void * config_ptr ) 
{
    mmio_uniformm_runtime_config_ptr_t config = (mmio_uniformm_runtime_config_ptr_t) config_ptr;
    // check whether interrupts are enabled
    #ifdef ARCH_PATMOS
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }
    #endif

    mmio_uniform_block_header_ptr_t header = (mmio_uniform_block_header_ptr_t) buffer; 
    header->next = NULL;

    if(config->free_list_last == NULL)
    {
        config->free_list_first = header;
        config->free_list_last = header;
    }   
    else
    {
        mmio_uniform_block_header_ptr_t old_last_buffer = config->free_list_last; 
        old_last_buffer->next = header;
        config->free_list_last = header;
    }

    #ifdef ARCH_PATMOS
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }
    #endif

    return;
}


// reallocate part of a buffer 
void * mmio_uniform_buffer_realloc(void * buffer, uint16_t size, void * config_ptr)
{
    return NULL;
}
