#include <stdio.h>
#include <stdlib.h>
#include "tlsf.h"
#include "../../inc/generic_types.h"
#include "../../inc/utils.h"

#ifdef ARCH_PATMOS
	#include <machine/exceptions.h>
#endif

static uint16_t get_fli_sli(uint16_t size, tlsf_runtime_config_prt_t config);
static uint16_t get_next_free_fli_sli(uint16_t size, tlsf_runtime_config_prt_t config);
static void remove_buffer_from_free_list(block_header_free_tlsf_ptr_t buffer, tlsf_runtime_config_prt_t config);
static void add_buffer_to_free_list(block_header_free_tlsf_ptr_t  buffer, tlsf_runtime_config_prt_t config);
static block_header_free_tlsf_ptr_t split_buffer(uint16_t size, block_header_free_tlsf_ptr_t buffer, tlsf_runtime_config_prt_t config);
static void merge_buffers(block_header_free_tlsf_ptr_t buffer, tlsf_runtime_config_prt_t config );
static block_header_free_tlsf_ptr_t get_buffer(uint16_t size, tlsf_runtime_config_prt_t config);

void * tlsf_init(buffer_management_config_t *config)
{
	if( (1 << config->tlsf_cfg.fl_lowest_cls) + sizeof(block_header_tlsf_t) < sizeof(block_header_free_tlsf_t)  )
	{
		return NULL;
	}

	// create config at start of managed memory
	tlsf_runtime_config_prt_t in_mem_cfg = (tlsf_runtime_config_prt_t) config->start_address;
	
	// adapt addresses
    in_mem_cfg->start_address = (ptr_t) (config->start_address + sizeof(tlsf_runtime_config_t));
    in_mem_cfg->end_address = (ptr_t)  config->end_address;
	
	// copy configuration values
	in_mem_cfg->fl_cls_cnt = config->tlsf_cfg.fl_cls_cnt;
	in_mem_cfg->sl_cls_cnt = config->tlsf_cfg.sl_cls_cnt;
	in_mem_cfg->fl_lowest_cls = config->tlsf_cfg.fl_lowest_cls;
	
	in_mem_cfg->sl_free = (ptr_16_t) in_mem_cfg->start_address;
    
	// add 32/64 bit alignment for a more efficient memory access 
    uint64_t offset = sizeof(uint16_t) * (uint32_t)in_mem_cfg->fl_cls_cnt * (uint32_t)(1 << in_mem_cfg->sl_cls_cnt);
    if((offset & (sizeof(ptr_t)-1)) != 0) { 
		offset += sizeof(ptr_t) - (offset & (sizeof(ptr_t)-1)); 
	}
    in_mem_cfg->start_address = in_mem_cfg->start_address + offset;
	
	// set pointers for free lists
    offset = in_mem_cfg->fl_cls_cnt * (1<<in_mem_cfg->sl_cls_cnt) * sizeof(block_header_free_tlsf_ptr_t);
	in_mem_cfg->free_list_first = (block_header_free_tlsf_ptr_t*) in_mem_cfg->start_address;
    in_mem_cfg->start_address 	= in_mem_cfg->start_address + offset;
    in_mem_cfg->free_list_last 	= (block_header_free_tlsf_ptr_t*) in_mem_cfg->start_address;
	in_mem_cfg->start_address 	= in_mem_cfg->start_address + offset;
	
	// Set init values
	in_mem_cfg->fl_free = 0;
	uint16_t free_list_id = 0;
	
	#pragma loopbound min 1 max 16
	for(uint16_t i=0; i < in_mem_cfg->fl_cls_cnt; i++)
	{	
		#pragma loopbound min 1 max 16
		for(uint16_t j=0; j <  (1 << in_mem_cfg->sl_cls_cnt); j++)
		{
			in_mem_cfg->free_list_first[free_list_id]=NULL;
			in_mem_cfg->free_list_last[free_list_id]=NULL;
			free_list_id++;
		}
		
		in_mem_cfg->sl_free[i] = 0;
	}
	
	// calculate offset to allign initial buffer for optimal memory access
	offset  = (uint64_t)in_mem_cfg->start_address ;
	offset &= (sizeof(ptr_t)-1);
    if(offset != 0) { 
		offset += sizeof(ptr_t) - (offset & (sizeof(ptr_t)-1)); 
	}

	// create initial buffer
	block_header_free_tlsf_ptr_t buffer = (block_header_free_tlsf_ptr_t) (in_mem_cfg->start_address + offset);
	uint16_t size = (uint16_t) ((ptr_t)in_mem_cfg->end_address - (ptr_t) in_mem_cfg->start_address- sizeof(block_header_tlsf_t));
	buffer->bh.size = (size >> 2);
	buffer->bh.alloc = 0;
	buffer->bh.last_block=1;
	buffer->bh.size_prev=0;
	
	add_buffer_to_free_list(buffer, in_mem_cfg);
	
	return (void *) in_mem_cfg;
}

uint16_t get_fli_sli(uint16_t size, tlsf_runtime_config_prt_t config)
{
	uint16_t fli = 16 - utils_nlz_16(size);
	uint16_t shift_amount = (fli - config->sl_cls_cnt-1);
	uint16_t tmp = (size) >> shift_amount;
	uint16_t sli = tmp & ((1 << config->sl_cls_cnt)-1);
	fli -= config->fl_lowest_cls;
	fli--;
	return (fli << 8) | (sli & 0xFF);
}

uint16_t get_next_free_fli_sli(uint16_t size, tlsf_runtime_config_prt_t config)
{	
	uint16_t fli=0;
	uint16_t sli=0;
	
	if(size > (1 << config->fl_lowest_cls))
	{
		fli = 16 - utils_nlz_16(size);
		uint16_t shift_amount = (fli - config->sl_cls_cnt);
		uint16_t tmp = (size) >> shift_amount;
		sli = tmp & ((1 << config->sl_cls_cnt));
		fli -= config->fl_lowest_cls;
		fli--;
	}

	// calculate masks if no free buffers in computed 
	// fli and sli 
	uint16_t new_fli_mask =  ~((1 << (fli+1)) - 1);
	uint16_t new_sli_mask =  ~((1 << (sli+1)) - 1);
	// check if there is no entry in calculated sli
	// but in a larger sli with current fli
	if(	(config->fl_free & (1 << fli)) != 0 &&
		(config->sl_free[fli] & 1 << sli )== 0 &&
		(config->sl_free[fli] & new_sli_mask) > 0
	)
	{
		uint16_t tmp = config->sl_free[fli] & new_sli_mask; 
		tmp &= (~tmp)+1; 
		sli = 16 - utils_nlz_16(tmp)-1;	
	}
	// check if there is a free buffer in a larger fli
	else if((config->fl_free & new_fli_mask) > 0)
	{
		// calculate new fli
		uint16_t tmp_fli = config->fl_free & new_fli_mask; 
		tmp_fli &= (~tmp_fli)+1; 
		fli = 16 - utils_nlz_16(tmp_fli)-1;
		
		// calculate new sli
		uint16_t tmp_sli = config->sl_free[fli];
		tmp_sli &= (~tmp_sli)+1; 
		sli = 16 - utils_nlz_16(tmp_sli)-1;
	}
	// check if there is no buffer left of at least required size
	else if ((config->fl_free & new_fli_mask) ==0 )
	{
		// no buffer free
		return 0xFFFF;
	}
	
	// merge fli and sli into single number
	return (fli << 8) | (sli & 0xFF);
}

void remove_buffer_from_free_list(block_header_free_tlsf_ptr_t buffer, tlsf_runtime_config_prt_t config)
{
	#ifdef ARCH_PATMOS
    // check whether interrupts are enabled
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }
    #endif

	block_header_free_tlsf_ptr_t prev_buffer = buffer->prev;
	block_header_free_tlsf_ptr_t next_buffer = buffer->next;
	
	uint16_t sli = get_fli_sli(buffer->bh.size<<2, config);
	uint16_t fli = (sli >> 8) & 0xFF; 
	sli &= 0xFF;
	
	uint16_t free_list_id = fli*( 1 << (uint16_t)config->sl_cls_cnt);
	free_list_id += sli;
	
	if(prev_buffer == NULL && next_buffer == NULL)
	{ 
		config->sl_free[fli] &= ~(1 << sli);
		config->free_list_first[free_list_id]	=NULL;
		config->free_list_last[free_list_id]	=NULL;
	}
	else if (prev_buffer == NULL)
	{
		config->free_list_first[free_list_id]=next_buffer;
		next_buffer->prev = NULL;
	}
	else if (next_buffer == NULL)
	{
		config->free_list_last[free_list_id]=prev_buffer;
		prev_buffer->next = NULL;
	}
	else
	{
		prev_buffer->next = next_buffer;
		next_buffer->prev = prev_buffer;
	}
	if(config->sl_free[fli] == 0) 
    {
        config->fl_free &= ~(1 << fli);
    }
	buffer->prev = NULL;
	buffer->next = NULL;

	#ifdef ARCH_PATMOS
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }
    #endif
}

void add_buffer_to_free_list(block_header_free_tlsf_ptr_t  buffer, tlsf_runtime_config_prt_t config)
{
	#ifdef ARCH_PATMOS
    // check whether interrupts are enabled
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }
    #endif

	uint16_t sli = get_fli_sli(buffer->bh.size<<2, config);
	uint16_t fli = (sli >> 8) & 0xFF; 
	sli &= 0xFF;

	uint16_t  free_list_id = fli*( 1 << (uint16_t )config->sl_cls_cnt);
	free_list_id += sli;
	
	// check if free list is empty
	if(config->free_list_last[free_list_id] == NULL || 
		config->free_list_first[free_list_id] == NULL)
	{
		config->free_list_first[free_list_id]= buffer;
		buffer->prev = NULL;
	}
	else
	{
		buffer->prev = config->free_list_last[free_list_id];
		config->free_list_last[free_list_id]->next = buffer;
	}
	
	// update last entry
    config->free_list_last[free_list_id] = buffer;
		
	buffer->next = NULL;

	// set availability bits
	config->fl_free   |= (1 << fli);
	config->sl_free[fli] |= (1 << sli);

	#ifdef ARCH_PATMOS
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }
    #endif
}

block_header_free_tlsf_ptr_t split_buffer(uint16_t size, block_header_free_tlsf_ptr_t buffer, tlsf_runtime_config_prt_t config)
{
	// calculate size of new buffer
	uint16_t new_buffer_size = buffer->bh.size << 2;
	// align size to memory address
	if (size <  (1 << config->fl_lowest_cls))
	{
		size = (1 << config->fl_lowest_cls);
	}
	else if(size & (sizeof(void *) -1))
	{
		size &= ~(sizeof(void *) -1);
		size += sizeof(void *);
	}
	
	new_buffer_size -= size;
	new_buffer_size -= sizeof(block_header_tlsf_t);

	// calculate address of new block
	ptr_t new_buffer_address = (ptr_t) buffer;
	new_buffer_address += sizeof(block_header_tlsf_t);
	new_buffer_address += size;
	block_header_free_tlsf_ptr_t buffer_new = (block_header_free_tlsf_ptr_t) new_buffer_address;

	// set sizes of this and new block
	buffer->bh.size = size >> 2; // -= (new_buffer_size >> 2);
	buffer_new->bh.size = (new_buffer_size >> 2);
	buffer_new->bh.size_prev = size >> 2;
	buffer_new->bh.last_block = 0;

	buffer_new->prev = NULL;
	buffer_new->next = NULL;
	buffer_new->bh.alloc = 0;
	
	// check if this is the last block in the physical memory
	if( buffer->bh.last_block == 0)
	{

		#ifdef ARCH_PATMOS
		// check whether interrupts are enabled
		unsigned intr_en = EXC_STATUS & 0x1;
		if(intr_en)
		{
			CLI();
		}
		#endif
		// get address of next block after new block
		ptr_t next_buffer_address = (ptr_t) buffer_new;
		next_buffer_address += buffer_new->bh.size << 2;
		next_buffer_address += sizeof(block_header_tlsf_t);
		block_header_free_tlsf_ptr_t buffer_next = (block_header_free_tlsf_ptr_t) next_buffer_address;
		
		// adapt relative addressing in next block
		buffer_next->bh.size_prev = buffer_new->bh.size;
		
		#ifdef ARCH_PATMOS
		if(intr_en)
		{
			SEI();
		}
		#endif
	}
	else 
	{
		// set that new buffer is new last block
		buffer_new->bh.last_block = 1;
		buffer->bh.last_block = 0;
	}
	
	return buffer_new;	
}

void merge_buffers(block_header_free_tlsf_ptr_t buffer, tlsf_runtime_config_prt_t config )
{
	block_header_free_tlsf_ptr_t buffer_merged = buffer;
	

	ptr_t prev_buffer_address = (ptr_t)buffer;
	prev_buffer_address -= buffer->bh.size_prev << 2;
	prev_buffer_address -= sizeof(block_header_tlsf_t);
	block_header_free_tlsf_ptr_t buffer_prev = (block_header_free_tlsf_ptr_t ) prev_buffer_address;

	ptr_t next_buffer_address = (ptr_t)buffer;
	next_buffer_address += buffer->bh.size <<2 ;
	next_buffer_address += sizeof(block_header_tlsf_t);

	block_header_free_tlsf_ptr_t buffer_next = (block_header_free_tlsf_ptr_t ) next_buffer_address;
	
	#ifdef ARCH_PATMOS
	// check whether interrupts are enabled
	unsigned intr_en = EXC_STATUS & 0x1;
	if(intr_en)
	{
		CLI();
	}
	#endif

	if(buffer->bh.size_prev > 0 && buffer_prev->bh.alloc ==0 )
	{
		remove_buffer_from_free_list(buffer_prev, config);

		// adjust size of previous buffer
		buffer_prev->bh.size += buffer->bh.size;
		buffer_prev->bh.size += sizeof(block_header_tlsf_t)>> 2;
		// adjust next buffer
		buffer_next->bh.size_prev = buffer_prev->bh.size;
		buffer_merged = buffer_prev;
		buffer_merged->bh.last_block = buffer->bh.last_block;
	}

	if(buffer->bh.last_block == 0 && buffer_next->bh.alloc == 0 )
	{
		remove_buffer_from_free_list(buffer_next, config);
		

		// adjust size of previous buffer
		buffer_merged->bh.size += buffer_next->bh.size;
		buffer_merged->bh.size += (sizeof(block_header_tlsf_t) >> 2);
		buffer_merged->bh.last_block = buffer_next->bh.last_block;

		if(buffer_next->bh.last_block == 0)
		{
			// adjust buffer after next buffer
			next_buffer_address += buffer_next->bh.size << 2;
			next_buffer_address += sizeof(block_header_tlsf_t);
			buffer_next = (block_header_free_tlsf_ptr_t ) next_buffer_address;
			buffer_next->bh.size_prev = buffer_merged->bh.size;
		}
	}
	
	#ifdef ARCH_PATMOS
	if(intr_en)
	{
		SEI();
	}
	#endif

	add_buffer_to_free_list(buffer_merged, config);
}

block_header_free_tlsf_ptr_t get_buffer(uint16_t size, tlsf_runtime_config_prt_t config)
{
	
	// TODO: refactor this possbile race between get free fli sli combination
	// and removing buffer from free list and adapting free list identifiers
	#ifdef ARCH_PATMOS
    // check whether interrupts are enabled
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }
    #endif

	block_header_free_tlsf_ptr_t buffer = NULL;
	uint16_t  sli = get_next_free_fli_sli(size, config);
	if(sli != 0xFFFF )
	{

		uint16_t  fli = sli >> 8;
		sli &= 0xFF;
		
		uint16_t  free_list_id = fli* (1 << (uint16_t )config->sl_cls_cnt);
		free_list_id += sli;
		buffer = config->free_list_first[free_list_id];

		// check if state of free list has changed between allocation of buffer
		// and calculation of indexes
		if(buffer != NULL)
		{
			buffer->bh.alloc =1;

			remove_buffer_from_free_list(buffer, config);
		}

	}
	
	#ifdef ARCH_PATMOS
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }
    #endif

	return buffer;
}

void *tlsf_malloc(uint16_t size, void * config_ptr)
{	
	tlsf_runtime_config_prt_t config = (tlsf_runtime_config_prt_t) config_ptr;
	#ifdef ARCH_PATMOS
	// I don't know why, but the interrupts need to be disabled for nearly the whole 
	// routine, otherwise patmos crashes 

	// check whether interrupts are enabled
	unsigned intr_en = EXC_STATUS & 0x1;
	if(intr_en)
	{
		CLI();
	}
	#endif

	uint32_t  fl_lowest_cls = 1<<(config->fl_lowest_cls );
	if(size < fl_lowest_cls )
	{
		size = fl_lowest_cls;
	}

	block_header_free_tlsf_ptr_t buffer = get_buffer(size, config);
	
	if(buffer == NULL)
	{
		#ifdef ARCH_PATMOS
		// check whether interrupts need to be reenabled
		if(intr_en)
		{
			SEI();
		}
		#endif
		return NULL;
	}

	uint32_t  min_space_required = fl_lowest_cls;
	min_space_required += sizeof(block_header_tlsf_t);

	if(size & (sizeof(void *) -1))
	{
		// add additional space required for memory address alignment
		min_space_required += sizeof(void *);
		min_space_required -= size & (sizeof(void *) -1);
	}
	
	uint16_t  buffer_size = buffer->bh.size<<2;
	
	if(buffer_size > size + min_space_required)
	{
		block_header_free_tlsf_ptr_t new_buffer = split_buffer(size, buffer, config);	
		add_buffer_to_free_list(new_buffer, config);
	}
	
	#ifdef ARCH_PATMOS
	if(intr_en)
	{
		SEI();
	}
	#endif
	
	// check whether interrupts need to be reenabled
	ptr_t buffer_address = (ptr_t) buffer;
	buffer_address+= sizeof(block_header_tlsf_t);
	

	return (void *) buffer_address;
}


// reallocate a part of a buffer, returns NULL if not possible 
void * tlsf_realloc(void * address, uint16_t size, void * config_ptr)
{
	tlsf_runtime_config_prt_t config = (tlsf_runtime_config_prt_t) config_ptr;

	uint32_t  fl_lowest_cls = (uint32_t) (1<<(config->fl_lowest_cls));
	uint32_t min_space_required = fl_lowest_cls;
	min_space_required += (uint32_t) sizeof(block_header_tlsf_t);
	min_space_required += min_space_required % sizeof(void *); // add overset for proper alignment

	ptr_t buffer_address = (ptr_t) address;
	buffer_address -= sizeof(block_header_tlsf_t);
	block_header_free_tlsf_ptr_t  buffer = (block_header_free_tlsf_ptr_t  )buffer_address;

	if(size < fl_lowest_cls)
	{
		size = fl_lowest_cls;
	}

	if(size & (sizeof(void *) -1))
	{
		// add additional space required for memory address alignment
		min_space_required += sizeof(void *);
		min_space_required -= size & (sizeof(void *) -1);
	}

	if(((buffer->bh.size << 2) - size) < min_space_required)
	{
		return NULL;
	}


	ptr_t next_buffer_address = (ptr_t) address;
	next_buffer_address += (buffer->bh.size<< 2);
	block_header_free_tlsf_ptr_t  buffer_next = (block_header_free_tlsf_ptr_t  ) next_buffer_address;

	block_header_free_tlsf_ptr_t new_buffer = split_buffer(size, buffer, config);	


	#ifdef ARCH_PATMOS
	// check whether interrupts are enabled
	unsigned intr_en = EXC_STATUS & 0x1;
	if(intr_en)
	{
		CLI();
	}
	#endif

	if( buffer->bh.last_block == 0 && buffer_next->bh.alloc == 0 )
	{
		remove_buffer_from_free_list(buffer_next, config);
		
		// adjust size of previous buffer
		new_buffer->bh.size += buffer_next->bh.size;
		new_buffer->bh.size += (sizeof(block_header_tlsf_t) >> 2);
		new_buffer->bh.last_block = buffer_next->bh.last_block;

		if(buffer_next->bh.last_block == 0)
		{
			// adjust buffer after next buffer
			next_buffer_address += buffer_next->bh.size << 2;
			next_buffer_address += sizeof(block_header_tlsf_t);
			buffer_next = (block_header_free_tlsf_ptr_t ) next_buffer_address;
			buffer_next->bh.size_prev = new_buffer->bh.size;
		}
	}
	
	#ifdef ARCH_PATMOS
	if(intr_en)
	{
		SEI();
	}
	#endif

	add_buffer_to_free_list(new_buffer, config);

	return address;
}

void tlsf_free(void * address, void * config_ptr)
{	
	#ifdef ARCH_PATMOS
	// check whether interrupts are enabled
	unsigned intr_en = EXC_STATUS & 0x1;
	if(intr_en)
	{
		CLI();
	}
	#endif

	tlsf_runtime_config_prt_t config = (tlsf_runtime_config_prt_t) config_ptr;
	ptr_t buffer_address = (ptr_t) address;
	buffer_address -= sizeof(block_header_tlsf_t);
	block_header_free_tlsf_ptr_t  buffer = (block_header_free_tlsf_ptr_t  )buffer_address;
	buffer->bh.alloc = 0;

	ptr_t prev_buffer_address = (ptr_t)buffer_address;
	prev_buffer_address -= (buffer->bh.size_prev << 2);
	prev_buffer_address -= sizeof(block_header_tlsf_t);

	block_header_free_tlsf_ptr_t  buffer_prev =(block_header_free_tlsf_ptr_t  ) prev_buffer_address;

	ptr_t next_buffer_address = (ptr_t) address;
	next_buffer_address += (buffer->bh.size<< 2);
	block_header_free_tlsf_ptr_t  buffer_next = (block_header_free_tlsf_ptr_t  ) next_buffer_address;

	if( (buffer->bh.size_prev > 0   && buffer_prev->bh.alloc == 0 ) || 
		(buffer->bh.last_block == 0 && buffer_next->bh.alloc == 0 ))
	{
		merge_buffers((block_header_free_tlsf_ptr_t )buffer, config);
	}
	else
	{
		add_buffer_to_free_list((block_header_free_tlsf_ptr_t )buffer, config);
	}

	#ifdef ARCH_PATMOS
	if(intr_en)
	{
		SEI();
	}
	#endif
}

// GCOVR_EXCL_START Code coverage exclusion marker start
void tlsf_memory_dump(void * config_ptr)
{	
	tlsf_runtime_config_prt_t config = (tlsf_runtime_config_prt_t) config_ptr;

	uint32_t  min_space_required = (uint32_t) (1<<(config->fl_lowest_cls));
	min_space_required += (uint32_t) sizeof(block_header_tlsf_t);
	min_space_required += min_space_required % sizeof(void *); // add overset for proper alignment
    utils_log("min_space_required %u \n", min_space_required );

    utils_log("\nConfig Memory Address:\n");
    utils_log("config: %p\n", (void *) config);
    utils_log("config->start_address: %p\n", (void *) config->start_address);
    utils_log("config->end_address: %p\n", (void *) config->end_address);
    utils_log("config->sl_free: %p\n", (void *) config->sl_free);
    utils_log("config->free_list_first: %p\n",(void *) config->free_list_first);
    utils_log("config->free_list_last: %p\n", (void *) config->free_list_last);

    utils_log("\nRuntime Values:\n");
    utils_log("config->fl_free: %x\n", config->fl_free);
    for(uint16_t i=0; i < config->fl_cls_cnt; i++)
    {
        utils_log("config->sl_free[%u]: %x\n", i, config->sl_free[i]);
		uint32_t fl_class_size_bottom =	(1 << (i+ config->fl_lowest_cls));
		for(uint16_t j=0; j < (1<<config->sl_cls_cnt); j++)
        {
			uint32_t start_buffer_size = fl_class_size_bottom + ( j << (config->sl_cls_cnt +i+1)) ; 
			uint32_t end_buffer_size = fl_class_size_bottom + ( (j+1) << (config->sl_cls_cnt +i+1)) -1; 
			uint32_t free_list_id = i* (uint32_t) ( 1 << (uint16_t)config->sl_cls_cnt) + j;
			uint8_t buffer_available = (config->sl_free[i] & (1 << j)) > 0;
			utils_log("\tFLI:%2u, SLI:%2u, Buffer available: %u ", i, j, buffer_available );
			utils_log("Freelist %u Buffers [%u,%u] \n", free_list_id, start_buffer_size, end_buffer_size);
        }
	}
    
    for(uint16_t i=0; i < config->fl_cls_cnt; i++)
    {
        uint32_t free_list_id = i* (uint32_t) ( 1 << (uint16_t)config->sl_cls_cnt);
        for(uint16_t j=0; j < (1<<config->sl_cls_cnt); j++)
        {
            utils_log("config->free_list_first[%3u]: %p, ", free_list_id, config->free_list_first[free_list_id]);
            utils_log("config->free_list_last[%3u] : %p\n", free_list_id, config->free_list_last[free_list_id]);
			free_list_id++;
        }
    }

	utils_log("\nMemory Content:\n");
	ptr_t buffer_address = config->start_address;
	uint16_t lastBlockRead = 0;
	 
	uint32_t buffer_cnt =0;
	while (lastBlockRead == 0)
	{
		block_header_free_tlsf_ptr_t buffer = (block_header_free_tlsf_ptr_t )buffer_address;
		utils_log("Buffer No. %u, address: %p, alloc: %u, last_block: %u, size: %u, size_prev: %u ", 
			buffer_cnt, buffer, buffer->bh.alloc, buffer->bh.last_block, buffer->bh.size<<2, buffer->bh.size_prev<<2
		);

		if(buffer->bh.alloc == 0)
		{
			block_header_free_tlsf_ptr_t  free_buffer = (block_header_free_tlsf_ptr_t )buffer;
			utils_log("freelist prev: %p, freelist next %p\n", 
				free_buffer->prev, free_buffer->next
			);
		}
		else
		{
			/*
			ptr_t data = buffer_address;
			data+= sizeof(block_header_tlsf_t); 
			for(uint32_t i=0; i < (uint32_t) (buffer->bh.size << 2); i++)
			{
				if(i % 16 == 0)
				{
					utils_log("\n\t");
				}
				utils_log("0x%x ", data[i]);
			}
			*/
			utils_log("\n");
		}
		if(buffer->bh.last_block == 1)
		{
			lastBlockRead = 0xFF;
		}
		buffer_address += sizeof(block_header_tlsf_t) + (buffer->bh.size << 2); 	
		buffer_cnt++;
	}
}
// GCOVR_EXCL_STOP overage exclusion marker end