/*
 * Driver for the Ethernet MAC
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 *          Christoph Lehr
 */

#ifdef ARCH_PATMOS

#include <stdio.h>

#include <machine/exceptions.h>
#include <machine/rtc.h>

#include "../mmio_utils.h"
#include "../../../inc/device.h"
#include "../../../inc/generic_types.h"
#include "../../../inc/net_types.h"
#include "../../../inc/buffer_management.h"
#include "patmos_eth_mac.h"

#ifndef PATMOS_ETH_MAC_CNT
    #define PATMOS_ETH_MAC_CNT 2
#endif

patmos_eth_mac_config_ptr_t configurations[PATMOS_ETH_MAC_CNT];
static uint8_t interface_cnt = 0;

///////////////////////////////////////////////////////////////
// Helper Function Prototypes 
///////////////////////////////////////////////////////////////

static network_error_t enqueue_tx_frame(mmio_ptr_t address, uint16_t frame_length, patmos_eth_mac_config_ptr_t config);

static network_error_t dequeue_tx_frame(queue_entry_t *entry, patmos_eth_mac_config_ptr_t config);

static char tx_queue_full(patmos_eth_mac_config_ptr_t config);

static char tx_queue_empty(patmos_eth_mac_config_ptr_t config);

static uint32_t get_next_tx_frame_length(patmos_eth_mac_config_ptr_t config);

static mmio_ptr_t get_next_tx_address(patmos_eth_mac_config_ptr_t config);

static mmio_ptr_t get_last_tx_address(patmos_eth_mac_config_ptr_t config);

static network_error_t enqueue_rx_frame(mmio_ptr_t address, uint16_t frame_length, patmos_eth_mac_config_ptr_t config);

static network_error_t dequeue_rx_frame(queue_entry_t *entry, patmos_eth_mac_config_ptr_t config);

static char rx_queue_full(patmos_eth_mac_config_ptr_t config);

static char rx_queue_empty(patmos_eth_mac_config_ptr_t config);

static uint32_t get_next_rx_frame_length(patmos_eth_mac_config_ptr_t config);

static mmio_ptr_t get_next_rx_address(patmos_eth_mac_config_ptr_t config);

///////////////////////////////////////////////////////////////
//High level functions 
///////////////////////////////////////////////////////////////

static void do_interrupt_stuff(void)
{
    patmos_eth_mac_config_ptr_t config =configurations[0];
    
    uint32_t interrupt_id= exc_get_source();
    #pragma loopbound min 1 max PATMOS_ETH_MAC_CNT
    for(uint8_t i=0;i < interface_cnt; i++)
    {  
        if(configurations[i] != NULL && configurations[i]->interupt_id == interrupt_id)
        {
            config = configurations[i];
            break;
        }
    }
    
    if(config != NULL)
    {   

        uint32_t interrupt_src = mmio_rd_32(config->start_register_address,INT_SOURCE_ADDR);
        // Check if interrupt was caused by transmission of a frame
        // and execute corresponding update routing
        if(interrupt_src & INT_SOURCE_TXB_BIT)
        {
            eth_mac_update_transmission((mmio_ptr_t) config);
            mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_TXB_BIT);
        }

        // Check if interrupt was caused by reception  of a frame
        // and execute corresponding update routing
        if(interrupt_src & INT_SOURCE_RXB_BIT)
        {
            eth_mac_update_reception((mmio_ptr_t) config);
            mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_RXB_BIT);
        }
    }
}

void eth_mac_handle_interrupt(void) __attribute__((naked));
void eth_mac_handle_interrupt(void)
{
    exc_prologue();

    do_interrupt_stuff();
    
    exc_epilogue();
} 



/**
 * @brief Initialize Ethernet MAC 
 *  
 * This function initializes the MAC Address and the Receive Buffer,
 * Configures the Ethernet MAC
 * 
 * @param char enable_int   : set to 1 to enable interrupt driven 
 *                            operation of ethernet driver.
 * 
 */
mmio_ptr_t eth_mac_initialize(device_config_t *config){ 
    // copy data into in memory configuration
    patmos_eth_mac_config_t _IODEV *in_mem_cfg = (patmos_eth_mac_config_t _IODEV *) config->buffer_mgmt_config.start_address;
    mmio_ptr_t start_address = (mmio_ptr_t) config->buffer_mgmt_config.start_address;
    
    // copy memory addresses
    in_mem_cfg->start_register_address = (mmio_ptr_t) config->mmio_config.start_register_address;
    in_mem_cfg->start_memory_address = start_address;

    // copy configuration
    in_mem_cfg->interupt_id = config->interupt_id;
    in_mem_cfg->interupt_ena = config->interupt_ena;
    in_mem_cfg->queue_length = config->queue_length;
    if(config->rx_buffers == 0 || config->rx_buffers > DEVICE_MAX_RX_BUFFERS)
    {
        in_mem_cfg->num_rx_buffers = 1;
    }
    else
    {
        in_mem_cfg->num_rx_buffers = config->rx_buffers;
    }

    // init variables
    in_mem_cfg->tx_queue_frames_read  = 0;
    in_mem_cfg->tx_queue_frames_write = 0;
    in_mem_cfg->rx_queue_frames_read  = 0;
    in_mem_cfg->rx_queue_frames_write = 0;
    	

    // copy mac address
    in_mem_cfg->mac_address.address[0] = config->mac_address.address[0];
    in_mem_cfg->mac_address.address[1] = config->mac_address.address[1];
    in_mem_cfg->mac_address.address[2] = config->mac_address.address[2];
    in_mem_cfg->mac_address.address[3] = config->mac_address.address[3];
    in_mem_cfg->mac_address.address[4] = config->mac_address.address[4];
    in_mem_cfg->mac_address.address[5] = config->mac_address.address[5];

    
    // reserve space for queuing frames
    start_address += sizeof(patmos_eth_mac_config_t);
    in_mem_cfg->tx_queue_frames = (queue_entry_ptr_t) start_address;
    start_address += sizeof(queue_entry_t) * in_mem_cfg->queue_length;
    in_mem_cfg->rx_queue_frames = (queue_entry_ptr_t) start_address;
    start_address += sizeof(queue_entry_t) * in_mem_cfg->queue_length;

    for(uint16_t i=0; i< in_mem_cfg->queue_length; i++)
    {
        in_mem_cfg->tx_queue_frames[i].frame_length = 0;
        in_mem_cfg->rx_queue_frames[i].frame_length = 0;
    }
    
    // Set MAC Address
	uint32_t reg_val = (config->mac_address.address[2] << 24);
    reg_val |= (config->mac_address.address[3] << 16);
    reg_val |= (config->mac_address.address[4] << 8);
    reg_val |= (config->mac_address.address[5]);
    
    mmio_wr_32(in_mem_cfg->start_register_address, MAC_ADDR0, reg_val);
    reg_val  = (config->mac_address.address[0] << 8 );
    reg_val |= (config->mac_address.address[1]);
	mmio_wr_32(in_mem_cfg->start_register_address, MAC_ADDR1, reg_val);

	
    //MODEREG: PAD|HUGEN|CRCEN|DLYCRCEN|-|FULLD|EXDFREN|NOBCKOF|LOOPBCK|IFG|PRO|IAM|BRO|NOPRE|TXEN|RXEN
    reg_val  = PAD_BIT | CRCEN_BIT | FULLD_BIT | PRO_BIT | BRO_BIT | TXEN_BIT | RXEN_BIT ; // == 0x0000A423
    mmio_wr_32(in_mem_cfg->start_register_address,MODER_ADDR, reg_val);
    
    // set new start address and register new buffer management entry 
    config->buffer_mgmt_config.start_address = (void *) start_address;
    in_mem_cfg->buffer_mgmt_id = buffer_management_register(&config->buffer_mgmt_config);
    
    uint32_t rx_bd_addr= RX_BD_ADDR_BASE(mmio_rd_32(in_mem_cfg->start_register_address, TX_BD_NUM_ADDR));

    reg_val = RX_BD_EMPTY_BIT | RX_BD_IRQEN_BIT;
    for(uint16_t i=0; i< in_mem_cfg->num_rx_buffers; i++)
    {
        // initialise default RX Buffer descriptor and request buffer from Buffer Management
        mmio_ptr_t rx_buffer = (mmio_ptr_t) buffer_management_malloc(ETHERNET_MAX_FRAME_SIZE, in_mem_cfg->buffer_mgmt_id);
        if(rx_buffer == NULL)
        {
            return NULL;
        }

        mmio_wr_32(in_mem_cfg->start_register_address, rx_bd_addr +4, (uint32_t) rx_buffer); 

        if(i == in_mem_cfg->num_rx_buffers-1)
        {
            // set wrap bit at last rx buffer
            reg_val |= RX_BD_WRAP_BIT;
        }
        mmio_wr_32(in_mem_cfg->start_register_address, rx_bd_addr, reg_val);
        rx_bd_addr +=8;
    }

    if(in_mem_cfg->interupt_ena != 0)
    {
        // enable receive Receive and transmit interruptsn
        exc_register(in_mem_cfg->interupt_id, &eth_mac_handle_interrupt);
        intr_unmask(in_mem_cfg->interupt_id);

        reg_val  = INT_MASK_RXF_BIT | INT_MASK_TXB_BIT;
        mmio_wr_32(in_mem_cfg->start_register_address,INT_MASK_ADDR , reg_val);
    }
    
    configurations[interface_cnt] = in_mem_cfg;
    interface_cnt++;

    return (mmio_ptr_t) in_mem_cfg;
}

network_error_t eth_mac_send(network_frame_t *frame, unsigned long long int timeout, mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;

    network_error_t ret = eth_mac_send_nb(frame, config_ptr);
    // return if an error occured
    if(ret != NET_ERR_OK || ret == NET_ERR_DRVR_FRAME_QUEUED)
    {
        return ret;
    }

    mmio_ptr_t buf_address;
    unsigned long long int start_time = get_cpu_usecs();
    if(ret == NET_ERR_DRVR_FRAME_QUEUED)
    {
        // frame was enqued get last enqueued frame
        buf_address = get_last_tx_address(config);
        // Wait until frame 
        if (timeout == 0){
            // Wait until this frame is next
            #pragma loopbound min 1 max 1  //1us - 10ms
            while (buf_address != get_next_tx_address(config)){};
            // Wait until frame has been dequeued by upated routines
            #pragma loopbound min 1 max 1  //1us - 10ms
            while (buf_address == get_next_tx_address(config)){};

        }
        // wait until a frame is received or timeout is exceeded
        else
        {
            // Wait until this frame is next
            #pragma loopbound min 1 max 1   //1us
            while (buf_address != get_next_tx_address(config))
            {
                // Check for timeout
                if((get_cpu_usecs()-start_time >= timeout))
                {
                    return NET_ERR_DRVR_FRAME_QUEUED;
                }
            }
            // Wait until frame has been dequeued by upated routines
            #pragma loopbound min 1 max 1   //1us
            while (buf_address == get_next_tx_address(config))
            {
                // check for t timeout
                if((get_cpu_usecs()-start_time >= timeout))
                {
                    return NET_ERR_DRVR_FRAME_QUEUED;
                }
            }
        }
        // configuration is not required as this will happen in the update
        // routines
    }
    else
    {
        // transmission was already started, get address from eth mac register
        buf_address = (mmio_ptr_t) mmio_rd_32(config->start_register_address, TX_BD_ADDR_BASE+4);
    }
    
    //Wait until is is done
    uint8_t done = 0;
    #pragma loopbound min 1 max 1
    do{
        unsigned intr_en = EXC_STATUS & 0x1;
        if(intr_en)
        {
            CLI();
        }
        // check interrupt bit if frame has been transmitted
        done  = (mmio_rd_32(config->start_register_address,INT_SOURCE_ADDR) & INT_SOURCE_TXB_BIT);
        // when using interrupts the, interrupt bit will be cleared
        // therefore additionally it has to be checked if the 
        // address inside the buffer descriptor has been changed
        done |= (mmio_ptr_t) mmio_rd_32(config->start_register_address,TX_BD_ADDR_BASE+4) != buf_address;
        if(intr_en)
        {
            SEI();
        }
        // check for t timeout
        if(timeout > 0 && (get_cpu_usecs()-start_time >= timeout))
        {
            return NET_ERR_DRVR_BUFFER_BUSY;
        }
    }while (done==0); 

    // Transmission has been finished, free buffer again as it
    // is not required anymore 
    
    buffer_management_free((void*)buf_address, config->buffer_mgmt_id);

    return NET_ERR_OK; // frame has been sent
}
  
network_error_t eth_mac_send_nb(network_frame_t *frame, mmio_ptr_t config_ptr)
{

    network_error_t ret = NET_ERR_OK;
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;

    // check if there is space left to queue the frame
    if (tx_queue_full(config) == 1)
    {
        return NET_ERR_DRVR_TX_QUEUE_FULL; // No space to queue the frame
    }

    uint32_t frame_length = frame->header_length + frame->user_data_length;
    mmio_ptr_t buf_address = (mmio_ptr_t) buffer_management_malloc(frame_length, config->buffer_mgmt_id);
    // check if buffer request was successfull
    if(buf_address == NULL )
    {

#ifndef WCET_ANALYSIS 
        buffer_management_memory_dump(config->buffer_mgmt_id);
#endif 
        return NET_ERR_DRVR_MEM_FATAL; // Unable to acquire buffer, return error 
    }

    mmio_ptr_t address = (mmio_ptr_t) buf_address;
    // write header data into buffer
    mmio_wr_block(address, frame->header_data, frame->header_length);
    // adapt address and get next frame segment
    address += frame->header_length;
    
    // user data can be null sometimes with TCP frames
    if(frame->user_data != NULL)
    {
        // write user data into buffer
        mmio_wr_block(address, frame->user_data, frame->user_data_length);    
    }
    
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    // enques the frame 
    ret = enqueue_tx_frame(buf_address , frame_length, config);
    // frame will be sent later if it could be enqueue
    if(ret == NET_ERR_OK)
    {
        ret = NET_ERR_DRVR_FRAME_QUEUED;
    }
    // if the frame was not queued, buffer has to be released again
    else
    {
        if(intr_en)
        {
            SEI();
        }
        buffer_management_free((void*)buf_address, config->buffer_mgmt_id);
        return ret;   
    }

    uint8_t in_use = (mmio_rd_32(config->start_register_address,TX_BD_ADDR_BASE) & TX_BD_READY_BIT); 
    mmio_ptr_t bd_send_buffer = (mmio_ptr_t)  mmio_rd_32(config->start_register_address,TX_BD_ADDR_BASE+4);
 
    // free memory address from sent buffer
    if(in_use == 0 && bd_send_buffer != NULL)
    {
        buffer_management_free((void*)bd_send_buffer, config->buffer_mgmt_id);
        // configure NULL address to indicate no buffer is in the bd
        mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE+4, 0);
    }
    
    if(in_use != 0)
    {
        if(intr_en)
        {
            SEI();
        }
        return ret;
    }

    // MAC is free equeue a waiting frame
    // get frame from queue
    queue_entry_t entry;
    ret = dequeue_tx_frame(&entry, config);
    if( ret == NET_ERR_DRVR_TX_QUEUE_EMPTY)
    {
        // frame was dequeued between enque and disabling of interrupts
        // no error
        //return NET_ERR_OK;
        if(intr_en)
        {
            SEI();
        }
        return NET_ERR_DRVR_QUEUE_FATAL;
    }
    else
    {
        if(buf_address == entry.buffer_address)
        {
            ret = NET_ERR_OK;
        }
        // configure interrupt, buffer descriptor and start transmission        
        buf_address = entry.buffer_address;
        frame_length =  entry.frame_length;
    }
    
    
    // clear interrupt and set buffer descriptor        
    mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_TXB_BIT);
    mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE+4, (uint32_t) buf_address);
    uint32_t txbd = frame_length<<16; 
    txbd |= TX_BD_READY_BIT | TX_BD_IRQEN_BIT | TX_BD_WRAP_BIT | TX_BD_PAD_EN_BIT;
    // 
    mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE, txbd);
    
    if(intr_en)
    {
        SEI();
    }

    return ret;   // frame is beeing sent
}

network_error_t eth_mac_update_transmission(mmio_ptr_t config_ptr)
{
    network_error_t ret = NET_ERR_OK;
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;

    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    // check if the buffer is ready for a new transmission 
    uint8_t in_use = (mmio_rd_32(config->start_register_address, TX_BD_ADDR_BASE) & TX_BD_READY_BIT); 
    if(in_use !=0)
    {
        ret = NET_ERR_DRVR_BUFFER_BUSY;
    }
    else
    {
        // Free transmission Buffer as no longer used
        mmio_ptr_t buffer =(mmio_ptr_t) mmio_rd_32(config->start_register_address, TX_BD_ADDR_BASE+4);
        if(buffer != NULL)
        {
            buffer_management_free((void*)buffer, config->buffer_mgmt_id);
        }

        // clear interrupt flag
        mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_TXB_BIT);
        // get next frame
        queue_entry_t entry;
        if(tx_queue_empty(config) != 0 || dequeue_tx_frame(&entry, config) == NET_ERR_DRVR_TX_QUEUE_EMPTY)
        {
            // configure NULL address to indicate no buffer is in the bd
            mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE+4, 0);
            ret = NET_ERR_DRVR_TX_QUEUE_EMPTY;
        }
        else
        {
            // configure interrupt, buffer descriptor and start transmission        
            mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE+4, (uint32_t) entry.buffer_address);
            uint32_t txbd = entry.frame_length<<16; 
            txbd |= TX_BD_READY_BIT | TX_BD_IRQEN_BIT | TX_BD_WRAP_BIT | TX_BD_PAD_EN_BIT;
            mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE, txbd);
        }
    }

    if(intr_en)
    {
        SEI();
    }

    return ret;
}

network_error_t eth_mac_receive(uint8_t *rx_buffer, uint16_t buffer_size, unsigned long long int timeout, mmio_ptr_t config_ptr)
{
    
    // return value of -1 indicates no frame waiting
    network_error_t ret=NET_ERR_DRVR_NO_FRAME_RECEIVED;
    // Wait until a frame is received
    if (timeout == 0){
        #pragma loopbound min 1 max 1  //1us - 10ms
		while (ret==NET_ERR_DRVR_NO_FRAME_RECEIVED){
            ret = eth_mac_receive_nb(rx_buffer, buffer_size, config_ptr);        
        };
	}
    // wait until a frame is received or timeout is exceeded
    else
    {
		unsigned long long int start_time = get_cpu_usecs();
        #pragma loopbound min 1 max 1   //1us
		while (ret == NET_ERR_DRVR_NO_FRAME_RECEIVED){
            if((get_cpu_usecs()-start_time >= timeout))
            {
                return NET_ERR_DRVR_NO_FRAME_RECEIVED;
            }
            ret = eth_mac_receive_nb(rx_buffer, buffer_size, config_ptr);
        }
	}
    return ret;
}

network_error_t eth_mac_receive_nb(uint8_t *rx_buffer, uint16_t buffer_size, mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;
    uint32_t rx_bd = RX_BD_ADDR_BASE(mmio_rd_32(config->start_register_address, TX_BD_NUM_ADDR)); 
    
    mmio_ptr_t rx_buffer_eth_mac;
    uint32_t frame_length;
    /*
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }
    // Store whether a frame has been received by the eth mac 
    uint32_t frame_in_eth_mac = mmio_rd_32(config->start_register_address,INT_SOURCE_ADDR) & INT_SOURCE_RXB_BIT; 
   
    // Check if frame is waiting in the ethernet MAC 
    if(frame_in_eth_mac!=0)
    {
        // clear interrupt bit
        mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_RXB_BIT);
        uint32_t reg_val = RX_BD_EMPTY_BIT | RX_BD_IRQEN_BIT;

        #pragma loopbound min 1 max DEVICE_MAX_RX_BUFFERS
        for(uint16_t i=0; i < config->num_rx_buffers; i++)
        {
            uint32_t buffer_in_bd = (mmio_rd_32(config->start_register_address,rx_bd) & RX_BD_EMPTY_BIT) == 0 ;
            if(buffer_in_bd)
            {
                // enque the received frame, a frame is always queued to simplify the logic
                rx_buffer_eth_mac = (mmio_ptr_t) mmio_rd_32(config->start_register_address,rx_bd+4);
                frame_length = (mmio_rd_32(config->start_register_address,rx_bd) & 0xFFFF0000) >> 16;
                // return if frame was not enqueued
                if(enqueue_rx_frame(rx_buffer_eth_mac, frame_length, config) != NET_ERR_OK)
                {
                    if(intr_en)
                    {
                        SEI();
                    }
                    return NET_ERR_DRVR_RX_QUEUE_FULL;
                }
                
                // request new buffer for receiving a new frame
                mmio_ptr_t new_rx_buffer_addr = (mmio_ptr_t) buffer_management_malloc(ETHERNET_MAX_FRAME_SIZE, config->buffer_mgmt_id);
                
                // check if buffer request was successfull
                if(new_rx_buffer_addr == NULL)
                {
                    if(intr_en)
                    {
                        SEI();
                    }
                    return NET_ERR_DRVR_NO_FREE_BUFFER; // Unable to acquire buffer, return error 
                }
                
                // clear interrupt and configure buffer descriptor
                mmio_wr_32(config->start_register_address,rx_bd+4, (uint32_t) new_rx_buffer_addr);
                if(i == config->num_rx_buffers-1)
                {
                    // set wrap bit in last buffer descriptor
                    reg_val |= RX_BD_WRAP_BIT;
                }
                mmio_wr_32(config->start_register_address,rx_bd, reg_val);
            }
            rx_bd += 8;
        }
    } 

    if(intr_en)
    {
        SEI();
    }
    */
    // get next frame of queue
    if(rx_queue_empty(config) == 0 )
    {    
        
        frame_length = get_next_rx_frame_length(config);
        // check if a frame has been stored
        if(frame_length == 0)
        {
            return NET_ERR_DRVR_NO_FRAME_RECEIVED;
        }
        // check if frame fits in buffer
        else if(frame_length > buffer_size)
        {
            return NET_ERR_BUFF_UNDERSIZED;
        }
        // remove frame from queue
        queue_entry_t entry;
        dequeue_rx_frame(&entry, config);
        rx_buffer_eth_mac = entry.buffer_address;
    }
    else
    {
        // If there was no frame received by the mac and queued
        // or there was no frame in the queue no framme 
        // is there to be processed
        return NET_ERR_DRVR_NO_FRAME_RECEIVED;
    }
    
    // copy frame to buffer
    mmio_rd_block(rx_buffer_eth_mac, rx_buffer, frame_length);

    // free buffer of received frame again
    buffer_management_free((void*)rx_buffer_eth_mac, config->buffer_mgmt_id);

    return NET_ERR_OK;
}

network_error_t eth_mac_update_reception(mmio_ptr_t config_ptr)
{
    network_error_t ret = NET_ERR_DRVR_NO_FRAME_RECEIVED;

    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;

    // Store whether a frame has been received by the eth mac 
    uint32_t rx_bd = RX_BD_ADDR_BASE(mmio_rd_32(config->start_register_address,TX_BD_NUM_ADDR)); 
    
    uint32_t frame_in_eth_mac = mmio_rd_32(config->start_register_address, INT_SOURCE_ADDR) & INT_SOURCE_RXB_BIT; 
    //uint32_t frame_in_eth_mac = (mmio_rd_32(config->start_register_address,rx_bd) & RX_BD_EMPTY_BIT ) == 0; 
    
    unsigned intr_en = EXC_STATUS & 0x1;
    // Check if frame is waiting in the ethernet MAC 
    if(frame_in_eth_mac!=0)
    {
        // clear interrupt bit
        mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_RXB_BIT);
        uint32_t reg_val = RX_BD_EMPTY_BIT | RX_BD_IRQEN_BIT;
        
        mmio_ptr_t rx_buffer_eth_mac;
        uint32_t frame_length;

        #pragma loopbound min 1 max DEVICE_MAX_RX_BUFFERS
        for(uint16_t i=0; i < config->num_rx_buffers; i++)
        {
            uint32_t buffer_in_bd = (mmio_rd_32(config->start_register_address,rx_bd) & RX_BD_EMPTY_BIT) == 0 ;
            if(buffer_in_bd)
            {
                if(intr_en)
                {
                    CLI();
                }
                
                // enque the received frame, a frame is always queued to simplify the logic
                rx_buffer_eth_mac = (mmio_ptr_t) mmio_rd_32(config->start_register_address,rx_bd+4);
                frame_length = (mmio_rd_32(config->start_register_address,rx_bd) & 0xFFFF0000) >> 16;
                // return if frame was not enqueued
                if(enqueue_rx_frame(rx_buffer_eth_mac, frame_length, config) != NET_ERR_OK)
                {
                    if(intr_en)
                    {
                        SEI();
                    }
                    return NET_ERR_DRVR_RX_QUEUE_FULL;
                }
                
                // request new buffer for receiving a new frame
                mmio_ptr_t new_rx_buffer_addr = (mmio_ptr_t) buffer_management_malloc(ETHERNET_MAX_FRAME_SIZE, config->buffer_mgmt_id);
                
                // check if buffer request was successfull
                if(new_rx_buffer_addr == NULL)
                {
                    if(intr_en)
                    {
                        SEI();
                    }
                    return NET_ERR_DRVR_NO_FREE_BUFFER; // Unable to acquire buffer, return error 
                }
                
                // clear interrupt and configure buffer descriptor
                mmio_wr_32(config->start_register_address,rx_bd+4, (uint32_t) new_rx_buffer_addr);
                if(i == config->num_rx_buffers-1)
                {
                    // set wrap bit in last buffer descriptor
                    reg_val |= RX_BD_WRAP_BIT;
                }
                mmio_wr_32(config->start_register_address,rx_bd, reg_val);

                if(intr_en)
                {
                    SEI();
                }
            }

            rx_bd += 8;
        }
    }
    
    return ret;
}


uint16_t eth_mac_has_frame(mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;
    
    uint32_t recv = 0;
    uint32_t rx_bd = RX_BD_ADDR_BASE(mmio_rd_32(config->start_register_address,TX_BD_NUM_ADDR)); 

    if(rx_queue_empty(config) == 0 )
    {
        recv= get_next_rx_frame_length(config);
    }
    else if ((mmio_rd_32(config->start_register_address,INT_SOURCE_ADDR) & INT_SOURCE_RXB_BIT) !=0){
        uint32_t frame_length = mmio_rd_32(config->start_register_address,rx_bd) >> 16;
        recv = (uint16_t) frame_length;
    }
    
    return recv;
}

///////////////////////////////////////////////////////////////
//  Configuration functions
///////////////////////////////////////////////////////////////

/**
 * @brief Enables Transmission
 *  
 * This function enables transmission of ethernet frames.
 */
void eth_mac_enable_tx(mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address, MODER_ADDR);
    mmio_wr_32(config->start_register_address, MODER_ADDR, data | TXEN_BIT);
}

/**
 * @brief Disables Transmission
 *  
 * This function disable transmission of ethernet frames.
 */
void eth_mac_disable_tx(mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address,MODER_ADDR);
    mmio_wr_32(config->start_register_address,MODER_ADDR, data & ~TXEN_BIT);
}

/**
 * @brief Transmission is enabled
 *  
 * This function returns whether transmission is enabled or
 * not.
 * 
 * @return char enabled     : returns 1 if transmission is 
 *                            enabled, 0 otherwise.
 */
char eth_mac_tx_enabled(mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address,MODER_ADDR);
    if(data & TXEN_BIT)
    {
        return 1;
    }
    return 0;
}

/**
 * @brief Enables Reception
 *  
 * This function enables reception of ethernet frames.
 */
void eth_mac_enable_rx(mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address,MODER_ADDR);
    mmio_wr_32(config->start_register_address,MODER_ADDR, data | RXEN_BIT);
}

/**
 * @brief Disables Reception
 *  
 * This function disable reception of ethernet frames.
 */
void eth_mac_disable_rx(mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address,MODER_ADDR);
    mmio_wr_32(config->start_register_address,MODER_ADDR, data & ~RXEN_BIT);
}

/**
 * @brief Reception is enabled
 *  
 * This function returns whether reception is enabled or
 * not.
 * 
 * @return char enabled     : returns 1 if reception is 
 *                            enabled, 0 otherwise.
 */
char eth_mac_rx_enabled(mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address,MODER_ADDR);
    if(data & RXEN_BIT)
    {
        return 1;
    }
    return 0;
}

/**
 * @brief Sets the MAC Address 
 *  
 * This function is used to set a the Ethernet MACs MAC-Address.
 * 
 * @param uint8_t *mac_addr   : Buffer which holds the MAC
 *                                    Address.
 */
void eth_mac_set_mac_address(uint8_t *mac_addr, mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;
	uint32_t reg_val = (mac_addr[2] << 24) | (mac_addr[3] << 16) | (mac_addr[4] << 8) | (mac_addr[5]);
    mmio_wr_32(config->start_register_address,MAC_ADDR0, reg_val);
    reg_val = (mac_addr[0] << 8 ) | (mac_addr[1]);
	mmio_wr_32(config->start_register_address,MAC_ADDR1, reg_val);
}

/**
 * @brief Gets the MAC Address 
 *  
 * This function reads out the MAC-Address from the Ethernet MAC.
 * 
 * @param uint8_t *mac_addr   : Buffer where the MAC
 *                                    Address will be stored.
 */
void eth_mac_get_mac_address(uint8_t *mac_addr, mmio_ptr_t config_ptr)
{
    patmos_eth_mac_config_ptr_t config = (patmos_eth_mac_config_ptr_t) config_ptr;

    unsigned long long my_mac = mmio_rd_32(config->start_register_address, MAC_ADDR0); // Getting MSB 2 byte
    mac_addr[5] = my_mac & 0xFF;
    my_mac >>= 8;
    mac_addr[4] = my_mac & 0xFF;
    my_mac >>= 8;
    mac_addr[3] = my_mac & 0xFF;
    my_mac >>= 8;
    mac_addr[2] = my_mac & 0xFF;
    
    my_mac = mmio_rd_32(config->start_register_address,MAC_ADDR1); // Getting MSB 2 byte
    mac_addr[1] = my_mac & 0xFF;
    my_mac >>= 8;
    mac_addr[0] = my_mac & 0xFF;
}

/////////////////////
// Help functions
/////////////////////

/**
 * @brief Get frame length of next TX frame
 *  
 * This function returns the length of the frame which will be 
 * dequeued next from the transmission queue.
 * 
 * @return uint32_t frame_length    : Length of the TX Frame
 */
static uint32_t get_next_tx_frame_length(patmos_eth_mac_config_ptr_t config)
{   
    return config->tx_queue_frames[config->tx_queue_frames_read].frame_length;
}

/**
 * @brief Get buffer address of next TX frame
 *  
 * This function returns the buffer address of the frame which will 
 * be dequeued next from the transmission queue.
 * 
 * @return uint32_t buffer_address    : Address of the TX Frame
 */
static mmio_ptr_t get_next_tx_address(patmos_eth_mac_config_ptr_t config)
{   
    return config->tx_queue_frames[config->tx_queue_frames_read].buffer_address;
}

/**
 * @brief Get buffer address of last TX frame
 *  
 * This function returns the buffer address of the frame which was 
 * enqueued last at the transmission queue.
 * 
 * @return uint32_t buffer_address    : Address of the TX Frame
 */
static mmio_ptr_t get_last_tx_address(patmos_eth_mac_config_ptr_t config)
{   
    
    uint16_t last_entry = (config->tx_queue_frames_write-1>= config->queue_length) ? config->queue_length-1 : config->tx_queue_frames_write-1;

    return config->tx_queue_frames[last_entry].buffer_address;
}

/**
 * @brief Dequeues a TX Frame
 *  
 * This function provides the data of the next tx frame queued which is ready
 * to be sent.
 * 
 * @param  queue_entry_t *entry     : Pointer where the data will be stored
 * @return uint8_t status     : Returns 0 if successful else -1
 */
static network_error_t dequeue_tx_frame(queue_entry_t *entry, patmos_eth_mac_config_ptr_t config)
{
    network_error_t ret=NET_ERR_DRVR_TX_QUEUE_EMPTY;
    // check whether interrupts are enabled
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    // check whether an entry is left
    if(config->tx_queue_frames[config->tx_queue_frames_read].frame_length != 0)
    {
        // get entry
        entry->buffer_address = config->tx_queue_frames[config->tx_queue_frames_read].buffer_address;
        entry->frame_length = config->tx_queue_frames[config->tx_queue_frames_read].frame_length;
        // mark entry as used
        config->tx_queue_frames[config->tx_queue_frames_read].frame_length = 0;
        // calculate next read pointer
        config->tx_queue_frames_read = (config->tx_queue_frames_read+1>= config->queue_length) ? 0 : config->tx_queue_frames_read+1;
        ret = NET_ERR_OK;
    }

    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return ret;
}

/**
 * @brief Enqueues a TX Frame
 *  
 * This function adds a frame to the tx queue if there is still space 
 * left.
 * 
 * @param uint32_t address      : Address where frame is stored
 * @param uint32_t frame_length : Length of the Frame
 * @return char status              : 0 if enqueueing was successful, 
                                      else -1.
 */
static network_error_t enqueue_tx_frame(mmio_ptr_t address, uint16_t frame_length, patmos_eth_mac_config_ptr_t config)
{
    // check whether interrupts are enabled
    network_error_t ret=NET_ERR_DRVR_TX_QUEUE_FULL;

    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    if(config->tx_queue_frames[config->tx_queue_frames_write].frame_length == 0)
    {
        // memory barrier 
        asm volatile ("" : : : "memory");
        // mark used buffer entry as available again
        config->tx_queue_frames[config->tx_queue_frames_write].buffer_address = address;
        config->tx_queue_frames[config->tx_queue_frames_write].frame_length = frame_length;
        // calculate next write pointer
        config->tx_queue_frames_write = (config->tx_queue_frames_write+1>= config->queue_length ) ? 0 : config->tx_queue_frames_write+1;
        ret = NET_ERR_OK;
    }
    
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return ret;
}

/**
 * @brief Check if tx queue is full
 *  
 * This function checks whether there is still space left in the tx queue.
 * 
 * @return char status          : Returns 1 if full, 0 otherwise
 */
static char tx_queue_full(patmos_eth_mac_config_ptr_t config)
{
    char ret=1;
    // check whether interrupts are enabled
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    // check whether an entry is left
    if(config->tx_queue_frames[config->tx_queue_frames_read].frame_length == 0)
    {
        ret =0;
    }
    
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return 0;
}

/**
 * @brief Check if the tx queue is empty
 *  
 * This function checks whether the tx queue is empty.
 * 
 * @return char status          : Returns 1 if empty, 0 otherwise
 */
static char tx_queue_empty(patmos_eth_mac_config_ptr_t config)
{
    // check whether interrupts are enabled,
    char ret =1;
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    // check whether an entry is left
    if(config->tx_queue_frames[config->tx_queue_frames_read].frame_length != 0)
    {
        ret = 0;
    }
    
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return ret;
}

/**
 * @brief Get frame length of next RX frame
 *  
 * This function returns the length of the frame which will be 
 * dequeued next from the reception queue.
 * 
 * @return uint32_t frame_length    : Length of the RX Frame
 */
static uint32_t get_next_rx_frame_length(patmos_eth_mac_config_ptr_t config)
{   
    return config->rx_queue_frames[config->rx_queue_frames_read].frame_length;
}

/**
 * @brief Get buffer address of next frame
 *  
 * This function returns the buffer address of the frame which will 
 * be dequeued next from the reception queue.
 * 
 * @return uint32_t buffer_address    : Address of the RX Frame
 */
static mmio_ptr_t get_next_rx_address(patmos_eth_mac_config_ptr_t config)
{   
    return config->rx_queue_frames[config->rx_queue_frames_read].buffer_address;
}

/**
 * @brief Dequeues a RX Frame
 *  
 * This function provides the data of the next rx frame queued which is ready
 * to be sent.
 * 
 * @param  queue_entry_t *entry     : Pointer where the data will be stored
 * @return uint8_t status     : Returns 0 if successful else -1
 */
static network_error_t dequeue_rx_frame(queue_entry_t *entry, patmos_eth_mac_config_ptr_t config)
{
    network_error_t ret=NET_ERR_DRVR_RX_QUEUE_EMPTY;
    // check whether interrupts are enabled
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    // check whether an entry is left
    if(config->rx_queue_frames[config->rx_queue_frames_read].frame_length != 0)
    {
        // get entry
        entry->buffer_address = config->rx_queue_frames[config->rx_queue_frames_read].buffer_address;
        entry->frame_length = config->rx_queue_frames[config->rx_queue_frames_read].frame_length;
        // mark entry as used
        config->rx_queue_frames[config->rx_queue_frames_read].frame_length = 0;
        // calculate next read pointer
        config->rx_queue_frames_read = (config->rx_queue_frames_read+1>= config->queue_length) ? 0 : config->rx_queue_frames_read+1;
        ret = NET_ERR_OK;
    }

    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return ret;
}

/**
 * @brief Enqueues a RX Frame
 *  
 * This function adds a frame to the rx queue if there is still space 
 * left.
 * 
 * @param uint32_t address      : Address where frame is stored
 * @param uint32_t frame_length : Length of the Frame
 * @return char status              : 0 if enqueueing was successful, 
                                      else -1.
 */
static network_error_t enqueue_rx_frame(mmio_ptr_t address, uint16_t frame_length, patmos_eth_mac_config_ptr_t config)
{
    // check whether interrupts are enabled
    network_error_t ret = NET_ERR_DRVR_RX_QUEUE_FULL;
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }
    
    if(config->rx_queue_frames[config->rx_queue_frames_write].frame_length == 0)
    {
        // memory barrier 
        asm volatile ("" : : : "memory");
        // mark used buffer entry as available again
        config->rx_queue_frames[config->rx_queue_frames_write].buffer_address = address;
        config->rx_queue_frames[config->rx_queue_frames_write].frame_length = frame_length;
        // calculate next write pointer
        config->rx_queue_frames_write = (config->rx_queue_frames_write+1>= config->queue_length) ? 0 : config->rx_queue_frames_write+1;
        ret = NET_ERR_OK;        
    }
    
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return ret;
}

/**
 * @brief Check if rx queue is full
 *  
 * This function checks whether there is still space left in the rx queue.
 * 
 * @return char status          : Returns 1 if full, 0 otherwise
 */
static char rx_queue_full(patmos_eth_mac_config_ptr_t config)
{
    char ret=1;
    // check whether interrupts are enabled
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    // check whether an entry is left
    if(config->rx_queue_frames[config->rx_queue_frames_read].frame_length == 0)
    {
        ret =0;
    }
    
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return 0;
}

/**
 * @brief Check if the rx queue is empty
 *  
 * This function checks whether the rx queue is empty.
 * 
 * @return char status          : Returns 1 if empty, 0 otherwise
 */
static char rx_queue_empty(patmos_eth_mac_config_ptr_t config)
{
    // check whether interrupts are enabled,
    char ret =1;
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    // check whether an entry is left
    if(config->rx_queue_frames[config->rx_queue_frames_read].frame_length != 0)
    {
        ret = 0;
    }
    
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return ret;
}

#endif