/*
 * Driver for the Ethernet MAC
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 *          Christoph Lehr
 */

#ifndef ETH_MAC_DRIVER_H_
#define ETH_MAC_DRIVER_H_

#ifdef ARCH_PATMOS

#include "../../../inc/net_types.h"
#include "../../../inc/buffer_management.h"

#define ETHERNET_MAX_FRAME_SIZE 1518
#define ETHERNET_MIN_FRAME_SIZE 64
#define ETH_MAC_QUEUE_COUNT     8
#define ETH_MAC_RX_QUEUE_SIZE   30
#define ETH_MAC_TX_QUEUE_SIZE   30

#define MODER_ADDR         0x00  // Mode
#define INT_SOURCE_ADDR    0x04  // Interrupt source
#define INT_MASK_ADDR      0x08  // Interrupt mask
#define IPGT_ADDR          0x0C  // Back to back inter packet gap
#define IPGR1_ADDR         0x10  // Non back to back inter packet gap
#define IPGR2_ADDR         0x14  // Non back to back inter packet gap
#define PACKETLEN_ADDR     0x18  // Packet length (minimum and maximum)
#define COLLCONF_ADDR      0x1C  // Collision and retry configuration
#define TX_BD_NUM_ADDR     0x20  // Transmit buffer descriptor number
#define CTRLMODER_ADDR     0x24  // Control module mode
#define MIIMODER_ADDR      0x28  // MII mode register
#define MIICOMMAND_ADDR    0x2C  // MII command
#define MIIADDRESS_ADDR    0x30  // MII address register containts the phy address 
                                 // and the register with the phy address
#define MIITX_DATA_ADDR    0x34  // 16-bit transmit data to selected phy register
#define MIIRX_DATA_ADDR    0x38  // 16-bit received data from selected phy register
#define MIISTATUS          0x3C  // MII Status
#define MAC_ADDR0          0x40  // MAC Address Bytes 2 to 5
#define MAC_ADDR1          0x44  // MAC Address Bytes 0 to 1
#define ETH_HASHO_ADR      0x48  // HASH0 Register
#define ETH_HASH1_ADR      0x4C  // HASH1 Register
#define TXCTRL             0x50  // TX Control Register

#define RECSMALL_BIT 0x10000
#define PAD_BIT      0x08000
#define HUGEN_BIT    0x04000
#define CRCEN_BIT    0x02000
#define DLYCRCEN_BIT 0x01000
#define FULLD_BIT    0x00400
#define EXDFREN_BIT  0x00200
#define NOBCKOF_BIT  0x00100
#define LOOPBCK_BIT  0x00080
#define IFG_BIT      0x00040
#define PRO_BIT      0x00020
#define IAM_BIT      0x00010
#define BRO_BIT      0x00008
#define NOPRE_BIT    0x00004
#define TXEN_BIT     0x00002
#define RXEN_BIT     0x00001

#define WCTRLDATA_BIT   0x4
#define RSTAT_BIT       0x2
#define SCANSTAT_BIT    0x1
#define NVALID_BIT      0x4
#define BUSY_BIT        0x2
#define LINKFAIL_BIT    0x1

#define TX_BD_ADDR_BASE             0x400
#define TX_BD_ADDR_END(TX_BD_NUM)   TX_BD_ADDR_BASE + TX_BD_NUM * 8

#define RX_BD_ADDR_BASE(TX_BD_NUM)  TX_BD_ADDR_BASE + TX_BD_NUM * 8
#define RX_BD_ADDR_END              0x7FF

#define INT_SOURCE_RXC_BIT          0x0040
#define INT_SOURCE_TXC_BIT          0x0020
#define INT_SOURCE_BUSY_BIT         0x0010
#define INT_SOURCE_RXE_BIT          0x0008
#define INT_SOURCE_RXB_BIT          0x0004
#define INT_SOURCE_TXE_BIT          0x0002
#define INT_SOURCE_TXB_BIT          0x0001

#define INT_MASK_RXC_BIT          0x0040
#define INT_MASK_TXC_BIT          0x0020
#define INT_MASK_BUSY_BIT         0x0010
#define INT_MASK_RXE_BIT          0x0008
#define INT_MASK_RXF_BIT          0x0004
#define INT_MASK_TXE_BIT          0x0002
#define INT_MASK_TXB_BIT          0x0001

#define RX_BD_EMPTY_BIT             0x8000
#define RX_BD_IRQEN_BIT             0x4000
#define RX_BD_WRAP_BIT              0x2000
#define RX_BD_CF_BIT                0x0100
#define RX_BD_MISS_BIT              0x0080
#define RX_BD_OR_BIT                0x0040
#define RX_BD_IS_BIT                0x0020
#define RX_BD_DN_BIT                0x0010
#define RX_BD_TL_BIT                0x0008
#define RX_BD_SF_BIT                0x0004
#define RX_BD_CRCERR_BIT            0x0002
#define RX_BD_LC_BIT                0x0001     

#define TX_BD_READY_BIT             0x8000
#define TX_BD_IRQEN_BIT             0x4000
#define TX_BD_WRAP_BIT              0x2000
#define TX_BD_PAD_EN_BIT            0x1000
#define TX_BD_CRC_EN_BIT            0x0800
#define TX_BD_UR_BIT                0x0100
#define TX_BD_RTRY_BIT              0x0010
#define TX_BD_LC_BIT                0x0008
#define TX_BD_DF_BIT                0x0004
#define TX_BD_CS_BIT                0x0001              

// Frame Queue entry 
typedef struct
{
    mmio_ptr_t buffer_address;
    uint16_t padding;
    uint16_t frame_length;
    time_stamp_t ts;
} queue_entry_t;

typedef queue_entry_t volatile _IODEV * queue_entry_ptr_t;

typedef struct 
{
    uint16_t interupt_ena:1;
    uint16_t interupt_id:15;
    uint16_t num_rx_buffers;
    uint16_t buffer_mgmt_id;
    uint16_t queue_length;
    uint16_t tx_queue_frames_read;
    uint16_t tx_queue_frames_write;
    uint16_t rx_queue_frames_read;
    uint16_t rx_queue_frames_write;
    mac_address_t mac_address;
    mmio_ptr_t start_register_address;
    mmio_ptr_t start_memory_address;
    queue_entry_ptr_t tx_queue_frames;
    queue_entry_ptr_t rx_queue_frames;
} patmos_eth_mac_config_t;

typedef patmos_eth_mac_config_t _IODEV * patmos_eth_mac_config_ptr_t;

///////////////////////////////////////////////////////////////
//High level functions 
///////////////////////////////////////////////////////////////


/**
 * @brief Initialize Ethernet MAC 
 *  
 * This function initializes the MAC Address and the Receive Buffer,
 * Configures the Ethernet MAC
 * 
 * 
 */
mmio_ptr_t eth_mac_initialize(device_config_t *config);

/**
 * @brief Sends a Ethernet Frame 
 *  
 * This function requests a buffer from the Ethernet MAC, transfers
 * data to the Ethernet MAC. If there is no buffer available or no 
 * space left inside the transmission queue, the function returns 
 * -1. If there is no frame waiting for transmission the frame will 
 * be transmitted immidiately and the function blocks until finished. 
 * If there is frame already waiting the frame will be queued and the 
 * function returns -2.
 * 
 */
network_error_t eth_mac_send(network_frame_t *seg, unsigned long long int timeout, mmio_ptr_t config_ptr);

/**
 * @brief Sends a Ethernet Frame without blocking
 *  
 * This function requests a buffer from the Ethernet MAC, transfers
 * data to the Ethernet MAC. If there is no buffer available or no 
 * space left inside the transmission queue, the function returns 
 * -1. If there is no frame waiting for transmission the frame will 
 * be transmitted immidiately and the function will return without  
 * blocking. If there is frame already waiting the frame will be  
 * queued and the function returns -2.
 * 
 */
network_error_t eth_mac_send_nb(network_frame_t *seg, mmio_ptr_t config_ptr);
 

/**
 * @brief Update Transmission 
 *  
 * This function is used to send enqueued frames and shall be 
 * called periodically. If a frame was transmitted and the buffer
 * is no longer busy, a buffer entry will be dequeued from the TX
 * Queue and sending will be initiated.   
 *
 */
network_error_t eth_mac_update_transmission(mmio_ptr_t config_ptr);


/**
 * @brief Receive a Ethernet Frame 
 *  
 * This function waits until receiving a ethernet frame. A timeout
 * can be provided how long this function shall wait, use 0 to wait
 * indefinately. When the timeout is exceeded the function returns -1
 * When a frame is received, it is checked if the frame will fit
 * in the provided buffer, if the buffer is not large enough -2 will
 * be return from the function. If a frame was received successfuly
 * the function returns 0. 
 * 
 */
network_error_t eth_mac_receive(uint8_t *rx_buffer, uint16_t buffer_size, unsigned long long int timeout, mmio_ptr_t config_ptr);

/**
 * @brief Receive a Ethernet Frame 
 *  
 * This function checks if a frame has been received by the
 * Ethernet MAC. If a frame is waiting the data will be copied
 * to the provided buffer and the function returns 0. If the
 * provided Buffer does not have sufficient size the function 
 * returns -2. If there is no frame waiting for reception the
 * function returns -1.
 * 
 */
network_error_t eth_mac_receive_nb(uint8_t *rx_buffer, uint16_t buffer_size, mmio_ptr_t config_ptr);

/**
 * @brief Update Reception 
 *  
 * This function is used to enqueued received frames from the
 * ethernet frame and shall be called periodically to prevent
 * dropping of frames. If a frame was received the function
 * adds it to the RX queue. A new buffer will be requested and
 * set as the receive buffer.
 *
 */
network_error_t eth_mac_update_reception(mmio_ptr_t config_ptr);

/**
 * @brief Check for Ethernet Frame 
 *  
 * This function checks whether the Ethernet has received a frame which
 * is ready to be copied to the internal Memory. If a frame was received
 * by the Ethernet MAC, the function returns the length of the frame, 
 * otherwise 0.
 * 
 * @return unsigned short recv: 1 if a frame has be received, 0 if not
 */
uint16_t eth_mac_has_frame(mmio_ptr_t config_ptr);

///////////////////////////////////////////////////////////////
//  Configuration functions
///////////////////////////////////////////////////////////////

/**
 * @brief Enables Transmission
 *  
 * This function enables transmission of ethernet frames.
 */
void eth_mac_enable_tx(mmio_ptr_t config_ptr);

/**
 * @brief Disables Transmission
 *  
 * This function disable transmission of ethernet frames.
 */
void eth_mac_disable_tx(mmio_ptr_t config_ptr);

/**
 * @brief Transmission is enabled
 *  
 * This function returns whether transmission is enabled or
 * not.
 * 
 * @return char enabled     : returns 1 if transmission is 
 *                            enabled, 0 otherwise.
 */
char eth_mac_tx_enabled(mmio_ptr_t config_ptr);

/**
 * @brief Enables Reception
 *  
 * This function enables reception of ethernet frames.
 */
void eth_mac_enable_rx(mmio_ptr_t config_ptr);

/**
 * @brief Disables Reception
 *  
 * This function disable reception of ethernet frames.
 */
void eth_mac_disable_rx(mmio_ptr_t config_ptr);

/**
 * @brief Reception is enabled
 *  
 * This function returns whether reception is enabled or
 * not.
 * 
 * @return char enabled     : returns 1 if reception is 
 *                            enabled, 0 otherwise.
 */
char eth_mac_rx_enabled(mmio_ptr_t config_ptr);



/**
 * @brief Sets the MAC Address 
 *  
 * This function is used to set a the Ethernet MACs MAC-Address.
 * 
 * @param unsigned char *mac_addr   : Buffer which holds the MAC
 *                                    Address.
 */
void eth_mac_set_mac_address(unsigned char *mac_addr, mmio_ptr_t config_ptr);

/**
 * @brief Gets the MAC Address 
 *  
 * This function reads out the MAC-Address from the Ethernet MAC.
 * 
 * @param unsigned char *mac_addr   : Buffer where the MAC
 *                                    Address will be stored.
 */
void eth_mac_get_mac_address(unsigned char *mac_addr, mmio_ptr_t config_ptr);

#endif 
#endif