// Copyright 2013 Stefan Hepp
// 
// This file is part of the newlib C library for the Patmos processor.
// 
// This file is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This code is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// (COPYING3.LIB). If not, see <http://www.gnu.org/licenses/>.
//
// This file contains Patmos RTC related functions. 


#ifndef PATMOS_ETH_MAC_TIMER
#define PATMOS_ETH_MAC_TIMER

#ifdef ARCH_PATMOS

#include <machine/patmos.h>

/**
 * Address offsets.
 */
#define __PATMOS_TIMER_HICLK_OFFSET      (0x00)
#define __PATMOS_TIMER_LOCLK_OFFSET      (0x04)
#define __PATMOS_TIMER_HIUSEC_OFFSET     (0x08)
#define __PATMOS_TIMER_LOUSEC_OFFSET     (0x0c)

/**
 * Get the current cycle counter value.
 */
static inline uint64_t patmos_get_cpu_cycles(void* timer_base) {
    uint32_t clo, chi;

    // Prevent the compiler from moving the read over other instructions 
    // or into a call delay slot behind the call miss stall
    asm volatile ("" : : : "memory");

    _iodev_ptr_t hi_clock = (_iodev_ptr_t)(timer_base + __PATMOS_TIMER_HICLK_OFFSET);
    _iodev_ptr_t lo_clock = (_iodev_ptr_t)(timer_base + __PATMOS_TIMER_LOCLK_OFFSET);

    // Order is important here
    clo = *lo_clock;
    chi = *hi_clock;

    asm volatile ("" : : : "memory");

    return (((uint64_t) chi) << 32) | clo;
}

/**
 * Get the current RTC microsecond value.
 */
static inline uint64_t patmos_get_cpu_usecs(void* timer_base) {
    uint32_t ulo, uhi;

    // Prevent the compiler from moving the read over other instructions 
    // or into a call delay slot behind the call miss stall
    asm volatile ("" : : : "memory");

    _iodev_ptr_t hi_usec = (_iodev_ptr_t)(timer_base + __PATMOS_TIMER_HIUSEC_OFFSET);
    _iodev_ptr_t lo_usec = (_iodev_ptr_t)(timer_base + __PATMOS_TIMER_LOUSEC_OFFSET);

    // Order is important here
    ulo = *lo_usec;
    uhi = *hi_usec;

    asm volatile ("" : : : "memory");

    return (((uint64_t) uhi) << 32) | ulo;
}

/**
 * Set the timeout for the clock timer. The RTC will trigger an interrupt once
 * the cycle counter reaches the given value.
 */
static inline void patmos_arm_clock_timer(void* timer_base, uint64_t timestamp) {
    // Prevent the compiler from moving the write over other instructions 
    // or into a call delay slot behind the call miss stall
    asm volatile ("" : : : "memory");

    _iodev_ptr_t hi_clock = (_iodev_ptr_t)(timer_base + __PATMOS_TIMER_HICLK_OFFSET);
    _iodev_ptr_t lo_clock = (_iodev_ptr_t)(timer_base + __PATMOS_TIMER_LOCLK_OFFSET);

    // Order is important here
    *lo_clock = (uint32_t)timestamp;
    *hi_clock = (uint32_t)(timestamp>>32);

    asm volatile ("" : : : "memory");
}

/**
 * Set the timeout for the clock timer. The RTC will trigger an interrupt once
 * the cycle counter reaches the given value.
 */
static inline void patmos_arm_usec_timer(void* timer_base, uint64_t timestamp) {
    // Prevent the compiler from moving the write over other instructions 
    // or into a call delay slot behind the call miss stall
    asm volatile ("" : : : "memory");

    _iodev_ptr_t hi_usec = (_iodev_ptr_t)(timer_base + __PATMOS_TIMER_HIUSEC_OFFSET);
    _iodev_ptr_t lo_usec = (_iodev_ptr_t)(timer_base + __PATMOS_TIMER_LOUSEC_OFFSET);

    // Order is important here
    *lo_usec = (uint32_t)timestamp;
    *hi_usec = (uint32_t)(timestamp>>32);

    asm volatile ("" : : : "memory");
}

#endif
#endif /* PATMOS_ETH_MAC_TIMER */
