/*
 * Driver for the Ethernet MAC
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 *          Christoph Lehr
 */

//#ifdef ARCH_PATMOS

#include <stdio.h>
#include <stdlib.h>

#include <machine/exceptions.h>
#include <machine/rtc.h>

#include "../mmio_utils.h"
#include "../../../inc/device.h"
#include "../../../inc/generic_types.h"
#include "../../../inc/net_types.h"
#include "../../../inc/buffer_management.h"
#include "../../../inc/utils.h"
#include "patmos_eth_mac.h"
#include "patmos_tsn_mac.h"
#include "patmos_timer.h"

#ifndef PATMOS_ETH_MAC_CNT
    #define PATMOS_ETH_MAC_CNT 2
#endif

tsn_runtime_config_ptr_t configurations[PATMOS_ETH_MAC_CNT];
static uint8_t interface_cnt = 0;

#define LEDS (*((volatile _IODEV unsigned *)  PATMOS_IO_LED))
///////////////////////////////////////////////////////////////
// Helper Function Prototypes 
///////////////////////////////////////////////////////////////

static network_error_t enqueue_frame(mmio_ptr_t address, uint16_t frame_length, time_stamp_t *ts, queue_ptr_t config);

static network_error_t dequeue_frame(queue_entry_t *entry, queue_ptr_t config);

static char queue_full(queue_ptr_t config);

static char queue_empty(queue_ptr_t config);

static uint32_t get_next_frame_length(queue_ptr_t config);

static mmio_ptr_t get_next_address(queue_ptr_t config);

static mmio_ptr_t get_last_address(queue_ptr_t config);


///////////////////////////////////////////////////////////////
// Network Clock functionality
///////////////////////////////////////////////////////////////
#define SEC_TO_NS   1000000000
#define SEC_TO_USEC    1000000
#define SEC_TO_MS         1000

#define MS_TO_SEC        0.001
#define MS_TO_USEC    1000
#define MS_TO_NS   1000000

#define USEC_TO_NS 1000

#define NS_TO_USEC 0.001
#define NS_TO_MS   0.000001
#define NS_TO_SEC  0.000000001

uint64_t get_netw_clk_usecs(mmio_ptr_t config_ptr);

///////////////////////////////////////////////////////////////
//High level functions 
///////////////////////////////////////////////////////////////

static void do_eth_interrupt_stuff(void)
{
    tsn_runtime_config_ptr_t config =configurations[0];

    uint32_t interrupt_id= exc_get_source();
    #pragma loopbound min 1 max PATMOS_ETH_MAC_CNT
    for(uint8_t i=0;i < interface_cnt; i++)
    {  
        if(configurations[i] != NULL && configurations[i]->interupt_id == interrupt_id)
        {
            config = configurations[i];
            break;
        }
    }
    
    if(config != NULL)
    {   

        //uint32_t interrupt_src = mmio_rd_32(config->start_register_address,INT_SOURCE_ADDR);
        // Check if interrupt was caused by transmission of a frame
        // and execute corresponding update routing
        //if(interrupt_src & INT_SOURCE_TXB_BIT)
        {
            tsn_mac_update_transmission((mmio_ptr_t) config);
        //    mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_TXB_BIT);
        }

        // Check if interrupt was caused by reception  of a frame
        // and execute corresponding update routing
        //if(interrupt_src & INT_SOURCE_RXB_BIT)
        {
            tsn_mac_update_reception((mmio_ptr_t) config);
        //    mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_RXB_BIT);
        }
        mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_RXB_BIT|INT_SOURCE_TXB_BIT);
        
    }
}

static void handle_eth_interrupt(void) __attribute__((naked));
static void handle_eth_interrupt(void)
{
    exc_prologue();

    // functions do ISR protection on their own, interrupts can be enabled
    //SEI();

    do_eth_interrupt_stuff();
    
    //CLI();

    exc_epilogue();
} 



/**
 * @brief Handle timer tick
 *  
 * Interrupt callback function to handle timer ticks
 */
static void handle_tsn_schedule()
{
    tsn_runtime_config_ptr_t config =configurations[0];
    
    uint32_t interrupt_id= exc_get_source();
    #pragma loopbound min 1 max PATMOS_ETH_MAC_CNT
    for(uint8_t i=0;i < interface_cnt; i++)
    {  
        if(configurations[i] != NULL && configurations[i]->timer_cfg.interupt_id == interrupt_id)
        {
            config = configurations[i];
            break;
        }
    }
    
    // If update time is less then the current time, the gcl shall be updated immidiately 
    time_stamp_t netw_clk_time;    
    (void)tsn_mac_get_netw_clk_timestamp( (mmio_ptr_t) config, &netw_clk_time);

    if((config->admin_schedule->update_time.seconds > 0 || config->admin_schedule->update_time.nanoseconds > 0) && 
        (config->admin_schedule->update_time.seconds <= netw_clk_time.seconds ||
        (config->admin_schedule->update_time.seconds == netw_clk_time.seconds && config->admin_schedule->update_time.nanoseconds <= netw_clk_time.nanoseconds)))
    {
        tsn_schedule_config_ptr_t tmp = config->oper_schedule;
        config->oper_schedule = config->admin_schedule;
        config->admin_schedule = tmp;
        
        // start new cycle now
        config->cycle_start = netw_clk_time;

        config->list_pointer = 0;
        config->admin_schedule->update_time.seconds = 0;
        config->admin_schedule->update_time.nanoseconds = 0;
    }

    time_interval_t until_cycle_start;


    // end of cycle reached
    if(config->list_pointer == config->oper_schedule->gcl_length)
    {
        //config->list_pointer=0;
        config->current_gate_state = 0; // last slot ended, close all gates    
        
        //current_time = patmos_get_cpu_usecs(config->timer_cfg.timer_base_address);
        (void)tsn_mac_get_netw_clk_timestamp((mmio_ptr_t)config, &netw_clk_time);

        time_stamp_t cycle_start = config->cycle_start;
        time_interval_t delta;
        delta.seconds = netw_clk_time.seconds;
        delta.nanoseconds = netw_clk_time.nanoseconds;

        delta.seconds -= cycle_start.seconds;
        delta.nanoseconds -= cycle_start.nanoseconds;
        if(delta.nanoseconds <= -SEC_TO_NS || (delta.nanoseconds< 0 && delta.seconds !=0))
        {
            delta.seconds--;
            delta.nanoseconds+=SEC_TO_NS;
        }
        
        if(delta.seconds>0)
        {
            uint32_t cycles_per_second   = SEC_TO_NS / config->oper_schedule->cycle_time;
            if(cycles_per_second *config->oper_schedule->cycle_time < SEC_TO_NS)
            {
                cycles_per_second++;
            }
            uint32_t overhang_per_second_ns = cycles_per_second * config->oper_schedule->cycle_time -  SEC_TO_NS ;
            uint32_t overhang_per_second_ms = overhang_per_second_ns / MS_TO_NS;
            
            uint32_t overhang_delta_ms=0;
            // Iterate that overhang does not exceed nanosecond portion, approximate by halving seconds 
            // loopbound 22, assuming clocks are not more than 48 days apart 
            #pragma loopbound min 0 max 22
            while (delta.seconds*overhang_per_second_ns > (uint64_t) delta.nanoseconds)
            {    
                overhang_delta_ms = (delta.seconds>>1) * overhang_per_second_ms;

                delta.seconds -= (delta.seconds >>1)+overhang_delta_ms/SEC_TO_MS;
                delta.nanoseconds -= (overhang_delta_ms % SEC_TO_MS)*MS_TO_NS;

                if(delta.nanoseconds <= -SEC_TO_NS || (delta.nanoseconds< 0 && delta.seconds !=0))
                {
                    delta.seconds--;
                    delta.nanoseconds+=SEC_TO_NS;
                }

            }
            // overhang does no long exceed nanosecond portion, therefore
            overhang_delta_ms = delta.seconds * overhang_per_second_ms;
            delta.seconds -= delta.seconds + overhang_delta_ms/SEC_TO_MS;
            delta.nanoseconds -= (overhang_delta_ms % SEC_TO_MS)*MS_TO_NS;
        }
        // delta is now less then a second
        uint32_t cycles = delta.nanoseconds/config->oper_schedule->cycle_time;
        cycles++;
        delta.nanoseconds = config->oper_schedule->cycle_time*cycles-delta.nanoseconds;

        cycle_start.seconds = netw_clk_time.seconds;
        cycle_start.nanoseconds = netw_clk_time.nanoseconds + delta.nanoseconds ;
        if(cycle_start.nanoseconds >= SEC_TO_NS)
        {
            cycle_start.nanoseconds -= SEC_TO_NS;
            cycle_start.seconds++;
        }

        uint64_t current_time_usecs = patmos_get_cpu_usecs(config->timer_cfg.timer_base_address);        
        if(delta.nanoseconds >= TSN_MAC_CLOCK_UPDATE_INTERVAL_US*USEC_TO_NS )
        {
            config->time_interval = TSN_MAC_CLOCK_UPDATE_INTERVAL_US;
            config->slot_end_time_stamp = current_time_usecs+ TSN_MAC_CLOCK_UPDATE_INTERVAL_US;
        }
        else
        {
            config->time_interval = delta.nanoseconds/USEC_TO_NS;
            
            config->slot_end_time_stamp = current_time_usecs+ config->time_interval;
            config->list_pointer++;
            config->cycle_start = cycle_start;
        }        
    }
    else if(config->list_pointer > config->oper_schedule->gcl_length)
    {
        config->list_pointer = 0;
    }

    // load next slot
    if(config->list_pointer < config->oper_schedule->gcl_length)
    {   
        //config->current_gate_state =  0xFF;
        config->current_gate_state = config->oper_schedule->gate_control_list[config->list_pointer].gate_state;
        config->time_interval = config->oper_schedule->gate_control_list[config->list_pointer].slot_duration;
        /*
        // check if slot ends after update time
        if((config->admin_schedule->update_time.seconds > 0 || config->admin_schedule->update_time.nanoseconds > 0) && 
            (config->admin_schedule->update_time.seconds <= netw_clk_time.seconds ||
            (config->admin_schedule->update_time.seconds == netw_clk_time.seconds && config->admin_schedule->update_time.nanoseconds <= netw_clk_time.nanoseconds)))
        {
            uint64_t interval = config->admin_schedule->update_time.seconds - netw_clk_time.seconds;
            interval = (interval << 10) - (interval << 4) - (interval << 3); // interval * 1024 - interval *16 - intervale *8 = interval(s) *1000 = interval(ms)
            interval = (interval << 10) - (interval << 4) - (interval << 3); // interval * 1024 - interval *16 - intervale *8 = interval(ms) *1000= interval(us)
            interval += (config->admin_schedule->update_time.nanoseconds - netw_clk_time.nanoseconds)/1000;
            config->time_interval =interval;
        }
        */
        config->slot_end_time_stamp += config->time_interval;

        // increment list_pointer
        config->list_pointer++;

        tsn_mac_update_transmission((mmio_ptr_t) config);
    }
    
#ifndef WCET_ANALYSIS 
    uint64_t check_time = patmos_get_cpu_usecs(config->timer_cfg.timer_base_address);
    if(config->slot_end_time_stamp <= check_time)
    {
        printf("slot_end_time_stamp:  %lld check_time:  %lld\n", config->slot_end_time_stamp, check_time);
        printf("config->cycle_start:  %llu.%09u \n", config->cycle_start.seconds, config->cycle_start.nanoseconds);
        printf("netw_clk_time:        %llu.%09u \n", netw_clk_time.seconds, netw_clk_time.nanoseconds);
        printf("until_cycle_start:    %lld.%09d \n", until_cycle_start.seconds, until_cycle_start.nanoseconds);
        printf("TSN-SLOT EXCEEDED, halting!!!!\n");
        while(1){}
    }
#endif
    // rearm RTC Clock interrupt
    patmos_arm_usec_timer(config->timer_cfg.timer_base_address, config->slot_end_time_stamp);
}


static void handle_timer_interrupt(void) __attribute__((naked));
static void handle_timer_interrupt(void)
{
    exc_prologue();
    
    handle_tsn_schedule();
    
    exc_epilogue();
} 


/**
 * @brief Initialize Ethernet MAC 
 *  
 * This function initializes the MAC Address and the Receive Buffer,
 * Configures the Ethernet MAC
 * 
 * @param char enable_int   : set to 1 to enable interrupt driven 
 *                            operation of ethernet driver.
 * 
 */
mmio_ptr_t tsn_mac_initialize(device_config_t *config){ 
    
    // copy data into in memory configuration
    tsn_runtime_config_ptr_t in_mem_cfg = (tsn_runtime_config_ptr_t) config->buffer_mgmt_config.start_address;
    mmio_ptr_t start_address = (mmio_ptr_t) config->buffer_mgmt_config.start_address;

    // copy memory addresses
    in_mem_cfg->start_register_address = (mmio_ptr_t) config->mmio_config.start_register_address;
    in_mem_cfg->start_memory_address = start_address;

    // copy configuration
    in_mem_cfg->interupt_id = config->interupt_id;
    in_mem_cfg->interupt_ena = config->interupt_ena;
    in_mem_cfg->list_pointer = 0;
    in_mem_cfg->current_gate_state = 0;
    in_mem_cfg->cycle_start.seconds=0;
    in_mem_cfg->cycle_start.nanoseconds=0;
    in_mem_cfg->timer_cfg.interupt_ena = config->timer_cfg.interupt_ena;
    in_mem_cfg->timer_cfg.interupt_id = config->timer_cfg.interupt_id;
    in_mem_cfg->timer_cfg.timer_base_address = config->timer_cfg.timer_base_address;

    if(config->rx_buffers == 0 || config->rx_buffers > DEVICE_MAX_RX_BUFFERS)
    {
        in_mem_cfg->num_rx_buffers = 1;
    }
    else
    {
        in_mem_cfg->num_rx_buffers = config->rx_buffers;
    }
    
    // copy mac address
    in_mem_cfg->mac_address.address[0] = config->mac_address.address[0];
    in_mem_cfg->mac_address.address[1] = config->mac_address.address[1];
    in_mem_cfg->mac_address.address[2] = config->mac_address.address[2];
    in_mem_cfg->mac_address.address[3] = config->mac_address.address[3];
    in_mem_cfg->mac_address.address[4] = config->mac_address.address[4];
    in_mem_cfg->mac_address.address[5] = config->mac_address.address[5];

    // reserve space for queuing frames
    start_address += sizeof(tsn_runtime_config_t);
    for(uint8_t i=0; i<TSN_MAC_QUEUE_COUNT; i++)
    {
        in_mem_cfg->traffic_queues[i] = (queue_ptr_t)start_address;
        start_address += sizeof(queue_t);

        in_mem_cfg->traffic_queues[i]->read_ptr=0;
        in_mem_cfg->traffic_queues[i]->write_ptr=0;
        in_mem_cfg->traffic_queues[i]->length=config->queue_length;
        in_mem_cfg->traffic_queues[i]->queue = (queue_entry_ptr_t)start_address;
        start_address += sizeof(queue_entry_t)*config->queue_length;
        for(uint16_t j=0; j<=config->queue_length; j++)
        {
            in_mem_cfg->traffic_queues[i]->queue[j].frame_length=0;
        }
    }
    
    // reserve space for queuing frames
    in_mem_cfg->receive_queue = (queue_ptr_t)start_address;
    start_address += sizeof(queue_t);

    in_mem_cfg->receive_queue->read_ptr=0;
    in_mem_cfg->receive_queue->read_ptr=0;
    in_mem_cfg->receive_queue->length=config->queue_length;
    in_mem_cfg->receive_queue->queue = (queue_entry_ptr_t)start_address;
    for(uint16_t j=0; j<=config->queue_length; j++)
    {
        in_mem_cfg->receive_queue->queue[j].frame_length=0;
    }
    start_address += sizeof(queue_entry_t) * config->queue_length;
    
    // Store whether a frame has been received by the eth mac 
    in_mem_cfg->last_filled_bd = RX_BD_ADDR_BASE(mmio_rd_32(in_mem_cfg->start_register_address,TX_BD_NUM_ADDR)); 

    // reserve space for schedule
    in_mem_cfg->admin_schedule = (tsn_schedule_config_ptr_t) start_address;
    start_address += sizeof(tsn_schedule_config_t) ;
    in_mem_cfg->admin_schedule->gate_control_list = (tsn_gcl_entry_ptr_t)start_address;
    start_address += sizeof(tsn_gcl_entry_t) * TSN_MAC_MAX_GCL_LENGTH;

    in_mem_cfg->oper_schedule = (tsn_schedule_config_ptr_t) start_address;
    start_address += sizeof(tsn_schedule_config_t) ;
    in_mem_cfg->oper_schedule->gate_control_list = (tsn_gcl_entry_ptr_t)start_address;
    start_address += sizeof(tsn_gcl_entry_t) * TSN_MAC_MAX_GCL_LENGTH;

    for(uint16_t i=0; i<=TSN_MAC_MAX_GCL_LENGTH; i++)
    {
        in_mem_cfg->admin_schedule->gate_control_list[i].gate_state=0;
        in_mem_cfg->admin_schedule->gate_control_list[i].slot_duration=0;
        in_mem_cfg->oper_schedule->gate_control_list[i].gate_state=0;
        in_mem_cfg->oper_schedule->gate_control_list[i].slot_duration=0;
    }

    // Set default schedule : all queues open
    in_mem_cfg->oper_schedule->gate_control_list[0].gate_state=0xFF;
    in_mem_cfg->current_gate_state = 0xFF;

    // Set MAC Address
	uint32_t reg_val = (config->mac_address.address[2] << 24);
    reg_val |= (config->mac_address.address[3] << 16);
    reg_val |= (config->mac_address.address[4] << 8);
    reg_val |= (config->mac_address.address[5]);

    mmio_wr_32(in_mem_cfg->start_register_address, MAC_ADDR0, reg_val);
    reg_val  = (config->mac_address.address[0] << 8 );
    reg_val |= (config->mac_address.address[1]);
	mmio_wr_32(in_mem_cfg->start_register_address, MAC_ADDR1, reg_val);

    //MODEREG: PAD|HUGEN|CRCEN|DLYCRCEN|-|FULLD|EXDFREN|NOBCKOF|LOOPBCK|IFG|PRO|IAM|BRO|NOPRE|TXEN|RXEN
    reg_val  = PAD_BIT | CRCEN_BIT | FULLD_BIT | PRO_BIT | BRO_BIT | TXEN_BIT | RXEN_BIT ; // == 0x0000A423
    mmio_wr_32(in_mem_cfg->start_register_address,MODER_ADDR, reg_val);
    
    // set new start address and register new buffer management entry 
    config->buffer_mgmt_config.start_address = (void *) start_address;
    in_mem_cfg->buffer_mgmt_id = buffer_management_register(&config->buffer_mgmt_config);
    
    uint32_t rx_bd_addr= RX_BD_ADDR_BASE(mmio_rd_32(in_mem_cfg->start_register_address, TX_BD_NUM_ADDR));

    reg_val = RX_BD_EMPTY_BIT | RX_BD_IRQEN_BIT;
    for(uint16_t i=0; i< in_mem_cfg->num_rx_buffers; i++)
    {
        // initialise default RX Buffer descriptor and request buffer from Buffer Management
        mmio_ptr_t rx_buffer = (mmio_ptr_t) buffer_management_malloc(ETHERNET_MAX_FRAME_SIZE, in_mem_cfg->buffer_mgmt_id);
        if(rx_buffer == NULL)
        {
            return NULL;
        }

        mmio_wr_32(in_mem_cfg->start_register_address, rx_bd_addr +4, (uint32_t) rx_buffer); 

        if(i == in_mem_cfg->num_rx_buffers-1)
        {
            // set wrap bit at last rx buffer
            reg_val |= RX_BD_WRAP_BIT;
        }
        mmio_wr_32(in_mem_cfg->start_register_address, rx_bd_addr, reg_val);
        rx_bd_addr +=8;
    }
    
    if(in_mem_cfg->interupt_ena != 0)
    {
        // enable receive Receive and transmit interruptsn
        exc_register(in_mem_cfg->interupt_id, &handle_eth_interrupt);
        intr_unmask(in_mem_cfg->interupt_id);

        reg_val  = INT_MASK_RXF_BIT | INT_MASK_TXB_BIT;
        mmio_wr_32(in_mem_cfg->start_register_address,INT_MASK_ADDR , reg_val);
    }

    if(in_mem_cfg->timer_cfg.interupt_ena != 0)
    {
        // enable receive Receive and transmit interruptsn
        exc_register(in_mem_cfg->timer_cfg.interupt_id, &handle_timer_interrupt);
        intr_unmask(in_mem_cfg->timer_cfg.interupt_id);
    }
    
    configurations[interface_cnt] = in_mem_cfg;
    interface_cnt++;
    
    #ifdef WCET_ANALYSIS
        handle_tsn_schedule();
        do_eth_interrupt_stuff();
    #endif

    return (mmio_ptr_t) in_mem_cfg;
}

network_error_t tsn_mac_update_gcl(tsn_gcl_config_t *gcl_config, mmio_ptr_t config_ptr)
{
    uint64_t cycle_duration=0;
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;

    if(gcl_config->gcl_length > TSN_MAC_MAX_GCL_LENGTH)
    {
        return NET_ERR_INPUT_VALUE;
    }

    #pragma loopbound min 1 max TSN_MAC_MAX_GCL_LENGTH
    for(uint32_t i=0; i<gcl_config->gcl_length; i++)
    {
        config->admin_schedule->gate_control_list[i].gate_state = gcl_config->entry_list[i].gate_state;
        // ignore this step during WCET analysis since the compiler has no annotated 64 bit divisions
        #ifndef WCET_ANALYSIS
        // Slot duration is given as ns but will be interpreted as us due to hardware limitations
        config->admin_schedule->gate_control_list[i].slot_duration = gcl_config->entry_list[i].slot_duration/1000; 
        #endif
        if(config->admin_schedule->gate_control_list[i].slot_duration < TSN_MAC_MIN_SLOT_INTERVAL_US)
        {
            return NET_ERR_DRVR_TIMESLOT_TO_SHORT;
        }
        
        cycle_duration += gcl_config->entry_list[i].slot_duration;
    }
    
    if(cycle_duration + gcl_config->min_guard_band_duration > gcl_config->cycle_duration ||
        cycle_duration > (uint64_t) SEC_TO_NS)
    {
        return NET_ERR_INPUT_VALUE;
    }


    config->admin_schedule->gcl_length = gcl_config->gcl_length;
    // Slot duration is given as ns but will be interpreted as us due to hardware limitations
    config->admin_schedule->cycle_time = (uint32_t) gcl_config->cycle_duration;
    config->admin_schedule->update_time = gcl_config->admin_cycle_time;
    
    // If update time is 0, the gcl shall be updated immidiately 
    // If update time is less then the current time, the gcl shall be updated when the next cycle starts
    time_stamp_t current_time;
    (void) tsn_mac_get_netw_clk_timestamp((mmio_ptr_t) config, & current_time);
    
    if( config->admin_schedule->update_time.seconds == 0 &&
        config->admin_schedule->update_time.nanoseconds == 0)
    {
        unsigned intr_en = EXC_STATUS & 0x1;
        if(intr_en)
        {
            CLI();
        }

        tsn_schedule_config_ptr_t tmp = config->oper_schedule;
        config->oper_schedule = config->admin_schedule;
        config->admin_schedule = tmp;
        
        config->cycle_start = current_time;
         
        config->list_pointer = 0;
        config->current_gate_state = config->oper_schedule->gate_control_list[0].gate_state;
        config->time_interval = config->oper_schedule->gate_control_list[0].slot_duration;
        
        config->slot_end_time_stamp = patmos_get_cpu_usecs(config->timer_cfg.timer_base_address) + config->time_interval;
        
        if(intr_en)
        {
            SEI();
        }
        
        patmos_arm_usec_timer(config->timer_cfg.timer_base_address, config->slot_end_time_stamp);
        tsn_mac_update_transmission((mmio_ptr_t) config);
    }
    else if(config->admin_schedule->update_time.seconds <= current_time.seconds || 
            (config->admin_schedule->update_time.seconds == current_time.seconds && config->admin_schedule->update_time.nanoseconds <= current_time.nanoseconds))
    {
        // Do nothing wait until schedule ends
    }
    else 
    {
        // calculate interval until update shall be executed
        time_interval_t until_cycle_start;
        until_cycle_start.seconds = config->admin_schedule->update_time.seconds - current_time.seconds;
        until_cycle_start.nanoseconds = config->admin_schedule->update_time.nanoseconds - current_time.nanoseconds;
        if(until_cycle_start.nanoseconds <0)
        {
            until_cycle_start.nanoseconds += SEC_TO_NS;
            until_cycle_start.seconds--;
        }

        uint64_t until_slot_end =  config->slot_end_time_stamp-patmos_get_cpu_usecs(config->timer_cfg.timer_base_address);
        until_slot_end *= USEC_TO_NS;

        if(until_cycle_start.seconds == 0 && until_cycle_start.nanoseconds < until_slot_end)
        {
            uint64_t timer_update_time = patmos_get_cpu_usecs(config->timer_cfg.timer_base_address);
            timer_update_time+= until_cycle_start.nanoseconds/USEC_TO_NS;

            patmos_arm_usec_timer(config->timer_cfg.timer_base_address, timer_update_time);
        }
        
    }
    
    return NET_ERR_OK;
}

network_error_t tsn_mac_send(network_frame_t *frame, uint64_t timeout, mmio_ptr_t config_ptr)
{
    return NET_ERR_NOT_SUPPORTED;
}
  
network_error_t tsn_mac_send_nb(network_frame_t *frame, mmio_ptr_t config_ptr)
{
    network_error_t ret = NET_ERR_OK;
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    if(frame->priority >= TSN_MAC_QUEUE_COUNT)
    {
        return NET_ERR_DRVR_PRIORITY;
    }
    
    // check if there is space left to queue the frame
    if (queue_full(config->traffic_queues[frame->priority]) == 1)
    {
        return NET_ERR_DRVR_TX_QUEUE_FULL; // No space to queue the frame
    }

    uint32_t frame_length = frame->header_length + frame->user_data_length;
    if(frame_length > ETHERNET_MAX_FRAME_SIZE)
    {
        return NET_ERR_DRVR_FRAME_TOO_LONG;
    }

    mmio_ptr_t buf_address = (mmio_ptr_t) buffer_management_malloc(frame_length, config->buffer_mgmt_id);
    // check if buffer request was successfull
    if(buf_address == NULL )
    {
#ifndef WCET_ANALYSIS
        buffer_management_memory_dump(config->buffer_mgmt_id);
#endif
        return NET_ERR_DRVR_MEM_FATAL; // Unable to acquire buffer, return error 
    }
    
    mmio_ptr_t address = (mmio_ptr_t) buf_address;
    // write header data into buffer
    mmio_wr_block(address, frame->header_data, frame->header_length);
    // adapt address and get next frame segment
    address += frame->header_length;
    
    // user data can be null sometimes with TCP frames
    if(frame->user_data != NULL)
    {
        // write user data into buffer
        mmio_wr_block(address, frame->user_data, frame->user_data_length);    
    }

    // enques the frame 
    time_stamp_t dummy_ts;
    ret = enqueue_frame(buf_address , frame_length, &dummy_ts, config->traffic_queues[frame->priority]);
    // frame will be sent later if it could be enqueue
    if(ret == NET_ERR_OK)
    {
        ret = NET_ERR_DRVR_FRAME_QUEUED;
    }
    // if the frame was not queued, buffer has to be released again
    else
    {
        buffer_management_free((void*)buf_address, config->buffer_mgmt_id);
        return ret;   
    }
    
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    uint8_t in_use = (mmio_rd_32(config->start_register_address,TX_BD_ADDR_BASE) & TX_BD_READY_BIT); 
    mmio_ptr_t bd_send_buffer = (mmio_ptr_t)  mmio_rd_32(config->start_register_address,TX_BD_ADDR_BASE+4);
 
    // free memory address from sent buffer
    if(in_use == 0 && bd_send_buffer != NULL)
    {
        buffer_management_free((void*)bd_send_buffer, config->buffer_mgmt_id);
        // configure NULL address to indicate no buffer is in the bd
        mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE+4, 0);

    }
    
    if(in_use != 0)
    {
        if(intr_en)
        {
            SEI();
        }
        return ret;
    }
    // MAC is free equeue a waiting frame
    // get frame from queue
    queue_entry_t entry;
    uint8_t queue_id;
    uint8_t cnt =0;

    #pragma loopbound min 1 max TSN_MAC_QUEUE_COUNT
    for(queue_id=TSN_MAC_QUEUE_COUNT-1; queue_id < TSN_MAC_QUEUE_COUNT; queue_id--)
    {
        if(config->current_gate_state & (1 << queue_id))
        {
            if(queue_empty(config->traffic_queues[queue_id]) == 0)
            {
                break;
            }
            cnt ++;
        }
    }

    if(queue_id >= TSN_MAC_QUEUE_COUNT )
    {
        // no open queue with a frame in it
        if(intr_en)
        {
            SEI();
        }
        
        if(cnt > 0)
        {
            return NET_ERR_DRVR_FRAME_QUEUED;
        }
        else
        {
            return NET_ERR_DRVR_TSN_QUEUES_CLOSED;
        }
    }
    
    frame_length = get_next_frame_length(config->traffic_queues[queue_id]);
    
    if(frame_length < ETHERNET_MIN_FRAME_SIZE){ frame_length=ETHERNET_MIN_FRAME_SIZE;};

    // send_time(us) = data_size(Byte)/100(MBit/s) 
    //               = (8*data_size) Bit *10^9ns / 10^8 Bit
    //               = 80*data_size*ns 
    //               = 80*data_size*us/1000
    // Additional overhead is added for code execution and FCS calculation
    int64_t send_time = (frame_length*80)/1000 + TSN_MAC_SEND_OVERHEAD_US;
    int64_t remaining_time =config->slot_end_time_stamp;
    remaining_time -= patmos_get_cpu_usecs(config->timer_cfg.timer_base_address);

    if(send_time > remaining_time && config->slot_end_time_stamp >0) 
    {
        // time slot is not long enough to send full frame
        if(intr_en)
        {
            SEI();
        }
        
        return NET_ERR_DRVR_TIMESLOT_TO_SHORT;
    }

    ret = dequeue_frame(&entry, config->traffic_queues[queue_id]);
    if( ret == NET_ERR_DRVR_TX_QUEUE_EMPTY)
    {
        // frame was dequeued between enque and disabling of interrupts
        // no error
        if(intr_en)
        {
            SEI();
        }
        
        return NET_ERR_OK;
    }
    else
    {
        if(buf_address == entry.buffer_address)
        {
            ret = NET_ERR_OK;
        }
        // configure interrupt, buffer descriptor and start transmission        
        buf_address = entry.buffer_address;
        frame_length =  entry.frame_length;
    }
    
    // clear interrupt and set buffer descriptor        
    mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_TXB_BIT);
    mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE+4, (uint32_t) buf_address);
    uint32_t txbd = frame_length<<16; 
    txbd |= TX_BD_READY_BIT | TX_BD_IRQEN_BIT | TX_BD_WRAP_BIT | TX_BD_PAD_EN_BIT;

    mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE, txbd);
    
    if(intr_en)
    {
        SEI();
    }
    
    return ret;   // frame is beeing sent
}

network_error_t tsn_mac_update_transmission(mmio_ptr_t config_ptr)
{
    network_error_t ret = NET_ERR_OK;
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    // check if the buffer is ready for a new transmission 
    uint8_t in_use = (mmio_rd_32(config->start_register_address, TX_BD_ADDR_BASE) & TX_BD_READY_BIT); 
    if(in_use !=0)
    {
        ret = NET_ERR_DRVR_BUFFER_BUSY;
    }
    else
    {
        // Free transmission Buffer as no longer used
        mmio_ptr_t buffer =(mmio_ptr_t) mmio_rd_32(config->start_register_address, TX_BD_ADDR_BASE+4);
        if(buffer != NULL)
        {
            buffer_management_free((void*)buffer, config->buffer_mgmt_id);
        }

        // clear interrupt flag
        //mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_TXB_BIT);
        
        // MAC is free equeue a waiting frame
        // get frame from queue
        queue_entry_t entry;
        uint8_t queue_id;
        
        uint8_t cnt=0;

        #pragma loopbound min 1 max TSN_MAC_QUEUE_COUNT
        for(queue_id=TSN_MAC_QUEUE_COUNT-1; queue_id < TSN_MAC_QUEUE_COUNT; queue_id--)
        {
            if(config->current_gate_state & (1 << queue_id))
            {
                if(queue_empty(config->traffic_queues[queue_id]) == 0)
                {
                    break;
                }

                cnt++;
            }
        }

        if(queue_id >= TSN_MAC_QUEUE_COUNT)
        {
            // configure NULL address to indicate no buffer is in the bd
            mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE+4, 0);
            if(cnt >0)
            {
                ret = NET_ERR_DRVR_TX_QUEUE_EMPTY;
            }
            else
            {
                ret = NET_ERR_DRVR_TSN_QUEUES_CLOSED;
            }
        }
        else
        {
            uint32_t frame_length = get_next_frame_length(config->traffic_queues[queue_id]);
        
            if(frame_length < ETHERNET_MIN_FRAME_SIZE){ frame_length=ETHERNET_MIN_FRAME_SIZE;};

            // send_time(us) = data_size(Byte)/100(MBit/s) 
            //               = (8*data_size) Bit *10^9ns / 10^8 Bit
            //               = 80*data_size*ns 
            //               = 80*data_size*us/1000
            // Additional overhead is added for code execution and FCS calculation
            int64_t send_time = (frame_length*80)/1000 + TSN_MAC_SEND_OVERHEAD_US;
            int64_t remaining_time =config->slot_end_time_stamp;
            remaining_time -= patmos_get_cpu_usecs(config->timer_cfg.timer_base_address);

            if(send_time > remaining_time) 
            {
                mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE+4, 0);
                ret = NET_ERR_DRVR_TIMESLOT_TO_SHORT;
            }
            else
            {
                // get next frame
                if(dequeue_frame(&entry, config->traffic_queues[queue_id]) == NET_ERR_DRVR_TX_QUEUE_EMPTY)
                {
                    // configure NULL address to indicate no buffer is in the bd
                    mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE+4, 0);
                    ret = NET_ERR_DRVR_TX_QUEUE_EMPTY;
                }
                else
                {
                    // configure interrupt, buffer descriptor and start transmission        
                    mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE+4, (uint32_t) entry.buffer_address);
                    uint32_t txbd = entry.frame_length<<16; 
                    txbd |= TX_BD_READY_BIT | TX_BD_IRQEN_BIT | TX_BD_WRAP_BIT | TX_BD_PAD_EN_BIT;
                    mmio_wr_32(config->start_register_address,TX_BD_ADDR_BASE, txbd);
                    ret = NET_ERR_OK;
                }
            }
        }
    }

    if(intr_en)
    {
        SEI();
    }
    return ret;
}

network_error_t tsn_mac_receive(uint8_t *rx_buffer, uint16_t buffer_size, uint64_t timeout, mmio_ptr_t config_ptr)
{
    time_stamp_t rx_ts;    
    // return value of -1 indicates no frame waiting
    network_error_t ret=NET_ERR_DRVR_NO_FRAME_RECEIVED;
    // Wait until a frame is received
    if (timeout == 0){
        #pragma loopbound min 1 max 1  //1us - 10ms
		while (ret==NET_ERR_DRVR_NO_FRAME_RECEIVED){
            ret = tsn_mac_receive_nb(rx_buffer, buffer_size,  &rx_ts, config_ptr);        
        };
	}
    // wait until a frame is received or timeout is exceeded
    else
    {
		uint64_t start_time = get_cpu_usecs();
        #pragma loopbound min 1 max 1   //1us
		while (ret == NET_ERR_DRVR_NO_FRAME_RECEIVED){
            if((get_cpu_usecs()-start_time >= timeout))
            {
                return NET_ERR_DRVR_NO_FRAME_RECEIVED;
            }
            ret = tsn_mac_receive_nb(rx_buffer, buffer_size, &rx_ts, config_ptr);
        }
	}
    return ret;
}

network_error_t tsn_mac_receive_nb(uint8_t *rx_buffer, uint16_t buffer_size, time_stamp_t *rx_ts, mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    uint32_t rx_bd = RX_BD_ADDR_BASE(mmio_rd_32(config->start_register_address, TX_BD_NUM_ADDR)); 
    
    mmio_ptr_t rx_buffer_tsn_mac;
    uint32_t frame_length;

    // get next frame of queue
    if(queue_empty(config->receive_queue) == 0 )
    {    
        
        frame_length = get_next_frame_length(config->receive_queue);
        // check if a frame has been stored
        if(frame_length == 0)
        {
            return NET_ERR_DRVR_NO_FRAME_RECEIVED;
        }
        // check if frame fits in buffer
        else if(frame_length > buffer_size)
        {
            return NET_ERR_BUFF_UNDERSIZED;
        }
        // remove frame from queue
        queue_entry_t entry;
        dequeue_frame(&entry, config->receive_queue);
        rx_buffer_tsn_mac = entry.buffer_address;
        rx_ts->nanoseconds = entry.ts.nanoseconds;
        rx_ts->seconds = entry.ts.seconds;
    }
    else
    {
        // If there was no frame received by the mac and queued
        // or there was no frame in the queue no framme 
        // is there to be processed
        return NET_ERR_DRVR_NO_FRAME_RECEIVED;
    }
    
    // copy frame to buffer
    mmio_rd_block(rx_buffer_tsn_mac, rx_buffer, frame_length);

    // free buffer of received frame again
    buffer_management_free((void*)rx_buffer_tsn_mac, config->buffer_mgmt_id);

    return NET_ERR_OK;
}

network_error_t tsn_mac_update_reception(mmio_ptr_t config_ptr)
{
    network_error_t ret = NET_ERR_DRVR_NO_FRAME_RECEIVED;

    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;

    uint32_t frame_in_tsn_mac = mmio_rd_32(config->start_register_address, INT_SOURCE_ADDR) & INT_SOURCE_RXB_BIT; 
    //uint32_t frame_in_tsn_mac = (mmio_rd_32(config->start_register_address,rx_bd) & RX_BD_EMPTY_BIT ) == 0; 
    
    unsigned intr_en = EXC_STATUS & 0x1;

    // Check if frame is waiting in the ethernet MAC
    //if(frame_in_tsn_mac!=0)
    {
        // clear interrupt bit
        mmio_wr_32(config->start_register_address,INT_SOURCE_ADDR, INT_SOURCE_RXB_BIT);
        uint32_t reg_val = RX_BD_EMPTY_BIT | RX_BD_IRQEN_BIT;
        
        mmio_ptr_t rx_buffer_tsn_mac;
        uint32_t frame_length;
        
        // Store whether a frame has been received by the eth mac 
        uint32_t first_rx_bd = RX_BD_ADDR_BASE(mmio_rd_32(config->start_register_address,TX_BD_NUM_ADDR)); 
        uint32_t last_rx_bd =  first_rx_bd + ((config->num_rx_buffers-1)<<3);
        uint32_t rx_bd = config->last_filled_bd+8;

        uint32_t last_filled_bd=config->last_filled_bd;
        //DEVICE_MAX_RX_BUFFERS
        #pragma loopbound min 1 max DEVICE_MAX_RX_BUFFERS
        for(uint16_t i=0; i < config->num_rx_buffers; i++)
        {
            if(rx_bd > last_rx_bd)
            {
                rx_bd = first_rx_bd;
            }

            uint32_t buffer_in_bd = (mmio_rd_32(config->start_register_address,rx_bd) & RX_BD_EMPTY_BIT) == 0 ;
            if(buffer_in_bd)
            {
                last_filled_bd = rx_bd;

                if(intr_en)
                {
                    CLI();
                }
                // enque the received frame, a frame is always queued to simplify the logic
                rx_buffer_tsn_mac = (mmio_ptr_t) mmio_rd_32(config->start_register_address,rx_bd+4);
                frame_length = (mmio_rd_32(config->start_register_address,rx_bd) & 0xFFFF0000) >> 16;
                
                time_stamp_t rx_ts;
                ret =tsn_mac_get_netw_clk_rx_timestamp(config_ptr, &rx_ts);
                if(ret == NET_ERR_TIMESTAMP_NOT_AVAILABLE)
                {
                    tsn_mac_get_netw_clk_timestamp(config_ptr, &rx_ts);
                }

                if(frame_length >= ETHERNET_MIN_FRAME_SIZE)
                {
                    // return if frame was not enqueued
                    if(enqueue_frame(rx_buffer_tsn_mac, frame_length, &rx_ts, config->receive_queue) != NET_ERR_OK)
                    {
                        if(intr_en)
                        {
                            SEI();
                        }
                        return NET_ERR_DRVR_RX_QUEUE_FULL;
                    }
                    
                    // request new buffer for receiving a new frame
                    mmio_ptr_t new_rx_buffer_addr = (mmio_ptr_t) buffer_management_malloc(ETHERNET_MAX_FRAME_SIZE, config->buffer_mgmt_id);
                    
                    // check if buffer request was successfull
                    if(new_rx_buffer_addr == NULL)
                    {
                        if(intr_en)
                        {
                            SEI();
                        }
                        return NET_ERR_DRVR_NO_FREE_BUFFER; // Unable to acquire buffer, return error 
                    }
                    mmio_wr_32(config->start_register_address,rx_bd+4, (uint32_t) new_rx_buffer_addr);
                }
                
                // clear interrupt and configure buffer descriptor
                //if(i == config->num_rx_buffers-1)
                if(rx_bd == last_rx_bd)
                {
                    // set wrap bit in last buffer descriptor
                    //reg_val |= RX_BD_WRAP_BIT;
                    mmio_wr_32(config->start_register_address,rx_bd, reg_val| RX_BD_WRAP_BIT);
                }
                else
                {
                    mmio_wr_32(config->start_register_address,rx_bd, reg_val);
                }

                if(intr_en)
                {
                    SEI();
                }   
            }
            rx_bd += 8;
        }
        config->last_filled_bd = last_filled_bd;
    }
    
    return ret;
}


uint16_t tsn_mac_has_frame(mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    
    uint16_t recv = ~0;
    uint32_t rx_bd = RX_BD_ADDR_BASE(mmio_rd_32(config->start_register_address,TX_BD_NUM_ADDR)); 

    if(queue_empty(config->receive_queue) == 0 )
    {
        recv= get_next_frame_length(config->receive_queue);
    }
    /*
    else if ((mmio_rd_32(config->start_register_address,INT_SOURCE_ADDR) & INT_SOURCE_RXB_BIT) !=0){
        uint32_t frame_length = mmio_rd_32(config->start_register_address,rx_bd) >> 16;
        recv = (uint16_t) frame_length;
    }
    */
    return recv;
}

///////////////////////////////////////////////////////////////
// Network Clock functionality
///////////////////////////////////////////////////////////////


uint64_t get_netw_clk_usecs(mmio_ptr_t config_ptr){
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;

    uint64_t seconds =  (uint64_t) TSN_MAC_RTC_TIME_SEC_HI(config->start_memory_address) << 32;
    seconds |= (uint64_t) TSN_MAC_RTC_TIME_SEC(config->start_memory_address);
    uint32_t microseconds = (uint32_t) TSN_MAC_RTC_TIME_NS(config->start_memory_address);
    microseconds /= 1000;
    
    uint64_t timestamp = seconds * SEC_TO_USEC + (uint64_t)microseconds;
	return timestamp;
}

network_error_t tsn_mac_get_netw_clk_timestamp(mmio_ptr_t config_ptr, time_stamp_t *ts){

    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;


    ts->nanoseconds = TSN_MAC_RTC_TIME_NS(config->start_memory_address);
    ts->seconds = (uint64_t) TSN_MAC_RTC_TIME_SEC_HI(config->start_memory_address) << 32;
    ts->seconds |= TSN_MAC_RTC_TIME_SEC(config->start_memory_address);


    return NET_ERR_OK;
}

network_error_t tsn_mac_get_netw_clk_rx_timestamp(mmio_ptr_t config_ptr, time_stamp_t *ts){

    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;

    uint32_t time_stamp_available = TSN_MAC_PTP_RXCHAN_STATUS(config->start_memory_address);

    ts->nanoseconds = TSN_MAC_PTP_RXCHAN_TIMESTAMP_NS(config->start_memory_address);
    ts->seconds = (uint64_t)TSN_MAC_PTP_RXCHAN_TIMESTAMP_SEC_HI(config->start_memory_address) << 32;
    ts->seconds |= TSN_MAC_PTP_RXCHAN_TIMESTAMP_SEC(config->start_memory_address);

    TSN_MAC_PTP_RXCHAN_STATUS(config->start_memory_address) = 0;
    if((time_stamp_available & 0x1) !=0)
    {
        return NET_ERR_OK;
    }
    else
    {
        return NET_ERR_TIMESTAMP_NOT_AVAILABLE;
    }
}

network_error_t tsn_mac_get_netw_clk_tx_timestamp(mmio_ptr_t config_ptr, time_stamp_t *ts){

    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;

    uint32_t time_stamp_available = TSN_MAC_PTP_TXCHAN_STATUS(config->start_memory_address);
    
    ts->nanoseconds = TSN_MAC_PTP_TXCHAN_TIMESTAMP_NS(config->start_memory_address);
    ts->seconds = (uint64_t)TSN_MAC_PTP_TXCHAN_TIMESTAMP_SEC_HI(config->start_memory_address) << 32;
    ts->seconds |= TSN_MAC_PTP_TXCHAN_TIMESTAMP_SEC(config->start_memory_address);

    TSN_MAC_PTP_TXCHAN_STATUS(config->start_memory_address) = 0;
    if((time_stamp_available & 0x1)!=0)
    {
        return NET_ERR_OK;
    }
    else
    {
        return NET_ERR_TIMESTAMP_NOT_AVAILABLE;
    }
}

network_error_t tsn_mac_set_netw_clk_timestamp(mmio_ptr_t config_ptr, time_stamp_t *time){
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;


    TSN_MAC_RTC_TIME_SEC_HI(config->start_memory_address) = (uint32_t) ((time->seconds >> 32) & 0xFFFFFFFF);	//reverse order to load time operand last
    TSN_MAC_RTC_TIME_SEC(config->start_memory_address) =(uint32_t) (time->seconds & 0xFFFFFFFF);	//reverse order to load time operand last
    TSN_MAC_RTC_TIME_NS(config->start_memory_address) = time->nanoseconds;	//reverse order to load time operand last
    
    return NET_ERR_OK;
}


network_error_t tsn_mac_netw_clk_correct_offset(mmio_ptr_t config_ptr, time_interval_t *offset){
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    
	if(offset->seconds != 0)
    {
    
        int64_t seconds_calc = (int64_t)TSN_MAC_PTP_TXCHAN_TIMESTAMP_SEC_HI(config->start_memory_address) << 32;
        seconds_calc |= TSN_MAC_PTP_TXCHAN_TIMESTAMP_SEC(config->start_memory_address);
        seconds_calc += offset->seconds;
        uint64_t seconds = (uint64_t) llabs(seconds_calc);

        // Check for overflow is not required, HW implementation handles that
		TSN_MAC_RTC_TIME_SEC_HI(config->start_memory_address) = (uint32_t) ((seconds >> 32)); // + TSN_MAC_RTC_TIME_SEC(config->start_memory_address));	//reverse order to load time operand last
		TSN_MAC_RTC_TIME_SEC(config->start_memory_address) =(uint32_t) ((seconds &0xFFFFFFFF)); // + TSN_MAC_RTC_TIME_SEC(config->start_memory_address));	//reverse order to load time operand last
        TSN_MAC_RTC_TIME_NS(config->start_memory_address) = offset->nanoseconds + TSN_MAC_RTC_TIME_NS(config->start_memory_address);	//reverse order to load time operand last
	} else {
        TSN_MAC_RTC_ADJUST_OFFSET(config->start_memory_address) = offset->nanoseconds;
	}
    
    return NET_ERR_OK;
}

///////////////////////////////////////////////////////////////
//  Configuration functions
///////////////////////////////////////////////////////////////

/**
 * @brief Enables Transmission
 *  
 * This function enables transmission of ethernet frames.
 */
void tsn_mac_enable_tx(mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address, MODER_ADDR);
    mmio_wr_32(config->start_register_address, MODER_ADDR, data | TXEN_BIT);
}

/**
 * @brief Disables Transmission
 *  
 * This function disable transmission of ethernet frames.
 */
void tsn_mac_disable_tx(mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address,MODER_ADDR);
    mmio_wr_32(config->start_register_address,MODER_ADDR, data & ~TXEN_BIT);
}

/**
 * @brief Transmission is enabled
 *  
 * This function returns whether transmission is enabled or
 * not.
 * 
 * @return char enabled     : returns 1 if transmission is 
 *                            enabled, 0 otherwise.
 */
char tsn_mac_tx_enabled(mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address,MODER_ADDR);
    if(data & TXEN_BIT)
    {
        return 1;
    }
    return 0;
}

/**
 * @brief Enables Reception
 *  
 * This function enables reception of ethernet frames.
 */
void tsn_mac_enable_rx(mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address,MODER_ADDR);
    mmio_wr_32(config->start_register_address,MODER_ADDR, data | RXEN_BIT);
}

/**
 * @brief Disables Reception
 *  
 * This function disable reception of ethernet frames.
 */
void tsn_mac_disable_rx(mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address,MODER_ADDR);
    mmio_wr_32(config->start_register_address,MODER_ADDR, data & ~RXEN_BIT);
}

/**
 * @brief Reception is enabled
 *  
 * This function returns whether reception is enabled or
 * not.
 * 
 * @return char enabled     : returns 1 if reception is 
 *                            enabled, 0 otherwise.
 */
char tsn_mac_rx_enabled(mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
    uint32_t data = mmio_rd_32(config->start_register_address,MODER_ADDR);
    if(data & RXEN_BIT)
    {
        return 1;
    }
    return 0;
}

/**
 * @brief Sets the MAC Address 
 *  
 * This function is used to set a the Ethernet MACs MAC-Address.
 * 
 * @param uint8_t *mac_addr   : Buffer which holds the MAC
 *                                    Address.
 */
void tsn_mac_set_mac_address(uint8_t *mac_addr, mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;
	uint32_t reg_val = (mac_addr[2] << 24) | (mac_addr[3] << 16) | (mac_addr[4] << 8) | (mac_addr[5]);
    mmio_wr_32(config->start_register_address,MAC_ADDR0, reg_val);
    reg_val = (mac_addr[0] << 8 ) | (mac_addr[1]);
	mmio_wr_32(config->start_register_address,MAC_ADDR1, reg_val);
}

/**
 * @brief Gets the MAC Address 
 *  
 * This function reads out the MAC-Address from the Ethernet MAC.
 * 
 * @param uint8_t *mac_addr   : Buffer where the MAC
 *                                    Address will be stored.
 */
void tsn_mac_get_mac_address(uint8_t *mac_addr, mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;

    uint64_t my_mac = mmio_rd_32(config->start_register_address, MAC_ADDR0); // Getting MSB 2 byte
    mac_addr[5] = my_mac & 0xFF;
    my_mac >>= 8;
    mac_addr[4] = my_mac & 0xFF;
    my_mac >>= 8;
    mac_addr[3] = my_mac & 0xFF;
    my_mac >>= 8;
    mac_addr[2] = my_mac & 0xFF;
    
    my_mac = mmio_rd_32(config->start_register_address,MAC_ADDR1); // Getting MSB 2 byte
    mac_addr[1] = my_mac & 0xFF;
    my_mac >>= 8;
    mac_addr[0] = my_mac & 0xFF;
}

/////////////////////
// Help functions
/////////////////////

/**
 * @brief Get frame length of next TX frame
 *  
 * This function returns the length of the frame which will be 
 * dequeued next from the transmission queue.
 * 
 * @return uint32_t frame_length    : Length of the TX Frame
 */
static uint32_t get_next_frame_length(queue_ptr_t config)
{   
    asm volatile ("" : : : "memory");
    return config->queue[config->read_ptr].frame_length;
}

/**
 * @brief Get buffer address of next TX frame
 *  
 * This function returns the buffer address of the frame which will 
 * be dequeued next from the transmission queue.
 * 
 * @return uint32_t buffer_address    : Address of the TX Frame
 */
static mmio_ptr_t get_next_address(queue_ptr_t config)
{   
    asm volatile ("" : : : "memory");
    return config->queue[config->read_ptr].buffer_address;
}

/**
 * @brief Get buffer address of last TX frame
 *  
 * This function returns the buffer address of the frame which was 
 * enqueued last at the transmission queue.
 * 
 * @return uint32_t buffer_address    : Address of the TX Frame
 */
static mmio_ptr_t get_last_address(queue_ptr_t config)
{   
    asm volatile ("" : : : "memory");
    uint16_t last_entry = (config->write_ptr-1>= config->length) ? config->length-1 : config->write_ptr-1;

    return config->queue[last_entry].buffer_address;
}

/**
 * @brief Dequeues a TX Frame
 *  
 * This function provides the data of the next tx frame queued which is ready
 * to be sent.
 * 
 * @param  queue_entry_t *entry     : Pointer where the data will be stored
 * @return uint8_t status     : Returns 0 if successful else -1
 */
static network_error_t dequeue_frame(queue_entry_t *entry, queue_ptr_t config)
{
    network_error_t ret=NET_ERR_DRVR_TX_QUEUE_EMPTY;
    // check whether interrupts are enabled
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    asm volatile ("" : : : "memory");
    
    // check whether an entry is left
    if(config->queue[config->read_ptr].frame_length != 0)
    {
        // get entry
        entry->buffer_address = config->queue[config->read_ptr].buffer_address;
        entry->frame_length = config->queue[config->read_ptr].frame_length;
        entry->ts.seconds = config->queue[config->read_ptr].ts.seconds;
        entry->ts.nanoseconds = config->queue[config->read_ptr].ts.nanoseconds;

        // mark entry as used
        config->queue[config->read_ptr].frame_length = 0;
        // calculate next read pointer
        config->read_ptr = (config->read_ptr+1>= config->length) ? 0 : config->read_ptr+1;
        ret = NET_ERR_OK;
    }

    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return ret;
}

/**
 * @brief Enqueues a TX Frame
 *  
 * This function adds a frame to the tx queue if there is still space 
 * left.
 * 
 * @param uint32_t address      : Address where frame is stored
 * @param uint32_t frame_length : Length of the Frame
 * @return char status              : 0 if enqueueing was successful, 
                                      else -1.
 */
static network_error_t enqueue_frame(mmio_ptr_t address, uint16_t frame_length, time_stamp_t *ts, queue_ptr_t config)
{
    // check whether interrupts are enabled
    network_error_t ret=NET_ERR_DRVR_TX_QUEUE_FULL;

    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    asm volatile ("" : : : "memory");

    if(config->queue[config->write_ptr].frame_length == 0)
    {
        // memory barrier 
        asm volatile ("" : : : "memory");
        // mark used buffer entry as available again
        config->queue[config->write_ptr].buffer_address = address;
        config->queue[config->write_ptr].frame_length = frame_length;
        config->queue[config->write_ptr].frame_length = frame_length;
        config->queue[config->write_ptr].ts.seconds = ts->seconds;
        config->queue[config->write_ptr].ts.nanoseconds = ts->nanoseconds;
        
        // calculate next write pointer
        config->write_ptr = (config->write_ptr+1>= config->length ) ? 0 : config->write_ptr+1;
        ret = NET_ERR_OK; 
    }
    
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return ret;
}

/**
 * @brief Check if tx queue is full
 *  
 * This function checks whether there is still space left in the tx queue.
 * 
 * @return char status          : Returns 1 if full, 0 otherwise
 */
static char queue_full(queue_ptr_t config)
{
    char ret=1;
    // check whether interrupts are enabled
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }

    asm volatile ("" : : : "memory");

    // check whether an entry is left
    if(config->queue[config->read_ptr].frame_length == 0)
    {
        ret =0;
    }
    
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return 0;
}

/**
 * @brief Check if the tx queue is empty
 *  
 * This function checks whether the tx queue is empty.
 * 
 * @return char status          : Returns 1 if empty, 0 otherwise
 */
static char queue_empty(queue_ptr_t config)
{
    // check whether interrupts are enabled,
    char ret =1;
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }
    
    asm volatile ("" : : : "memory");

    // check whether an entry is left
    if(config->queue[config->read_ptr].frame_length != 0)
    {
        ret = 0;
    }
    
    // check whether interrupts need to be reenabled
    if(intr_en)
    {
        SEI();
    }

    return ret;
}
///////////////////////////////////////////////////////////////
// Debug Functions
///////////////////////////////////////////////////////////////

void tsn_mac_print_queue_states(mmio_ptr_t config_ptr)
{
    tsn_runtime_config_ptr_t config = (tsn_runtime_config_ptr_t) config_ptr;

    #pragma loopbound min TSN_MAC_QUEUE_COUNT max TSN_MAC_QUEUE_COUNT
    for(uint8_t i=0; i<TSN_MAC_QUEUE_COUNT; i++)
    {
        uint8_t read_ptr = config->traffic_queues[i]->read_ptr;
        uint8_t write_ptr = config->traffic_queues[i]->write_ptr;
        uint8_t length =config->traffic_queues[i]->length;
        uint8_t num_entries=0;
        if( read_ptr > write_ptr)
        {
            num_entries = length-write_ptr+read_ptr;
        }
        else
        {
            num_entries = write_ptr-read_ptr;
        }
        utils_log("Queue %u: %u of %u entries read_ptr: %u write_ptr: %u\n", i,num_entries, length, read_ptr, write_ptr);
    }

    uint8_t read_ptr = config->receive_queue->read_ptr;
    uint8_t write_ptr = config->receive_queue->write_ptr;
    uint8_t length =config->receive_queue->length;
    uint8_t num_entries=0;
    if( read_ptr > write_ptr)
    {
        num_entries = length-read_ptr+write_ptr;
    }
    else
    {
        num_entries = write_ptr-read_ptr;
    }
    utils_log("Receive Queue: %u of %u entries read_ptr: %u write_ptr: %u\n",num_entries, length, read_ptr, write_ptr);
}

//#endif