/*
 * Driver for the Ethernet MAC
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 *          Christoph Lehr
 */

#ifndef TSN_MAC_DRIVER_H_
#define TSN_MAC_DRIVER_H_

#ifdef ARCH_PATMOS

#include "../../../inc/device.h"
#include "../../../inc/net_types.h"
#include "../../../inc/buffer_management.h"
#include "patmos_eth_mac.h"

#define TSN_MAC_QUEUE_COUNT     8
#define TSN_MAC_MAX_GCL_LENGTH 100
#define TSN_MAC_MAX_GCL_DURATION 1000000 // Maximum schedule duration 1 second
#define TSN_MAC_RX_QUEUE_SIZE   30
#define TSN_MAC_TX_QUEUE_SIZE   30

#define TSN_MAC_SEND_OVERHEAD_US 200 // add 10us overhead required for sending 

#ifndef TSN_MAC_CLOCK_UPDATE_INTERVAL_US
    #define TSN_MAC_CLOCK_UPDATE_INTERVAL_US 10000 // Update Intervall when it will be synced with the network interface clock
#endif

#ifndef TSN_MAC_MIN_SLOT_INTERVAL_US
    #define TSN_MAC_MIN_SLOT_INTERVAL_US 1000 // Update Intervall when it will be synced with the network interface clock
#endif


#define TSN_MAC_PTP_RXCHAN_TIMESTAMP_NS(base)       *((volatile _IODEV unsigned int *) (base+0xE000))
#define TSN_MAC_PTP_RXCHAN_TIMESTAMP_SEC(base)      *((volatile _IODEV unsigned int *) (base+0xE004))
#define TSN_MAC_PTP_RXCHAN_TIMESTAMP_SEC_HI(base)   *((volatile _IODEV unsigned int *) (base+0xE008))
#define TSN_MAC_PTP_RXCHAN_STATUS(base)             *((volatile _IODEV unsigned int *) (base+0xE00C))

#define TSN_MAC_PTP_TXCHAN_TIMESTAMP_NS(base)       *((volatile _IODEV unsigned int *) (base+0xE400))
#define TSN_MAC_PTP_TXCHAN_TIMESTAMP_SEC(base)      *((volatile _IODEV unsigned int *) (base+0xE404))
#define TSN_MAC_PTP_TXCHAN_TIMESTAMP_SEC_HI(base)   *((volatile _IODEV unsigned int *) (base+0xE408))
#define TSN_MAC_PTP_TXCHAN_STATUS(base)             *((volatile _IODEV unsigned int *) (base+0xE40C))

#define TSN_MAC_RTC_TIME_NS(base)                   *((volatile _IODEV unsigned int *) (base+0xE800))
#define TSN_MAC_RTC_TIME_SEC(base)                  *((volatile _IODEV unsigned int *) (base+0xE804))
#define TSN_MAC_RTC_TIME_SEC_HI(base)               *((volatile _IODEV unsigned int *) (base+0xE808))
#define TSN_MAC_RTC_ADJUST_OFFSET(base)             *((volatile _IODEV signed int *)   (base+0xE820))

#define TSN_MAC_RTC_NS_OFFSET_THRESHOLD 10000000

typedef struct
{
    queue_entry_ptr_t queue;
    uint8_t read_ptr;
    uint8_t write_ptr;
    uint8_t length;
} queue_t;

typedef queue_t volatile _IODEV * queue_ptr_t;

typedef tsn_gcl_entry_t volatile _IODEV * tsn_gcl_entry_ptr_t;

typedef struct {
    uint32_t gcl_length;
    tsn_gcl_entry_ptr_t gate_control_list;
    uint32_t cycle_time;
    uint32_t guard_band;
    time_stamp_t update_time;
} tsn_schedule_config_t;

typedef tsn_schedule_config_t volatile _IODEV * tsn_schedule_config_ptr_t;

typedef struct {
    uint16_t interupt_ena:1;
    uint16_t interupt_id:15;
    uint16_t num_rx_buffers;
    uint16_t buffer_mgmt_id;
    
    mac_address_t mac_address;
    mmio_ptr_t start_register_address;
    mmio_ptr_t start_memory_address;

    tsn_schedule_config_ptr_t admin_schedule;
    tsn_schedule_config_ptr_t oper_schedule;
    uint64_t time_interval;
    time_stamp_t cycle_start;
    uint64_t tick_length_ns;
    uint64_t slot_end_time_stamp;
    
    uint32_t list_pointer;
    uint8_t current_gate_state;
    
    uint32_t last_filled_bd;
    queue_ptr_t traffic_queues[TSN_MAC_QUEUE_COUNT];
    queue_ptr_t receive_queue;
    

    timer_cfg_t timer_cfg;
} tsn_runtime_config_t;

typedef tsn_runtime_config_t _IODEV * tsn_runtime_config_ptr_t;

///////////////////////////////////////////////////////////////
//High level functions 
///////////////////////////////////////////////////////////////


/**
 * @brief Initialize Ethernet MAC 
 *  
 * This function initializes the MAC Address and the Receive Buffer,
 * Configures the Ethernet MAC
 * 
 * 
 */
mmio_ptr_t tsn_mac_initialize(device_config_t *config);


network_error_t tsn_mac_update_gcl(tsn_gcl_config_t *gcl_config, mmio_ptr_t config_ptr);

/**
 * @brief Sends a Ethernet Frame 
 *  
 * This function requests a buffer from the Ethernet MAC, transfers
 * data to the Ethernet MAC. If there is no buffer available or no 
 * space left inside the transmission queue, the function returns 
 * -1. If there is no frame waiting for transmission the frame will 
 * be transmitted immidiately and the function blocks until finished. 
 * If there is frame already waiting the frame will be queued and the 
 * function returns -2.
 * 
 */
network_error_t tsn_mac_send(network_frame_t *seg, unsigned long long int timeout, mmio_ptr_t config_ptr);

/**
 * @brief Sends a Ethernet Frame without blocking
 *  
 * This function requests a buffer from the Ethernet MAC, transfers
 * data to the Ethernet MAC. If there is no buffer available or no 
 * space left inside the transmission queue, the function returns 
 * -1. If there is no frame waiting for transmission the frame will 
 * be transmitted immidiately and the function will return without  
 * blocking. If there is frame already waiting the frame will be  
 * queued and the function returns -2.
 * 
 */
network_error_t tsn_mac_send_nb(network_frame_t *seg, mmio_ptr_t config_ptr);
 

/**
 * @brief Update Transmission 
 *  
 * This function is used to send enqueued frames and shall be 
 * called periodically. If a frame was transmitted and the buffer
 * is no longer busy, a buffer entry will be dequeued from the TX
 * Queue and sending will be initiated.   
 *
 */
network_error_t tsn_mac_update_transmission(mmio_ptr_t config_ptr);


/**
 * @brief Receive a Ethernet Frame 
 *  
 * This function waits until receiving a ethernet frame. A timeout
 * can be provided how long this function shall wait, use 0 to wait
 * indefinately. When the timeout is exceeded the function returns -1
 * When a frame is received, it is checked if the frame will fit
 * in the provided buffer, if the buffer is not large enough -2 will
 * be return from the function. If a frame was received successfuly
 * the function returns 0. 
 * 
 */
network_error_t tsn_mac_receive(uint8_t *rx_buffer, uint16_t buffer_size, unsigned long long int timeout, mmio_ptr_t config_ptr);

/**
 * @brief Receive a Ethernet Frame 
 *  
 * This function checks if a frame has been received by the
 * Ethernet MAC. If a frame is waiting the data will be copied
 * to the provided buffer and the function returns 0. If the
 * provided Buffer does not have sufficient size the function 
 * returns -2. If there is no frame waiting for reception the
 * function returns -1.
 * 
 */
network_error_t tsn_mac_receive_nb(uint8_t *rx_buffer, uint16_t buffer_size, time_stamp_t *rx_ts, mmio_ptr_t config_ptr);

/**
 * @brief Update Reception 
 *  
 * This function is used to enqueued received frames from the
 * ethernet frame and shall be called periodically to prevent
 * dropping of frames. If a frame was received the function
 * adds it to the RX queue. A new buffer will be requested and
 * set as the receive buffer.
 *
 */
network_error_t tsn_mac_update_reception(mmio_ptr_t config_ptr);

/**
 * @brief Check for Ethernet Frame 
 *  
 * This function checks whether the Ethernet has received a frame which
 * is ready to be copied to the internal Memory. If a frame was received
 * by the Ethernet MAC, the function returns the length of the frame, 
 * otherwise 0.
 * 
 * @return unsigned short recv: 1 if a frame has be received, 0 if not
 */
uint16_t tsn_mac_has_frame(mmio_ptr_t config_ptr);

///////////////////////////////////////////////////////////////
//  Configuration functions
///////////////////////////////////////////////////////////////

/**
 * @brief Enables Transmission
 *  
 * This function enables transmission of ethernet frames.
 */
void tsn_mac_enable_tx(mmio_ptr_t config_ptr);

/**
 * @brief Disables Transmission
 *  
 * This function disable transmission of ethernet frames.
 */
void tsn_mac_disable_tx(mmio_ptr_t config_ptr);

/**
 * @brief Transmission is enabled
 *  
 * This function returns whether transmission is enabled or
 * not.
 * 
 * @return char enabled     : returns 1 if transmission is 
 *                            enabled, 0 otherwise.
 */
char tsn_mac_tx_enabled(mmio_ptr_t config_ptr);

/**
 * @brief Enables Reception
 *  
 * This function enables reception of ethernet frames.
 */
void tsn_mac_enable_rx(mmio_ptr_t config_ptr);

/**
 * @brief Disables Reception
 *  
 * This function disable reception of ethernet frames.
 */
void tsn_mac_disable_rx(mmio_ptr_t config_ptr);

/**
 * @brief Reception is enabled
 *  
 * This function returns whether reception is enabled or
 * not.
 * 
 * @return char enabled     : returns 1 if reception is 
 *                            enabled, 0 otherwise.
 */
char tsn_mac_rx_enabled(mmio_ptr_t config_ptr);



/**
 * @brief Sets the MAC Address 
 *  
 * This function is used to set a the Ethernet MACs MAC-Address.
 * 
 * @param unsigned char *mac_addr   : Buffer which holds the MAC
 *                                    Address.
 */
void tsn_mac_set_mac_address(unsigned char *mac_addr, mmio_ptr_t config_ptr);

/**
 * @brief Gets the MAC Address 
 *  
 * This function reads out the MAC-Address from the Ethernet MAC.
 * 
 * @param unsigned char *mac_addr   : Buffer where the MAC
 *                                    Address will be stored.
 */
void tsn_mac_get_mac_address(unsigned char *mac_addr, mmio_ptr_t config_ptr);

void tsn_mac_print_queue_states(mmio_ptr_t config_ptr);

network_error_t tsn_mac_get_netw_clk_timestamp(mmio_ptr_t config_ptr, time_stamp_t *ts);

network_error_t tsn_mac_get_netw_clk_rx_timestamp(mmio_ptr_t config_ptr, time_stamp_t *ts);

network_error_t tsn_mac_get_netw_clk_tx_timestamp(mmio_ptr_t config_ptr, time_stamp_t *ts);

network_error_t tsn_mac_netw_clk_correct_offset(mmio_ptr_t config_ptr, time_interval_t *offset);

network_error_t tsn_mac_set_netw_clk_timestamp(mmio_ptr_t config_ptr, time_stamp_t *time);

#endif 
#endif