/*
 * IO functions to read froma and write to the Etherent MAC 
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 *          Christoph Lehr
 */

#include <stdio.h>
#include "../../inc/generic_types.h"
#include "mmio_utils.h"

// Write a byte in rx-tx buffer
void mmio_wr_8(mmio_ptr_t address, uint16_t offset ,uint8_t data) {
    mmio_ptr_8_t ptr = (mmio_ptr_8_t) (address+offset);
    *ptr = data;
    return;
}

// Write a byte in rx-tx buffer
uint8_t mmio_rd_8(mmio_ptr_t address, uint16_t offset) {
    mmio_ptr_8_t ptr = (mmio_ptr_8_t) (address+offset);
    return *ptr;
}

// Write a byte in rx-tx buffer
void mmio_wr_16(mmio_ptr_t address, uint16_t offset ,uint8_t data) {
    mmio_ptr_16_t ptr = (mmio_ptr_16_t) (address+offset);
    *ptr = data;
    return;
}

// Write a byte in rx-tx buffer
uint16_t mmio_rd_16(mmio_ptr_t address, uint16_t offset) {
    mmio_ptr_16_t ptr = (mmio_ptr_16_t) (address+offset);
    return *ptr;
}


// Write to ethernet controller
void mmio_wr_32(mmio_ptr_t address, uint16_t offset , uint32_t data) {
    //mmio_ptr_32_t ptr = (mmio_ptr_32_t) (address+offset);
    //*ptr= data;
    *((mmio_ptr_32_t) (address+offset))= data;
    return;
}      

// Read ethernet controller
uint32_t mmio_rd_32(mmio_ptr_t address, uint16_t offset) {
    mmio_ptr_32_t ptr = (mmio_ptr_32_t) (address+offset);
    return *ptr;
}      


// Write a block of data to the ethernet buffer
void mmio_wr_block(mmio_ptr_t address, uint8_t *data, uint16_t len) 
{    
    uint64_t start_address = (uint64_t) address;
    uint64_t end_address = (uint64_t)(address+len);
    unsigned int write_burst_len = len;
    write_burst_len -= end_address& (sizeof(mmio_ptr_t)-1);
    write_burst_len -= start_address& (sizeof(mmio_ptr_t)-1);
    unsigned int data_offset=0;
    
    // align start address
    if((start_address& (sizeof(mmio_ptr_t)-1)) !=0)
    {
        data_offset += start_address& (sizeof(mmio_ptr_t)-1);
        #pragma loopbound min 1 max 3
        for(uint8_t i=0; i < (start_address&0x3); i++)
        {
            mmio_wr_8(address, i, data[i]);
        }
    }

    // write to memory in word bursts
    // 379.5 = 1520 Bytes (Max MTU) / 4 (Bytes per Word)
    // The function will be executed twice for each frame
    // one time for the headers, one time for the data 
    // therefore the total cycles get halfed :380/2 = 190
    #pragma loopbound min 3 max MMIO_WCET_MAX_WRITE_DATA_SIZE
    for(uint16_t i=0; i < write_burst_len ; i+=4)
    {
        // correct byte ordering, I don't know why but the bytes have to be read out like this
        unsigned int word = data[data_offset+i] << 24 | data[data_offset+i+1] << 16 |data[data_offset+i+2] << 8 | data[data_offset+i+3];  
        mmio_wr_32(address, data_offset+i, word);
    }
    
    // write trailing bytes
    if(((end_address& (sizeof(mmio_ptr_t)-1)))!=0)
    {
        data_offset += write_burst_len;
        #pragma loopbound min 1 max 3
        for(unsigned i=0; i < end_address%4; i++)
        {
            mmio_wr_8(address, data_offset+i, data[data_offset+i]);
        }
    }
    
    return;
}


// read a block of data from the ethernet controller buffer
void mmio_rd_block(mmio_ptr_t address, uint8_t *data, uint16_t len) 
{
 
    uint64_t start_address = (uint64_t) address;
    uint64_t end_address = (uint64_t)(address+len);
    unsigned int read_burst_len = len;
    read_burst_len -= end_address& (sizeof(mmio_ptr_t)-1);
    read_burst_len -= start_address& (sizeof(mmio_ptr_t)-1);
    unsigned int data_offset=0;
    
    // align start address
    if((start_address& (sizeof(mmio_ptr_t)-1)) !=0)
    {        
        data_offset += start_address& (sizeof(mmio_ptr_t)-1);
        #pragma loopbound min 1 max 3
        for(uint8_t i=0; i < (start_address&0x3); i++)
        {
            data[i] = mmio_rd_8(address,i);
        }
    }
    
    uint32_t *word_data = (uint32_t *)(&data[data_offset]);

    // read from memory in word bursts
    // Maximum length of an Ethernet frame is 1518 Bytes
    // 379.5 = 1500 Bytes (Max MTU) / 4 (Bytes per Word)
    #pragma loopbound min 3 max MMIO_WCET_MAX_READ_DATA_SIZE 
    for(uint16_t i=0; i < read_burst_len ; i+=4)
    {
        word_data[i>>2] = mmio_rd_32(address, data_offset+i);
    }

    // write trailing bytes
    if((end_address& (sizeof(mmio_ptr_t)-1)) !=0)
    {
        data_offset += read_burst_len;
        #pragma loopbound min 1 max 3
        for(unsigned i=0; i < end_address%4; i++)
        {
            data[data_offset+i] = mmio_rd_8(address, data_offset+i);
        }
    }

    return;
}
