/*
 * Central Network Interface Management Module 
 * 
 * Authors: Lehr Christoph
 */

#include "stdio.h"
#include "stdlib.h"
#include "string.h"

#include "../inc/net_types.h"
#include "../inc/network.h"
#include "../inc/socket.h"
#include "../inc/utils.h"

#include "driver/ethernet/patmos_eth_mac.h"
#include "driver/ethernet/patmos_tsn_mac.h"
#include "protocols/arp.h"
#include "protocols/tcp.h"

typedef struct
{
    uint8_t debug_output:1;
} network_interface_aux_options_t;

typedef struct 
{
    mmio_ptr_t device_runtime_config;
    void* auxiliary_layer2_config;              // e.g. tsn config, vlan
    network_interface_types_t interface_type;   // what kind of interface is connected
    network_address_t address;                  // sets layer 3 addresses
    network_layer2_config_t layer2_config;
    network_interface_aux_options_t options;
    char if_name[MAX_NETWORK_IF_NAME_LENGTH]; 
} runtime_network_interface_config_t;


static runtime_network_interface_config_t config_array[MAX_NETWORK_IF_CONFIGS];
static volatile uint16_t config_wr_ptr=0;
static volatile uint8_t global_logging=0;
static volatile uint16_t buffer_mgmt_id=0xFFFF;

static network_error_t patmos_network_interface_get_message(uint16_t interface_id);

void network_init(buffer_management_config_t *buffer_mgmt_cfg)
{
    for(int i = 0; i < MAX_NETWORK_IF_CONFIGS; i++) {
        config_array[i].interface_type = NETWORK_INTERFACE_NONE;
    }
    arp_init();
    buffer_mgmt_id = buffer_management_register(buffer_mgmt_cfg);
}

uint16_t network_register_interface(network_interface_config_t* config)
{
    // find next free entry
    char found =0;
    uint16_t if_id=0xFFFF;
    #pragma loopbound min 1 max MAX_NETWORK_IF_CONFIGS 
    for(uint16_t i=config_wr_ptr; i < MAX_NETWORK_IF_CONFIGS; i++)
    {
        if(config_array[i].interface_type == NETWORK_INTERFACE_NONE)
        {
            config_wr_ptr = i;
            if_id =i;
            found =1;
            break;
        }
    }
    // check if buffer was found
    if(found == 0 )
    {
        #pragma loopbound min 0 max MAX_NETWORK_IF_CONFIGS 
        for(uint16_t i=0; i < config_wr_ptr; i++)
        {
            if(config_array[i].interface_type == NETWORK_INTERFACE_NONE)
            {
                config_wr_ptr = i;
                if_id =i;
                found =1;
                break;
            }
        }   
        // check if free index was found
        if(found == 0)
        {
            if(global_logging)
            {
                utils_log("No free index found for new network interface!\n");
            }
            return ~0;
        }
    }

    if(global_logging)
    {
        utils_log("Found free index for new network interface: %u\n", config_wr_ptr);
    }

    mmio_ptr_t ret_config_ptr = NULL;

    // Init Interface
#ifdef ARCH_PATMOS
    if(config->interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET)
    {   
        if(global_logging)
        {
            utils_log("Initializing patmos ethernet MAC\n");
        }
        ret_config_ptr= eth_mac_initialize(&config->device_config);
        if(global_logging)
        {
            utils_log(" Configuration stored at %p\n", ret_config_ptr);
        }
    }
    else if(config->interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
    {   
        if(global_logging)
        {
            utils_log("Initializing patmos TSN MAC\n");
        }
        ret_config_ptr= tsn_mac_initialize(&config->device_config);
        if(global_logging)
        {
            utils_log(" Configuration stored at %p\n", ret_config_ptr);
        }
    }
    else
#endif
    {
        if(global_logging)
        {
            utils_log("No matching Interface Type\n");
        }
        return  ~0;
    }
    
    if(ret_config_ptr != NULL)
    {
        config_array[if_id].device_runtime_config = ret_config_ptr;
    }
    else
    {
        if(global_logging)
        {
            utils_log("Invalid Interface config was detected\n");
        }
        return  ~0;
    }


    char *if_str = (char *) &config_array[if_id].if_name;
    strncpy(if_str, config->interface_name, MAX_NETWORK_IF_NAME_LENGTH);


    if(global_logging)
    {
        utils_log("Registering interface with name %s\n", if_str);
    }

    config_array[if_id].interface_type = config->interface_type;
    config_array[if_id].address = config->address;
    config_array[if_id].layer2_config = config->layer2_config;

    return if_id;
}

uint16_t network_duplicate_interface(uint16_t interface_id, network_interface_config_t* config)
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {

        if(global_logging)
        {
            utils_log("Invalid Interface was provided: %u\n", interface_id);
        }

        return ~0;
    }

    // find next free entry
    char found =0;
    #pragma loopbound min 1 max MAX_NETWORK_IF_CONFIGS 
    for(uint16_t i=config_wr_ptr; i < MAX_NETWORK_IF_CONFIGS; i++)
    {
        if(config_array[config_wr_ptr].device_runtime_config == NULL)
        {
            config_wr_ptr = i;
            found =1;
        }
    }
    // check if buffer was found
    if(found == 0 )
    {
        #pragma loopbound min 0 max MAX_NETWORK_IF_CONFIGS 
        for(uint16_t i=0; i < config_wr_ptr; i++)
        {
            if(config_array[config_wr_ptr].device_runtime_config == NULL)
            {
                config_wr_ptr = i;
                found =1;
            }
        }   
        // check if free index was found
        if(found == 0)
        {
            if(global_logging)
            {
                utils_log("No free index found for new network interface!\n");
            }
            return 0xFFFF;
        }
    }

    if(global_logging)
    {
        utils_log("Found free index for new network interface: %u\n", config_wr_ptr);
    }

    // copy interface data
    config_array[config_wr_ptr] = config_array[interface_id];
    // set new interface name and new network address

    char *if_str = (char *) &config_array[config_wr_ptr].if_name;
    strncpy(if_str, config->interface_name, MAX_NETWORK_IF_NAME_LENGTH);

    if(global_logging)
    {
        utils_log("Registering interface with name %s\n", if_str);
    }

    config_array[config_wr_ptr].address = config->address;

    return config_wr_ptr;
}

network_error_t network_set_option(uint16_t interface_id, network_options_t option, void* value, uint32_t value_size)
{
    network_error_t ret= NET_ERR_OK;
    if(interface_id == 0xFFFF)
    {
        if(option == NETWORK_OPTION_LOGGING)
        {
            uint8_t* val = (uint8_t*) value;
            global_logging = *val;
            return NET_ERR_OK;
        }

        return NET_ERR_INPUT_VALUE;
    }

    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }

    if(option == NETWORK_OPTION_TSN_CFG)
    {
        if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
        {
            tsn_gcl_config_t *tsn_conf = (tsn_gcl_config_t *) value;
            ret= tsn_mac_update_gcl(tsn_conf, config_array[interface_id].device_runtime_config);
        }
        else
        {
            ret=  NET_ERR_INPUT_VALUE;
        }
    }
    else if(option == NETWORK_OPTION_LOGGING)
    {
        uint8_t* val = (uint8_t*) value;
        config_array[interface_id].options.debug_output = *val;
    }
    else if(option == NETWORK_OPTION_PRINT_DRIVER_DBG_INFO)
    {
        if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
        {
            tsn_mac_print_queue_states(config_array[interface_id].device_runtime_config);
        }
        else
        {
            ret=  NET_ERR_NOT_SUPPORTED;
        }
    }
    else if(option == NETWORK_OPTION_SET_PTP_CLOCK_OFFSET) 
    {
        if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
        {
            ret= tsn_mac_netw_clk_correct_offset(config_array[interface_id].device_runtime_config,
                (time_interval_t*) value);
        }
        else
        {
            ret=  NET_ERR_NOT_SUPPORTED;
        }
    }
    else if(option == NETWORK_OPTION_SET_PTP_CLOCK)
    {
        if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
        {
            ret= tsn_mac_set_netw_clk_timestamp(config_array[interface_id].device_runtime_config, (time_stamp_t*) value);
        }
        else
        {
            ret=  NET_ERR_NOT_SUPPORTED;
        }
    }
    else
    {
        ret=  NET_ERR_INPUT_VALUE;
    }

    return ret;
}

network_error_t network_get_option(uint16_t interface_id, network_options_t option, void* value, uint32_t *value_size)
{
    network_error_t ret= NET_ERR_OK;
    if(interface_id == 0xFFFF)
    {
        if(option == NETWORK_OPTION_LOGGING)
        {
            uint8_t* val = (uint8_t*) value;
            global_logging = *val;
            return NET_ERR_OK;
        }

        return NET_ERR_INPUT_VALUE;
    }

    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }

    if(option == NETWORK_OPTION_GET_PTP_CLOCK)
    {
        if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
        {
            ret= tsn_mac_get_netw_clk_timestamp(config_array[interface_id].device_runtime_config,
                (time_stamp_t*) value);
            *value_size = sizeof(time_stamp_t);
        }
        else
        {
            ret=  NET_ERR_NOT_SUPPORTED;
        }
    }
    else if(option == NETWORK_OPTION_GET_PTP_LAST_RX_TIMESTAMP) 
    {
        if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
        {
            ret= tsn_mac_get_netw_clk_rx_timestamp(config_array[interface_id].device_runtime_config,
                (time_stamp_t*) value);
            *value_size = sizeof(time_stamp_t);
        }
        else
        {
            ret=  NET_ERR_NOT_SUPPORTED;
        }
    }
    else if(option == NETWORK_OPTION_GET_PTP_LAST_TX_TIMESTAMP) 
    {
        if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
        {
            ret= tsn_mac_get_netw_clk_tx_timestamp(config_array[interface_id].device_runtime_config,
                (time_stamp_t*) value);
            *value_size = sizeof(time_stamp_t);
        }
        else
        {
            ret=  NET_ERR_NOT_SUPPORTED;
        }
    }
    else
    {
        ret=  NET_ERR_INPUT_VALUE;
    }

    return ret;
}

void *network_get_buffer(uint16_t size){
    return buffer_management_malloc(size, buffer_mgmt_id);
}

void network_return_buffer(void *buffer)
{
    buffer_management_free(buffer, buffer_mgmt_id);
}


// this function shall be called periodically to keep sending and receiving packets
void network_update(void)
{
    #pragma loopbound min 1 max MAX_NETWORK_IF_CONFIGS 
    for(uint16_t i=0; i < MAX_NETWORK_IF_CONFIGS; i++)
    {
        // if(config_array[i].device_runtime_config != NETWORK_INTERFACE_NONE)
        if(config_array[i].interface_type != NETWORK_INTERFACE_NONE)
        {
            #ifdef ARCH_PATMOS
            if(config_array[i].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET)
            {
                if(global_logging || config_array[i].options.debug_output == NETWORK_OPTION_LOGGING )
                {
                    utils_log("Updating Reception and Transmission of interface %s\n", config_array[i].if_name);
                }
                eth_mac_update_transmission(config_array[i].device_runtime_config);
                eth_mac_update_reception(config_array[i].device_runtime_config);
            }
            else if(config_array[i].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
            {
                if(global_logging || config_array[i].options.debug_output == NETWORK_OPTION_LOGGING )
                {
                    utils_log("Updating Reception and Transmission of interface %s\n", config_array[i].if_name);
                }
                tsn_mac_update_transmission(config_array[i].device_runtime_config);
                tsn_mac_update_reception(config_array[i].device_runtime_config);
            }
            #endif
        }
    }
    // update the server TCP if exist
    tcp_update();
}

void network_read_interfaces(void)
{
    #pragma loopbound min 1 max MAX_NETWORK_IF_CONFIGS 
    for(uint16_t i=0; i < MAX_NETWORK_IF_CONFIGS; i++)
    {
        // if(config_array[i].device_runtime_config != NETWORK_INTERFACE_NONE)
        if(config_array[i].interface_type != NETWORK_INTERFACE_NONE)
        {
            network_read_interface(i);
        }
    }
}

network_error_t network_read_interface(uint16_t interface_id)
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }
   
    if(global_logging || config_array[interface_id].options.debug_output == NETWORK_OPTION_LOGGING )
    {
        utils_log("Reading from interface %s\n", config_array[interface_id].if_name);
    }
    network_error_t ret;
    network_error_t func_ret;

    #pragma loopbound min 1 max WCET_MAXIMUM_QUEUE_LENGTH
    do
    {
        ret = network_interface_get_message(interface_id);
        if(ret != NET_ERR_MGMT_NO_MSG && ret != NET_ERR_DRVR_NO_FRAME_RECEIVED)
        {
            func_ret = ret;
        }
    }while(ret != NET_ERR_MGMT_NO_MSG && ret != NET_ERR_DRVR_NO_FRAME_RECEIVED && ret != NET_ERR_MGMT_MEM_FULL && ret != NET_ERR_INPUT_VALUE);

    return func_ret;
}

network_error_t network_interface_get_message(uint16_t interface_id)
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }
    
    if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET || 
        config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
    {
        return patmos_network_interface_get_message(interface_id);
    }
    
    return NET_ERR_MGMT_UNKNOWN;
}

network_error_t patmos_network_interface_get_message(uint16_t interface_id)
{
    #ifdef ARCH_PATMOS  

    uint16_t frame_size =0;
    if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET)
    {
        frame_size = eth_mac_has_frame(config_array[interface_id].device_runtime_config);
    }
    else if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
    {
        frame_size = tsn_mac_has_frame(config_array[interface_id].device_runtime_config);
    }
    else
    {
        // unknown interface type
        return NET_ERR_INPUT_VALUE;
    }
     
    if(frame_size > ETHERNET_MAX_FRAME_SIZE || frame_size ==0 )
    {
        return NET_ERR_MGMT_NO_MSG;
    }

    if(global_logging || config_array[interface_id].options.debug_output == NETWORK_OPTION_LOGGING )
    {
        utils_log("Interface %s received a frame of length %u \n", config_array[interface_id].if_name, frame_size);
    }
    network_error_t ret = NET_ERR_OK;

    uint8_t *rx_ptr = (uint8_t*) buffer_management_malloc(frame_size, buffer_mgmt_id);
    
    if(rx_ptr == NULL)
    {

#ifndef WCET_ANALYSIS 
        buffer_management_memory_dump(buffer_mgmt_id);
#endif

        if(global_logging || config_array[interface_id].options.debug_output == NETWORK_OPTION_LOGGING )
        {
            utils_log("No free memory left in network stack\n");
        }
        return NET_ERR_MGMT_MEM_FULL;
    }
    
    time_stamp_t rx_ts;
    if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET)
    {
        ret = eth_mac_receive_nb(rx_ptr, frame_size, config_array[interface_id].device_runtime_config);
    }
    else if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
    {
        ret = tsn_mac_receive_nb(rx_ptr, frame_size, &rx_ts, config_array[interface_id].device_runtime_config);
    }
    else
    {
        buffer_management_free(rx_ptr, buffer_mgmt_id);
        // unknown interface type
        return NET_ERR_INPUT_VALUE;
    }
    
    if(ret != NET_ERR_OK)
    {
        if(global_logging || config_array[interface_id].options.debug_output == NETWORK_OPTION_LOGGING )
        {
            utils_log("Error 0x%04hx while getting frame of interface %s\n", ret, config_array[interface_id].if_name);
        }
        buffer_management_free(rx_ptr, buffer_mgmt_id);
    }
    else
    {
        if(global_logging || config_array[interface_id].options.debug_output == NETWORK_OPTION_LOGGING )
        {
            utils_log("Handing over frame of interface %s to socket\n", config_array[interface_id].if_name);
        }
        
        ret = socket_handle_frame(interface_id, rx_ptr, frame_size, &rx_ts);
        if(ret != NET_ERR_OK)
        {
            if(global_logging || config_array[interface_id].options.debug_output == NETWORK_OPTION_LOGGING )
            {
                utils_log("Error 0x%04hx while parsing frame of interface %s\n", ret, config_array[interface_id].if_name);
            }
            buffer_management_free(rx_ptr, buffer_mgmt_id);
        } 
        
    }
                
    return ret;
    #else
    // If this runs not on the patmos platform this function shall always return that no msg has been received
    return NET_ERR_MGMT_NO_MSG;
    #endif
}

uint16_t network_get_interface_by_name(char *name)
{
    #pragma loopbound min 1 max MAX_NETWORK_IF_CONFIGS 
    for (uint16_t i = 0; i < MAX_NETWORK_IF_CONFIGS; i++)
    {
        if(config_array[i].interface_type != NETWORK_INTERFACE_NONE)
        {
            char *if_name = (char *)&config_array[config_wr_ptr].if_name;
            if(strcmp(if_name, name ) == 0)
            {
                return i;
            }
        }
    }
    return ~0;
}

uint16_t network_get_interface_by_ip_address(network_address_t *address)
{

    #pragma loopbound min 1 max MAX_NETWORK_IF_CONFIGS 
    for (uint16_t i = 0; i < MAX_NETWORK_IF_CONFIGS; i++)
    {
        if(config_array[i].interface_type != NETWORK_INTERFACE_NONE)
        {
            if(address->type == config_array[i].address.type && 
                address->ip.ip_address.version == config_array[i].address.ip.ip_address.version)
            {
                uint8_t equal = utils_compare_ip(&address->ip.ip_address, &config_array[i].address.ip.ip_address);
                
                if(equal == 0)
                {
                    return i;
                }
            }
        }
    }
    
    return ~0;
}

uint16_t network_get_interface_by_ip_netmask(ip_address_t *address)
{
    #pragma loopbound min 1 max MAX_NETWORK_IF_CONFIGS 
    for (uint16_t i = 0; i < MAX_NETWORK_IF_CONFIGS; i++)
    {
        if(config_array[i].interface_type != NETWORK_INTERFACE_NONE)
        {
            if(NETWORK_ADDRESS_IP == config_array[i].address.type && 
                address->version == config_array[i].address.ip.ip_address.version)
            {
                
                uint8_t equal = utils_compare_ip_netmask(address, &config_array[i].address.ip.ip_address, config_array[i].address.ip.netmask);
                
                if(equal == 0)
                {
                    return i;
                }                
            }
        }
    }
    
    return ~0;
}

network_error_t network_set_interface_address(uint16_t interface_id, network_address_t *address)
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }

    config_array[config_wr_ptr].address = *address;

    return NET_ERR_OK;
}

network_error_t network_get_interface_address(uint16_t interface_id, network_address_t* addr)
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }
    *addr = config_array[interface_id].address;
    return NET_ERR_OK;
}

network_error_t network_get_interface_ip_address(uint16_t interface_id, ip_address_t* addr)
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }
    if(config_array[interface_id].address.type != NETWORK_ADDRESS_IP)
    {
        return NET_ERR_ADDR_INCOMPATIBLE;
    }
    *addr = config_array[interface_id].address.ip.ip_address;
    return NET_ERR_OK;
}

uint16_t network_get_interface_ip_netmask(uint16_t interface_id)
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return 0;
    }
    if(config_array[interface_id].address.type != NETWORK_ADDRESS_IP)
    {
        return 0;
    }
    
    return config_array[interface_id].address.ip.netmask;
}

network_error_t network_get_mac_address(uint16_t interface_id, mac_address_t* mac_addr)
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }

    #ifdef ARCH_PATMOS
    if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET)
    {   
        mac_addr->length=6;
        eth_mac_get_mac_address(mac_addr->address , config_array[interface_id].device_runtime_config);
    }
    else if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
    {
        mac_addr->length=6;
        tsn_mac_get_mac_address(mac_addr->address , config_array[interface_id].device_runtime_config);
    }
    #endif

    return NET_ERR_OK;
}

network_error_t network_get_l2_config(uint16_t interface_id, network_layer2_config_t* config)
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }

    *config = config_array[interface_id].layer2_config;

    return NET_ERR_OK;
}


network_error_t network_interface_exists(uint16_t interface_id)
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }
    return NET_ERR_OK;
}

network_error_t network_send(uint16_t interface_id, network_frame_t *frame, uint8_t blocking )
{
    if(interface_id >= MAX_NETWORK_IF_CONFIGS ||
       config_array[interface_id].interface_type == NETWORK_INTERFACE_NONE)
    {
        return NET_ERR_MGMT_UNKNOWN;
    }
    network_error_t ret;
    
    #ifdef ARCH_PATMOS
    if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET)
    {
        if(global_logging || config_array[interface_id].options.debug_output == NETWORK_OPTION_LOGGING )
        {
            utils_log("Sending frame with header lenght %u and data length %u from interface %s\n", 
                frame->header_length, frame->user_data_length, config_array[interface_id].if_name);
        }

        if(blocking)
        {
            ret = eth_mac_send(frame , 0, config_array[interface_id].device_runtime_config);
        }
        else
        {
            ret = eth_mac_send_nb(frame , config_array[interface_id].device_runtime_config);
        }
    }
    else if(config_array[interface_id].interface_type == NETWORK_INTERFACE_PATMOS_ETHERNET_TSN)
    {
        if(global_logging || config_array[interface_id].options.debug_output == NETWORK_OPTION_LOGGING )
        {
            utils_log("Sending frame with header lenght %u and data length %u from interface %s\n", 
                frame->header_length, frame->user_data_length, config_array[interface_id].if_name);
        }

        if(blocking)
        {
            ret = tsn_mac_send(frame , 0, config_array[interface_id].device_runtime_config);
        }
        else
        {
            ret = tsn_mac_send_nb(frame , config_array[interface_id].device_runtime_config);
        }
    }
    else
    #endif
    {
        ret = NET_ERR_NOT_SUPPORTED;
    }
    return ret;
}