/*
 * ARP section 
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 * 			Christoph Lehr
 */

#include "arp.h"
#include <stdio.h>
#include "../../inc/net_types.h"
#include "../../inc/generic_types.h"
#include "../../inc/utils.h"

///////////////////////////////////////////////////////////////
//Functions related to the ARP table
///////////////////////////////////////////////////////////////
struct arp_table_entry{
	unsigned char used;//0 if empty 
	ip_address_t ip_addr;
	mac_address_t mac_addr;
};

struct arp_table_entry arp_table[ARP_TABLE_SIZE];

//This function remove ARP table entries for the given IP. If something is removed it returns 1. If not it returns 0.
network_error_t arp_table_delete_entry(ip_address_t *ip_addr);

//This function initilize the ARP table.
void arp_init(){
	for (int i=0; i<ARP_TABLE_SIZE; i++){
		arp_table[i].used = 0;
	}
	return;
}

//This function searches in the ARP table for the given IP. If it exists it returns 1 and the MAC. If not it returns 0.
network_error_t arp_table_search(ip_address_t *ip_addr, mac_address_t *mac_addr){
	
	if(ip_addr->version != 4)
	{
		return NET_ERR_IP_VERSION;
	}

	#pragma loopbound min 1 max ARP_TABLE_SIZE
	for (int i=0; i<ARP_TABLE_SIZE; i++){
		if (arp_table[i].used == 1 && 
			utils_compare_ip(&arp_table[i].ip_addr, ip_addr) ==0
		){
			
			*mac_addr = arp_table[i].mac_addr;

			return NET_ERR_OK;
		}
	}
	mac_addr->length = 0;
	#pragma loopbound min 6 max 6
	for(uint8_t i=0; i< 6; i++)
	{
		mac_addr->address[i] = 0xFF;
	}
	return NET_ERR_ARP_UNKNOWN_ADDR;
}

//This function remove ARP table entries for the given IP. If something is removed it returns 1. If not it returns 0.
network_error_t arp_table_replace_entry(ip_address_t *ip_addr, mac_address_t *mac_addr)
{
	#pragma loopbound min 1 max ARP_TABLE_SIZE
	for (int i=0; i<ARP_TABLE_SIZE; i++){
		if (arp_table[i].used == 1 && 
			utils_compare_ip(&arp_table[i].ip_addr, ip_addr) ==0
		){
			arp_table[i].mac_addr = *mac_addr ;
			return NET_ERR_OK;
		}
	}
	return NET_ERR_ARP_UNKNOWN_ADDR;
}

//This function remove ARP table entries for the given IP. If something is removed it returns 1. If not it returns 0.
network_error_t arp_table_delete_entry(ip_address_t *ip_addr){
	network_error_t ans = NET_ERR_ARP_UNKNOWN_ADDR;	
	#pragma loopbound min 1 max ARP_TABLE_SIZE
	for (int i=0; i<ARP_TABLE_SIZE; i++){
		if (arp_table[i].used == 1 && 
			utils_compare_ip(&arp_table[i].ip_addr, ip_addr) ==0
		){
			arp_table[i].used=0;
			ans = NET_ERR_OK;
		}
	}
	return ans;
}

//This function insert a new entry in the ARP IP/MAC table. If an entry was already there the fields are updated. If we are out of space, the first entry is replaced.
network_error_t arp_table_new_entry(ip_address_t *ip_addr, mac_address_t *mac_addr){
	mac_address_t previous_mac;
	int i = 0;
	if(arp_table_search(ip_addr, &previous_mac) == NET_ERR_OK){
		return arp_table_replace_entry(ip_addr, mac_addr);
	}
	
	#pragma loopbound min 1 max ARP_TABLE_SIZE
	while(arp_table[i].used == 1 && i<ARP_TABLE_SIZE){
		i++;
	}

	if (i == ARP_TABLE_SIZE){
		i = 0;//table is full
		return NET_ERR_ARP_FULL;
	}

	arp_table[i].used = 1;
	arp_table[i].ip_addr = *ip_addr;
	arp_table[i].mac_addr = *mac_addr;

	return NET_ERR_OK;
}

//This ugly function prints the ARP table for debug purposes.
void arp_table_print(){
	utils_log("ARP IP/MAC table\n#\tUsed\tIP\t\tMAC\n");
	for (int i=0; i<ARP_TABLE_SIZE; i++){
		utils_log("%d\t%d\t%d.%d.%d.%d\t", i, arp_table[i].used, arp_table[i].ip_addr.ipv4_addr[0], arp_table[i].ip_addr.ipv4_addr[1], arp_table[i].ip_addr.ipv4_addr[2], arp_table[i].ip_addr.ipv4_addr[3]);
		utils_log("%02X:%02X:%02X:%02X:%02X:%02X\n", arp_table[i].mac_addr.address[0], arp_table[i].mac_addr.address[1], arp_table[i].mac_addr.address[2], arp_table[i].mac_addr.address[3], arp_table[i].mac_addr.address[4], arp_table[i].mac_addr.address[5]);
	}
	return;
}


///////////////////////////////////////////////////////////////
//Support functions related to the ARP protocol
///////////////////////////////////////////////////////////////

//This function process a received ARP package. If it is a request and we are the destination (IP) it reply the ARP request and returns 1. If it is a reply and we are the destination (IP) it inserts an entry in the ARP table and returns 2. Otherwise it returns 0.
network_error_t arp_handle_message(uint8_t *buffer, uint16_t size, ip_address_t* host_ip, mac_address_t* host_mac, arp_msg_t *arp_response)
{

	arp_msg_t *msg = (arp_msg_t *) buffer;
	
	// check if size of buffer is correct
	if(size < sizeof(arp_msg_t))
	{
		return NET_ERR_BUFF_UNDERSIZED;
	}

	// always add address to arp table for reply and response
	ip_address_t spa;
	mac_address_t sha;
	spa.version = 4;
	sha.length = 6;
	spa.ipv4_addr[0] = msg->spa[0];
	spa.ipv4_addr[1] = msg->spa[1];
	spa.ipv4_addr[2] = msg->spa[2];
	spa.ipv4_addr[3] = msg->spa[3];
	sha.address[0] = msg->sha[0];
	sha.address[1] = msg->sha[1];
	sha.address[2] = msg->sha[2];
	sha.address[3] = msg->sha[3];
	sha.address[4] = msg->sha[4];
	sha.address[5] = msg->sha[5];

	arp_table_new_entry(&spa, &sha);

	// build request to reply
	if(msg->oper == ARP_OPERATION_REQUEST)
	{	

		ip_address_t tpa;
		tpa.version = 4;
		tpa.ipv4_addr[0] = msg->tpa[0];
		tpa.ipv4_addr[1] = msg->tpa[1];
		tpa.ipv4_addr[2] = msg->tpa[2];
		tpa.ipv4_addr[3] = msg->tpa[3];

		if(utils_compare_ip(host_ip, &tpa) != 0)
		{
			return NET_ERR_ARP_WRONG_PADDR;
		}

		// Build Response
		arp_response->htype = 1; 	// 1 is for ethernet
		arp_response->ptype = ETHERTYPE_IPv4;
		arp_response->hlen = 6;	// length of MAC Address
		arp_response->plen = 4;	// length of IP Address
		arp_response->oper = ARP_OPERATION_RESPONSE;

		//MAC addrs
		#pragma loopbound min 6 max 6
		for (int i=0; i<6; i++){
			arp_response->sha[i] = host_mac->address[i];
			arp_response->tha[i] = msg->sha[i];
		}
		//IP addrs
		#pragma loopbound min 4 max 4
		for (int i=0; i<4; i++){
			arp_response->tpa[i] = msg->spa[i];
			arp_response->spa[i] = host_ip->ipv4_addr[i];
		}

		return NET_ERR_ARP_RESPONSE;
	}

	return NET_ERR_OK;
}

//This function builds an ARP request packet starting at tx_addr and targeting the target_ip
network_error_t arp_build_request(ip_address_t *target_ip, ip_address_t* host_ip, mac_address_t* host_mac, arp_msg_t * request)
{
	if(target_ip->version != IP_VERSION_4)
	{
		return NET_ERR_ADDR_INCOMPATIBLE;
	}

	unsigned char dest_mac[6] = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};

	request->htype = 1; 	// 1 is for ethernet
	request->ptype = ETHERTYPE_IPv4;
	request->hlen = 6;	// length of MAC Address
	request->plen = 4;	// length of IP Address
	request->oper = ARP_OPERATION_REQUEST;

	//MAC address
	#pragma loopbound min 6 max 6
	for (int i=0; i<6; i++){
		request->sha[i] = host_mac->address[i];
		request->tha[i] = dest_mac[i];
	}
	//IP address
	#pragma loopbound min 4 max 4
	for (int i=0; i<4; i++){
		request->spa[i] = host_ip->ipv4_addr[i];
		request->tpa[i] = target_ip->ipv4_addr[i];
	}  

	return NET_ERR_OK;
}

