/*
 * IPv4 section of ethlib (ethernet library)
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 * 			Christoph Lehr
 */

#include <stdio.h>
#include "ipv4.h"
#include "../../inc/net_types.h"
#include "../../inc/generic_types.h"
#include "../../inc/utils.h"
#include "arp.h"
#include "ethernet.h"

static unsigned short identification = IPv4_IDENTIFICATION_SEED;

//Write Mac Header to Buffer, returns length of written header length. Header offset is the offset to tx address
network_error_t ipv4_build_datagram(ip_address_t *dst_ip, ip_address_t *src_ip, ip_protocol_t protocol, uint8_t *buffer, uint16_t len)
{	
	if(dst_ip->version != 4 || src_ip->version != 4)
	{
		return NET_ERR_IP_VERSION;
	}

	ipv4_header_t *ip_hdr = (ipv4_header_t * )buffer;
	
	ip_hdr->version = 4;
	ip_hdr->ihl = 5; // header consists of 5 32 Bit words	
	ip_hdr->dscp =0;	// only relevant for VoIP
	ip_hdr->ecn	=0;	// Explicit congestion notification not implemented
	ip_hdr->total_length = len + sizeof(ipv4_header_t);	// IP Header is 20 Bytes long
	identification += IPv4_IDENTIFICATION_INC;	// only really needed for fragementation
	ip_hdr->identification = identification;
	ip_hdr->flags = 0b010;	// Set dont fragment bit
	ip_hdr->fragment_offset = 0;	// No Fragmentation no offset
	ip_hdr->ttl = 0xFF;
	ip_hdr->protocol = protocol;
	#pragma loopbound min 4 max 4
	for(unsigned char i=0; i<4; i++)
	{
		ip_hdr->destination_ip[i] = dst_ip->ipv4_addr[i];
		ip_hdr->source_ip[i] = src_ip->ipv4_addr[i];
	}

	ip_hdr->header_checksum = ipv4_compute_checksum(ip_hdr);
	
	// convert header to network byte order
	#pragma loopbound min 5 max 5
	for(uint8_t i=0; i<5; i++)
	{
		ip_hdr->raw[i] =utils_htonl(ip_hdr->raw[i]);
	}

	return NET_ERR_OK;
}

network_error_t ipv4_handle_packet(uint8_t*buffer, uint16_t size, ip_protocol_t *protocol, 
	ip_address_t *src_ip, ip_address_t *dst_ip, uint16_t *payload_length)
{
	// convert buffer to ethernet frame header
	ipv4_header_t *ip_hdr = (ipv4_header_t *) buffer; 

	// convert header from network byte order to host byte order
	#pragma loopbound min 5 max 5
	for(uint8_t i=0; i<5; i++)
	{
		ip_hdr->raw[i] =utils_ntohl(ip_hdr->raw[i]);
	}

	// check whether udp packet length is set correctly
	if( (ip_hdr->total_length > size  && size <= IP_MIN_PACKET_SIZE) || 
		(ip_hdr->total_length != size && size >  IP_MIN_PACKET_SIZE) )
	{
		return NET_ERR_IP_PACKET_LENGTH;
	}

	if(ipv4_verify_checksum(ip_hdr))
	{
		return NET_ERR_IP_HDR_CHCKSUM;
	}

	// check whether the receiver of the packet is correct
	dst_ip->version =4;
	dst_ip->ipv4_addr[0] = ip_hdr->destination_ip[0];
	dst_ip->ipv4_addr[1] = ip_hdr->destination_ip[1];
	dst_ip->ipv4_addr[2] = ip_hdr->destination_ip[2];
	dst_ip->ipv4_addr[3] = ip_hdr->destination_ip[3];

	// Fragmentation is not supported
	if(ip_hdr->flags & 0x1 || ip_hdr->fragment_offset >0)
	{
		return NET_ERR_IP_FRAGMENTED;
	}

	*protocol = ip_hdr->protocol;
	if(payload_length != NULL) {
		*payload_length = ip_hdr->total_length - sizeof(ipv4_header_t);
	}

	// register addr ip src like ipv4
	src_ip->version = IP_VERSION_4;
	for(uint8_t i = 0; i < 4; i++)
	{
		src_ip->ipv4_addr[i] = ip_hdr->source_ip[i];
	}

	return NET_ERR_OK;
}


///////////////////////////////////////////////////////////////
//Support functions related to the IPv4 protocol
///////////////////////////////////////////////////////////////


//This function compute and returns the IP header checksum. The function ignore the field checksum. Uses the Buffer Address as input
unsigned short ipv4_compute_checksum(ipv4_header_t *ipv4_header){

	unsigned int checksum = 0;
	unsigned short *pkt_hdr = (unsigned short*) ipv4_header;

	// calculate sum ignoring values of checksum field
	#pragma loopbound min 10 max 10
	for (int i = 0; i<10; i++){
		if(i!=5)
		{
			checksum = checksum + pkt_hdr[i];
		}
	}

	// iterate over sum until there is no overflow
	#pragma loopbound min 0 max 2
	while ((checksum & 0xFFFF0000) > 0)
	{
		checksum = (checksum & 0xFFFF) + (checksum >> 16);
	}
	// calculate ones's complement
	checksum = ((~checksum) & 0xFFFF);
	return (unsigned short) checksum;		
}

//This function verify the IP header checksum. If the checksum is correct it returns 1, otherwise it returns 0.
char ipv4_verify_checksum(ipv4_header_t *ipv4_header){
	unsigned int checksum = 0;
	unsigned short *pkt_hdr = (unsigned short*) ipv4_header;

	// calculate sum including checksum value
	#pragma loopbound min 10 max 10
	for (int i = 0; i<10; i++){
		checksum = checksum + pkt_hdr[i];
	}

	// iterate over sum until there is no overflow
	#pragma loopbound min 0 max 2
	while ((checksum & 0xFFFF0000) > 0)
	{
		checksum = (checksum & 0xFFFF) + (checksum >> 16);
	}
	// calculate ones's complement
	checksum = ((~checksum) & 0xFFFF);

	// If checksum is valid, the result is 0
	if (checksum != 0){
		return 1;
	}else{
		return 0;
	}	
}