/*
 * TCP section
 * 
 * Authors: Christoph Lehr
 */

#include <stdio.h>
#include <stdlib.h>
#include "tcp.h"
#include "../../inc/net_types.h"
#include "../../inc/generic_types.h"
#include "../../inc/utils.h"

uint16_t socket_serv_id = 0xFFFF;

// headers received
tcp_frame_saved_t hdr_rcv[TCP_MAX_HDR_SAVE];
uint8_t rd_ptr = 0;
uint8_t wr_ptr = 0;

// prototype function use in tcp
void tcp_deep_copy_hdr(tcp_header_t* hdr1, tcp_header_t* hdr2);
uint8_t find_connection(socket_t* server, ip_address_t* addr, uint16_t port);
void update_seq_ack(socket_connection_t* conn, tcp_frame_saved_t* frame);
void tcp_build_pure_ack(tcp_header_t* tcp_hdr, uint16_t destination_port, uint16_t source_port, socket_connection_t* conn);

///////////////////////////////////////////////////////////////
//Support functions related to the TCP protocol
///////////////////////////////////////////////////////////////

// Init the TCP socket server, listening for a connection
void tcp_init_server(socket_t* socket_server, uint16_t socket_id)
{
	for(uint8_t i = 0; i < MAX_CONNECTIONS_PER_SOCKET; i++)
	{
		socket_server->connections[i].state = TCP_STATE_LISTENING;
		socket_server->connections[i].response = NULL;
	}
	for(uint8_t i = 0; i < TCP_MAX_HDR_SAVE; i++)
	{
		hdr_rcv[i].free = 1;
	}
	socket_serv_id = socket_id;
}

// Update the TCP socket server, entry point on TCP server algo
void tcp_update() {
	if(socket_serv_id >= MAX_SOCKETS) return;			// check valid ID
	socket_t* socket_serv = socket_get(socket_serv_id);	// get socket
	if(socket_serv == NULL) return;						// check valid socket
	if(socket_serv->type != SOCKET_TYPE_STREAM) return;	// check socket type TCP

	tcp_frame_saved_t* frame;
	socket_connection_t* conn;	// create socket_connection
	network_error_t ret;

	while(hdr_rcv[rd_ptr].free == 0)
	{
		frame = &hdr_rcv[rd_ptr];
		if((frame->hdr.flags & SYN_MSK) > 0)
		{
			utils_log("rcv SYN\n");
			uint8_t found = 0;
			for(int i = 0; i < MAX_CONNECTIONS_PER_SOCKET; i++)
			{
				conn = &socket_serv->connections[i];
				if(conn->state == TCP_STATE_LISTENING)
				{
					found = 1;
					i = MAX_CONNECTIONS_PER_SOCKET;
				}
			}
			if(found == 1)
			{
				conn->state = TCP_STATE_OPENING;
				update_seq_ack(conn, frame);
				conn->tx.acknowledgement_number = conn->rx.sequence_number + 1;
				conn->tx.sequence_number = 20316;
				conn->tx.first_seq_num = conn->tx.sequence_number;
				socket_deep_copy_addr(&frame->addr, &conn->peer);
				ret = socket_send(socket_serv_id, &conn->peer, NULL, 0);
				if(ret != NET_ERR_OK)
				{
					utils_log("prb send tcp : 0x%04hx\n", ret);
				}
				else
				{
					utils_log("send SYN ACK\n");
				}
			}
		}
		else
		{
			uint8_t id_connection = find_connection(socket_serv, &frame->addr.ip_address, frame->addr.port);
			if(id_connection < MAX_CONNECTIONS_PER_SOCKET)
			{
				conn = &socket_serv->connections[id_connection];
				update_seq_ack(conn, frame);
				switch(conn->state)
				{
				case TCP_STATE_OPENING:
					if(	((frame->hdr.flags & ACK_MSK) > 0) && 
						(frame->payload_length == 0) && 
						(frame->hdr.acknowledgement_number == conn->tx.sequence_number)
					)
					{
						utils_log("rcv ACK\n");
						conn->state = TCP_STATE_WAITING;
					}
					else
					{
						utils_log("error incoherent frame received connection state oppening\n");
					}
					break;
				case TCP_STATE_WAITING:
					if(frame->payload_length == 0)
					{
						if((frame->hdr.flags & FIN_MSK) != 0)
						{
							utils_log("rcv FIN ACK\n");
							conn->tx.acknowledgement_number = conn->rx.sequence_number + 1;
							conn->state = TCP_STATE_CLOSING;
							ret = socket_send(socket_serv_id, &conn->peer, NULL, 0);
							if(ret != NET_ERR_OK)
							{
								utils_log("prb send tcp : 0x%04hx\n", ret);
							}
							else
							{
								utils_log("send FIN ACK\n");
							}
						}
						else
						{
							utils_log("error incoherent frame received during state wating\n");
						}
					}
					else
					{
						utils_log("rcv request\n");
						conn->tx.acknowledgement_number = conn->rx.sequence_number + frame->payload_length;
						ret = socket_send(socket_serv_id, &conn->peer, NULL, 0);
						if(ret != NET_ERR_OK)
						{
							utils_log("prb send tcp : 0x%04hx\n", ret);
						}
						else
						{
							utils_log("send ACK\n");
						}
					}
					break;
				case TCP_STATE_SENDING:
					utils_log("error frame received during sending state, maybe the request again befor the response\n");
					break;
				case TCP_STATE_SEND_NEXT:
					if(frame->payload_length != 0 || ((frame->hdr.flags & ACK_MSK) == 0))
					{
						utils_log("error incoherent frame received connection state send_next\n");
					}
					else
					{
						conn->tx.acknowledgement_number = conn->rx.sequence_number;
						if((conn->rx.acknowledgement_number - conn->tx.first_seq_num) > conn->response_len)
						{
							utils_log("all response ack by client\n");
							conn->response = NULL;
							conn->response_len = 0;
							conn->state = TCP_STATE_WAITING;
						}
						else
						{
							conn->tx.sequence_number = conn->rx.acknowledgement_number;
							ret = socket_send(socket_serv_id, &conn->peer, NULL, 0);
							if(ret != NET_ERR_OK)
							{
								utils_log("prb send tcp : 0x%04hx\n", ret);
							}
							else
							{
								utils_log("send next part response 1\n");
							}
						}
					}
					break;

				case TCP_STATE_CLOSING:
					if((frame->payload_length == 0) && ((frame->hdr.flags & ACK_MSK) != 0))
					{
						if(conn->rx.acknowledgement_number == conn->tx.sequence_number)
						{
							utils_log("rcv ACK\n");
							utils_log("end of communication, free connection -> LISTENING\n");
							conn->response = NULL;
							conn->response_len = 0;
							conn->rx.acknowledgement_number = 0;
							conn->rx.sequence_number = 0;
							conn->tx.acknowledgement_number = 0;
							conn->tx.first_seq_num = 0;
							conn->tx.sequence_number = 0;
							conn->peer.port = 0;
							conn->state = TCP_STATE_LISTENING;
							break;
						}
					}
					utils_log("error incoherent frame received connection state closing\n");
					break;
				default:
					utils_log("error tcp_update() connection state unknow\n");
					break;
				}
			}
			else
			{
				utils_log("error tcp_update() connection not found\n");
			}
		}

		frame->free = 1;
		rd_ptr = ((rd_ptr + 1) > (TCP_MAX_HDR_SAVE - 1)) ? 0 : (rd_ptr + 1);
	}
	for(uint8_t i = 0; i < MAX_CONNECTIONS_PER_SOCKET; i++)
	{
		conn = &socket_serv->connections[i];
		if(conn->state == TCP_STATE_SEND_NEXT)
		{
			if((conn->tx.sequence_number - conn->tx.first_seq_num) <= conn->response_len)
			{
				ret = socket_send(socket_serv_id, &conn->peer, NULL, 0);
				if(ret == NET_ERR_OK)
				{
					utils_log("send next part response 2\n");
				}
				else
				{
					utils_log("prb try send next : 0x%04x\n", ret);
				}
			}
		}
	}
}

uint16_t tcp_compute_checksum(tcp_header_t *tcp_hdr, ip_address_t *dst_ip, ip_address_t *src_ip, uint8_t *data, uint16_t data_length){
	
	uint32_t sum = 0;

	uint16_t tcp_hdr_len = (tcp_hdr->flags >> DATA_OFFSET_SHT) * 4;
	uint16_t tcp_segment_len = tcp_hdr_len + data_length;

	sum += src_ip->ipv4_addr[0] << 8 | src_ip->ipv4_addr[1];
	sum += src_ip->ipv4_addr[2] << 8 | src_ip->ipv4_addr[3];
	sum += dst_ip->ipv4_addr[0] << 8 | dst_ip->ipv4_addr[1];
	sum += dst_ip->ipv4_addr[2] << 8 | dst_ip->ipv4_addr[3];

	sum += IP_PROTOCOL_TCP;
	sum += tcp_segment_len;

	tcp_hdr->checksum = 0;
	
	uint16_t* ptr_hdr = (uint16_t*) tcp_hdr;
	while(tcp_hdr_len > 1)
	{
		sum += *ptr_hdr;
		ptr_hdr++;
		tcp_hdr_len -= 2;
	}

	// Calculate sum of data
	// convert bytes into 16 bit words for checksum calculation
	uint16_t *data_words = (uint16_t*) data;

	while(data_length > 1)
	{
		sum += *data_words;
		data_words++;
		data_length -= 2;
	}

	if(data_length > 0)
	{
		sum += ((*((uint8_t*)data_words)) << 8);
	}

	// iterate over sum until there is no overflow
	#pragma loopbound min 0 max 2
	while((sum & 0xFFFF0000) > 0)
	{
		sum = (sum >> 16) + (sum & 0xFFFF);
	}
	sum = (uint16_t)(~sum);

	return sum;		
}

// compute and returns tcp checksum. The function ignore the the field checksum.
uint8_t tcp_verify_checksum(tcp_header_t *tcp_hdr, ip_address_t *dest_ip, ip_address_t *src_ip, unsigned char *data, unsigned short data_length){
	
	uint16_t check_sum = tcp_hdr->checksum;
	uint16_t result = tcp_compute_checksum(tcp_hdr, dest_ip, src_ip, data, data_length);

	if(check_sum == 0 || result != check_sum)
	{
		return 0;
	}
	return 1;
}

// build TCP header for the good destination    
network_error_t tcp_build_datagram(ip_address_t *destination_ip, 
	uint16_t destination_port, ip_address_t *source_ip, uint16_t source_port, 
	uint8_t** data, uint16_t* data_length, uint8_t *buffer)
{
	if(socket_serv_id >= MAX_SOCKETS) return NET_ERR_TCP_NO_SERVER;				// check valid id
	socket_t* socket_serv = socket_get(socket_serv_id);							// get socket
	if(socket_serv == NULL) return NET_ERR_TCP_NO_SERVER;						// check valid socket
	if(socket_serv->type != SOCKET_TYPE_STREAM) return NET_ERR_TCP_NO_SERVER;	// check socket type TCP

	socket_connection_t* conn;							// creat socket_connection
	tcp_header_t* tcp_hdr = (tcp_header_t*) buffer;		// creat tcp header

	// look for the socket_connection with matching IP and port
	uint8_t num_connection = find_connection(socket_serv, destination_ip, destination_port);
	if(num_connection < MAX_CONNECTIONS_PER_SOCKET)
	{
		conn = &socket_serv->connections[num_connection];
		switch(conn->state)
		{
		case TCP_STATE_OPENING:
			// build header
			tcp_build_pure_ack(tcp_hdr, destination_port, source_port, conn);
			tcp_hdr->flags |= (5 << DATA_OFFSET_SHT) | SYN_MSK; // 5 has to be changed if option

			// compute and set checksum
			tcp_hdr->checksum = tcp_compute_checksum(tcp_hdr, destination_ip, source_ip, *data, *data_length);

			// update sequence number
			conn->tx.sequence_number += 1;	// += data_lenght when sending a response to a request
			break;
		
		case TCP_STATE_WAITING:
			if(*data == NULL || *data_length == 0)
			{
				// build header
				tcp_build_pure_ack(tcp_hdr, destination_port, source_port, conn);

				// compute and set checksum
				tcp_hdr->checksum = tcp_compute_checksum(tcp_hdr, destination_ip, source_ip, *data, *data_length);
				break;
			}
			else
			{
				conn->state = TCP_STATE_SENDING;
			}

		case TCP_STATE_SENDING:
			conn->response_len = *data_length;
			conn->response = *data;
			conn->tx.first_seq_num = conn->tx.sequence_number - 1;
			tcp_build_pure_ack(tcp_hdr, destination_port, source_port, conn);
			if(*data_length > TCP_MAX_DATA_LENGHT)
			{
				// change data lenght for this frame
				*data_length = TCP_MAX_DATA_LENGHT;
			}
			tcp_hdr->checksum = tcp_compute_checksum(tcp_hdr, destination_ip, source_ip, *data, *data_length);
			conn->tx.sequence_number += *data_length;
			conn->state = TCP_STATE_SEND_NEXT;
			utils_log("begin send resp\n");
			break;

		case TCP_STATE_SEND_NEXT:
			tcp_build_pure_ack(tcp_hdr, destination_port, source_port, conn);
			*data_length = conn->response_len - ((conn->tx.sequence_number - 1) - conn->tx.first_seq_num);
			if(*data_length > TCP_MAX_DATA_LENGHT)
			{
				*data_length = TCP_MAX_DATA_LENGHT;
			}
			*data = (conn->response + ((conn->tx.sequence_number - 1) - conn->tx.first_seq_num));
			tcp_hdr->checksum = tcp_compute_checksum(tcp_hdr, destination_ip, source_ip, *data, *data_length);
			conn->tx.sequence_number += *data_length;
			break;

		case TCP_STATE_CLOSING:
			tcp_build_pure_ack(tcp_hdr, destination_port, source_port, conn);
			tcp_hdr->flags |= FIN_MSK;
			tcp_hdr->checksum = tcp_compute_checksum(tcp_hdr, destination_ip, source_ip, *data, *data_length);
			conn->tx.sequence_number += 1; // because use FIN flag
			break;
		default:
			utils_log("error try send frame state unknow\n");
			utils_log("conn state : %d\n", conn->state);
			return NET_ERR_TCP_CONNECTION_STATE;
		}
	}
	else
	{
		return NET_ERR_TCP_CONNECTION_NOT_FOUND;	// no socket_connection with good IP and port
	}

	// change order of bytes from lib order to network order
	for(uint8_t i = 0; i < 5; i++)
	{
		tcp_hdr->raw[i] = utils_htonl(tcp_hdr->raw[i]);
	}

	return NET_ERR_OK;
}

network_error_t tcp_handle_datagram(uint8_t *buffer_tcp, ip_address_t *destination_ip, 
	uint16_t *destination_port, socket_address_t* src_addr, uint16_t* socket_tcp_serv_id, 
	uint16_t* payload_length, time_stamp_t* rx_ts)
{
	if(hdr_rcv[wr_ptr].free == 1)
	{
		tcp_header_t* tcp_hdr = (tcp_header_t*) buffer_tcp;
		
		// change the order of bytes from network order to lib order
		for(uint8_t i = 0; i < 5; i++)
		{
			tcp_hdr->raw[i] = utils_htonl(tcp_hdr->raw[i]);
		}

		uint16_t hdr_len = ((tcp_hdr->flags) >> DATA_OFFSET_SHT) * 4;

		// verifying the checksun
		if(tcp_verify_checksum(tcp_hdr, destination_ip, &src_addr->ip_address, &buffer_tcp[hdr_len], *payload_length - hdr_len) == 0)
		{
			utils_log("parsing error tcp checksum\n");
			return NET_ERR_TCP_CHCKSUM;
		}

		// return information of the header
		src_addr->port = tcp_hdr->src_port;								// src port
		*destination_port = tcp_hdr->dest_port;							// dst port				
		*payload_length -= ((tcp_hdr->flags) >> DATA_OFFSET_SHT) * 4;	// payload lenght

		// save infos for handle it later during tcp update
		tcp_deep_copy_hdr(tcp_hdr, &hdr_rcv[wr_ptr].hdr);				// header
		hdr_rcv[wr_ptr].payload_length = *payload_length;				// payload lenght
		hdr_rcv[wr_ptr].ts = *rx_ts;									// time stamp
		socket_deep_copy_addr(src_addr, &hdr_rcv[wr_ptr].addr);			// src address

		// return id of socket tcp server to save the payload (if payload lenght > 0)
		*socket_tcp_serv_id = socket_serv_id;
		
		// update list of saved headers
		hdr_rcv[wr_ptr].free = 0;
		wr_ptr = ((wr_ptr + 1) > (TCP_MAX_HDR_SAVE - 1)) ? 0 : (wr_ptr + 1);

		return NET_ERR_OK;
	}

	return NET_ERR_TCP_SAVE_HDR;	// No more place to save TCP header
}

void tcp_deep_copy_hdr(tcp_header_t* hdr1, tcp_header_t* hdr2)
{
	hdr2->src_port 					= hdr1->src_port;
	hdr2->dest_port 				= hdr1->dest_port;
	hdr2->sequence_number 			= hdr1->sequence_number;
	hdr2->acknowledgement_number 	= hdr1->acknowledgement_number;
	hdr2->flags 					= hdr1->flags;
	hdr2->window_size 				= hdr1->window_size;
	hdr2->checksum 					= hdr1->checksum;
	hdr2->urgent_pointer 			= hdr1->urgent_pointer;
}

uint8_t find_connection(socket_t* server, ip_address_t* addr, uint16_t port)
{
	for(uint8_t i = 0; i < MAX_CONNECTIONS_PER_SOCKET; i++)
	{
		if(server->connections[i].state != TCP_STATE_LISTENING)
		{
			if((utils_compare_ip(&server->connections[i].peer.ip_address, addr) == 0)
				&& server->connections[i].peer.port == port
			)
			{
				return i;
			}
		}
	}
	return 0xFF;
}

void update_seq_ack(socket_connection_t* conn, tcp_frame_saved_t* frame)
{
	conn->rx.sequence_number = frame->hdr.sequence_number;
	conn->rx.acknowledgement_number = frame->hdr.acknowledgement_number;
}

void tcp_build_pure_ack(tcp_header_t* tcp_hdr, uint16_t destination_port, uint16_t source_port, socket_connection_t* conn)
{
	tcp_hdr->dest_port = destination_port;
	tcp_hdr->src_port = source_port;
	tcp_hdr->sequence_number = conn->tx.sequence_number;
	tcp_hdr->acknowledgement_number = conn->tx.acknowledgement_number;
	tcp_hdr->flags = (5 << DATA_OFFSET_SHT) | ACK_MSK; // 5 has to be changed if option

	tcp_hdr->window_size = 1000;
	tcp_hdr->urgent_pointer = 0;
}


network_error_t tcp_establish_connection(socket_connection_t *conn, uint8_t *options_buffer, uint8_t *options_length)
{

	return NET_ERR_OK;
}

network_error_t tcp_register_connection(socket_connection_t *conn, uint8_t *options_buffer, uint8_t *options_length)
{

	return NET_ERR_OK;
}

	// // OLD IMPLEMENTATION
	// update last received sequence number
	// conn->rx.sequence_number = tcp_hdr->sequence_number;

	// if(tcp_hdr->flags.ack == 1)
	// {
	// 	message acknowledges data

	// 	update acknowledgement number 
	// 	conn->rx.acknowledgement_number = tcp_hdr->acknowledgement_number;
	// }


	// uint32_t msg_length = tcp_hdr->sequence_number - conn->tx.acknowledgement_number + sizeof(tcp_header_t);

	// // check whether tcp packet length is set correctly
	// if( (msg_length > size  && size <= TCP_MIN_PACKET_SIZE) || 
	// 	(msg_length != size && size >  TCP_MIN_PACKET_SIZE) )
	// {
	// 	return NET_ERR_TCP_MSG_LENGTH;
	// }