
/*
 * TCP section
 * 
 * Authors: Christoph Lehr
 */

#ifndef _TCP_H_
#define _TCP_H_

#include "../../inc/net_types.h"
#include "../../inc/generic_types.h"
#include "../../inc/socket.h"

#ifndef TCP_WCET_MAX_DATA_SIZE
   // Implementation does not support Ethernet jumbo frames 
	// or IPv4 Fragmentation. Max Ethernet user data 1500 Byte
	// minus IPv4 Header 20 Bytes and TCP Header 20 Bytes, therefore
	// the maximum payload of TCP is 1460 Byte, which comes down to
	// 365 32 bit words. (1500 - 20 - 20) / 4 = 365
   #define TCP_WCET_MAX_DATA_SIZE 365
#endif

// An ethernet frame has a minimum size of 64 bytes
// Ethernet Header + IPv4 = 30 Bytes => min size = 34 bytes
#define TCP_MIN_PACKET_SIZE 34
#define TCP_MAX_DATA_LENGHT 1448

#define TCP_MAX_HDR_SAVE 2 * SOCKET_RECV_QUEUE_LENGTH

typedef struct
{
	tcp_header_t hdr;
	time_stamp_t ts;
	uint16_t payload_length;
	socket_address_t addr;
	uint8_t free;
} tcp_frame_saved_t;

enum tcp_flag_msk_t
{
	FIN_MSK			=	0x1,	// 1 bit  (00000000 00000001)
	SYN_MSK		 	=	0x2, 	// 1 bit  (00000000 00000010)
	RST_MSK		 	=	0x4, 	// 1 bit  (00000000 00000100)
	PSH_MSK		 	=	0x8, 	// 1 bit  (00000000 00001000)
	ACK_MSK		 	=	0x10,	// 1 bit  (00000000 00010000)
	URG_MSK		 	=	0x20,	// 1 bit  (00000000 00100000)
	ECE_MSK 	 	=	0x40,	// 1 bit  (00000000 01000000)
	CWR_MSK 	 	=	0x80,	// 1 bit  (00000000 10000000)
	NS_MSK		 	=	0x100,	// 1 bit  (00000001 00000000)
	RESERVED_MSK 	=	0xE00,	// 3 bits (00001110 00000000)
	DATA_OFFSET_MSK =	0xF000	// 4 bits (11110000 00000000)
};

enum tcp_flag_shift_t
{
	FIN_SHT			=	0,	// 1st  bit (00000000 00000001)
	SYN_SHT		 	=	1, 	// 2nd  bit (00000000 00000010)
	RST_SHT		 	=	2, 	// 3rd  bit (00000000 00000100)
	PSH_SHT		 	=	3, 	// 4th  bit (00000000 00001000)
	ACK_SHT		 	=	4,	// 5th  bit (00000000 00010000)
	URG_SHT		 	=	5,	// 6th  bit (00000000 00100000)
	ECE_SHT 	 	=	6,	// 7th  bit (00000000 01000000)
	CWR_SHT 	 	=	7,	// 8th  bit (00000000 10000000)
	NS_SHT		 	=	8,	// 9th  bit (00000001 00000000)
	RESERVED_SHT 	=	9,	// 10th bit (00000010 00000000)
	DATA_OFFSET_SHT =	12	// 13th bit (00010000 00000000)
};

// Init the TCP socket server, listening for a connection
void tcp_init_server(socket_t* socket_server, uint16_t socket_id);

// Update the TCP socket server, entry point on TCP server algo
void tcp_update();

// compute the tcp checksum for a packet
uint16_t tcp_compute_checksum(tcp_header_t *tcp_hdr, ip_address_t *dst_ip, ip_address_t *src_ip, uint8_t *data, uint16_t data_length);

// This function computes the TCP checksum and returns whether it is correct
uint8_t tcp_verify_checksum(tcp_header_t *tcp_hdr, ip_address_t *dest_ip, ip_address_t *src_ip, unsigned char *data, unsigned short data_length);

network_error_t tcp_establish_connection(socket_connection_t *conn, uint8_t *options_buffer, uint8_t *options_length);

network_error_t tcp_register_connection(socket_connection_t *conn, uint8_t *options_buffer, uint8_t *options_length);

// This function sends an TCP packet to the dstination IP.
network_error_t tcp_build_datagram(ip_address_t *destination_ip, 
	uint16_t destination_port, ip_address_t *source_ip, uint16_t source_port, 
	uint8_t** data, uint16_t* data_length, uint8_t *buffer);

// This function parses a received tcp packet and checks wether the checksum is correct
network_error_t tcp_handle_datagram(uint8_t *buffer_tcp, ip_address_t *destination_ip, 
	uint16_t *destination_port, socket_address_t* src_addr, uint16_t* socket_tcp_serv_id,
	uint16_t* payload_length, time_stamp_t* rx_ts);

#endif
