/*
 * UDP section
 * 
 * Authors: Christoph Lehr
 */

#include <stdio.h>
#include <stdlib.h>
#include "udp.h"
#include "../../inc/net_types.h"
#include "../../inc/generic_types.h"
#include "../../inc/utils.h"

///////////////////////////////////////////////////////////////
//Support functions related to the UDP protocol
///////////////////////////////////////////////////////////////

unsigned short udp_compute_checksum(udp_header_t *udp_hdr, ip_address_t *dest_ip, ip_address_t *src_ip, unsigned char *data, unsigned short data_length){
	
	unsigned int checksum=0;
	
	// UDP Pseudo Header : IP Header fields
	if(src_ip->version == IP_VERSION_4)
	{
		checksum += dest_ip->ipv4_addr[0] << 8 | dest_ip->ipv4_addr[1];
		checksum += src_ip->ipv4_addr[0] << 8 | src_ip->ipv4_addr[1];
		checksum += dest_ip->ipv4_addr[2] << 8 | dest_ip->ipv4_addr[3];
		checksum += src_ip->ipv4_addr[2] << 8 | src_ip->ipv4_addr[3];
	}
	else
	{
		return NET_ERR_IP_VERSION;
	}

	checksum += IP_PROTOCOL_UDP;
	checksum += udp_hdr->length;	
	
	// UDP Header
	checksum += udp_hdr->src_port;
	checksum += udp_hdr->dest_port;
	checksum += udp_hdr->length;	

	// Calculate Checksum of data
	// convert bytes into 16 bit words for checksum calculation
	unsigned short *data_words = (unsigned short*) data;
	unsigned short data_words_len = data_length >> 1;
	
	// Implementation does not support Ethernet jumbo frames 
	// or IPv4 Fragmentation. Max Ethernet user data 1500 Byte
	// minus IPv4 Header 20 Bytes and UDP Header 8 Bytes, therefore
	// the maximum payload of UDP is 1472 Byte, which comes down to
	// 368 32 bit words. (1500 -20-8) /4 = 368
	// For WCET calculation this value can be overwritten with by
	// defining UDP_WCET_MAX_DATA_SIZE
	#pragma loopbound min 1 max UDP_WCET_MAX_DATA_SIZE
	for(unsigned short i=0; i <data_words_len; i+=2)
	{
		checksum += data_words[i] + data_words[i+1] ;
	}

	// check if odd number of data words
	if(data_words_len & 0x01)
	{
		checksum += data_words[data_words_len-1];
	}

	// check if odd number of data bytes
	if(data_length & 0x01)
	{
		checksum += ((unsigned short) data[data_length-1]) <<8;
	}

	// iterate over sum until there is no overflow
	#pragma loopbound min 0 max 2
	while ((checksum & 0xFFFF0000) > 0)
	{
		checksum = (checksum & 0xFFFF) + (checksum >> 16);
	}
	checksum = ((~checksum) & 0xFFFF);

	// If the calculated checksum is 0, the calculation shall return
	// 0xFFFF as 0 indicates that no checksum was calculated. In the 
	// ones complement there is no difference between 0 and 0xFFFF.
	if(checksum == 0)
	{
		return 0xFFFF;
	}
	return checksum;		
}

//This function compute and returns the UDP checksum. The function ignore the the field checksum.
unsigned char udp_verify_checksum(udp_header_t *udp_hdr, ip_address_t *dest_ip, ip_address_t *src_ip, unsigned char *data, unsigned short data_length){
	unsigned int checksum=0;
	
	// UDP Pseudo Header : IP Header fields
	if(src_ip->version == IP_VERSION_4)
	{
		checksum += dest_ip->ipv4_addr[0] << 8 | dest_ip->ipv4_addr[1];
		checksum += src_ip->ipv4_addr[0] << 8 | src_ip->ipv4_addr[1];
		checksum += dest_ip->ipv4_addr[2] << 8 | dest_ip->ipv4_addr[3];
		checksum += src_ip->ipv4_addr[2] << 8 | src_ip->ipv4_addr[3];
	}

	checksum += IP_PROTOCOL_UDP;
	checksum += udp_hdr->length;	
	
	// UDP Header
	checksum += udp_hdr->src_port;
	checksum += udp_hdr->dest_port;
	checksum += udp_hdr->length;
	checksum += udp_hdr->checksum;

	// Calculate Checksum of data
	// convert bytes into 16 bit words for checksum calculation
	unsigned short *data_words = (unsigned short*) data;
	unsigned short data_words_len = data_length >> 1;
	
	// Implementation does not support Ethernet jumbo frames 
	// or IPv4 Fragmentation. Max Ethernet user data 1500 Byte
	// minus IPv4 Header 20 Bytes and UDP Header 8 Bytes, therefore
	// the maximum payload of UDP is 1472 Byte, which comes down to
	// 368 32 bit words. (1500 -20-8) /4 = 368
	// For WCET calculation this value can be overwritten with by
	// defining UDP_WCET_MAX_DATA_SIZE
	#pragma loopbound min 1 max UDP_WCET_MAX_DATA_SIZE
	for(unsigned short i=0; i <data_words_len; i+=2)
	{
		checksum += data_words[i] + data_words[i+1] ;
	}

	// check if odd number of data words
	if(data_words_len & 0x01)
	{
		checksum += data_words[data_words_len-1];
	}

	// check if odd number of data bytes
	if(data_length & 0x01)
	{
		checksum += ((unsigned short) data[data_length-1]) <<8;
	}
	
	// iterate over sum until there is no overflow
	#pragma loopbound min 0 max 2
	while ((checksum & 0xFFFF0000) > 0)
	{
		checksum = (checksum & 0xFFFF) + (checksum >> 16);
	}
	checksum = ((~checksum) & 0xFFFF);

	if(checksum != 0)
	{
		return 1;
	}

	return 0;
}

//This function sends an UDP packet to the dstination IP.
network_error_t udp_build_datagram(ip_address_t *destination_ip, uint16_t destination_port, ip_address_t *source_ip, uint16_t source_port, uint8_t *data, uint16_t data_length, uint8_t *buffer)
{
	udp_header_t * udp_hdr = (udp_header_t * ) buffer;

	udp_hdr->dest_port = destination_port;
	udp_hdr->src_port = source_port;
	udp_hdr->length = data_length + sizeof(udp_header_t);
	udp_hdr->checksum = udp_compute_checksum(udp_hdr, destination_ip, source_ip, data, data_length);
	
	udp_hdr->raw[0] = utils_htonl(udp_hdr->raw[0]);
	udp_hdr->raw[1] = utils_htonl(udp_hdr->raw[1]);

	return NET_ERR_OK;
}

network_error_t udp_handle_datagram(uint8_t *buffer, uint16_t size, 
	ip_address_t *destination_ip, uint16_t *destination_port, 
	ip_address_t *source_ip, uint16_t *source_port)
{
	udp_header_t *udp_hdr = (udp_header_t*) buffer;
	
	udp_hdr->raw[0] = utils_ntohl(udp_hdr->raw[0]);
	udp_hdr->raw[1] = utils_ntohl(udp_hdr->raw[1]);

	// check whether udp packet length is set correctly
	if( (udp_hdr->length > size  && size <= UDP_MIN_PACKET_SIZE) || 
		(udp_hdr->length != size && size > UDP_MIN_PACKET_SIZE) )
	{
		return NET_ERR_UDP_MSG_LENGTH;
	}

	if(udp_hdr->checksum !=0 && udp_verify_checksum(udp_hdr, destination_ip, source_ip, &buffer[8], udp_hdr->length) == 0)
	{
		return NET_ERR_UDP_CHCKSUM;
	}
    *source_port = udp_hdr->src_port;
    *destination_port = udp_hdr->dest_port;
	return NET_ERR_OK;
}
