#include <stdlib.h>
#include "ptp.h"
#include "../../../inc/services/time_synchronisation.h"

#include "../../../inc/generic_types.h"
#include "../../../inc/net_types.h"
#include "../../../inc/network.h"
#include "../../../inc/socket.h"
#include "../../../inc/utils.h"

#include <machine/patmos.h>


///////////////////////////////////////////////////////////////////////////
// Type definitions
///////////////////////////////////////////////////////////////////////////
typedef struct
{
    uint16_t netw_if;
    uint16_t socket_id;
    ptp_port_identity_t host_port_identity;
    uint8_t domainNumber;
    int8_t logSyncInterval;
    int8_t logAnnounceInterval;
    uint8_t announceTimeout;
    int8_t logDelayReqInterval;
    int8_t logDelayRespTimeout;
    uint8_t priority1;
    uint8_t priority2;
    uint8_t clockClass;
    ptp_clock_accuracy_t clock_accuracy;
    ptp_time_source_t timeSource;
    uint16_t offsetScaledLogVariance;
    time_sync_forced_state_t forced_state;
} time_synchronisation_runtime_config_t;

typedef struct
{
    uint16_t sequenceId_sync;
    uint16_t sequenceId_delay_req;
    uint16_t sequenceId_announce;
    time_stamp_t ts_t1;
    time_stamp_t ts_t2;
    time_stamp_t ts_t3;
    time_stamp_t ts_t4;
    ptp_timestamp_t ts_sync_send;
    ptp_timestamp_t ts_delay_req_rcv;
    time_stamp_t next_announce;
    time_stamp_t next_sync;
    time_stamp_t next_delay_req;
    time_stamp_t delayRespTimeout;
    time_interval_t mean_path_delay;
    time_interval_t offset;
    time_interval_t delta;
    uint8_t announce_receveid;
    uint8_t announceTimeoutCnt;
    time_sync_state_t current_state;
    ptp_port_state_t current_ptp_state;
    struct
    {
        uint8_t grandmasterPriority1;
        uint8_t grandmasterPriority2;
        uint8_t timeSource;
        uint16_t stepsRemoved;
        ptp_clock_identity_t grandmasterIdentity;
        ptp_clock_quality_t grandmasterClockQuality;
    } announce_data;
    struct
    {
        ptp_port_identity_t requestingPortIdentity;
        uint16_t sequenceId_delay_req_rcv;
        int64_t correctionField_rcv;
    } pdelay_response;
    struct 
    {
        time_stamp_t ts_sync_rcv;
        int64_t correctionField_sync_rcv;
        uint16_t sequence_id;
    } sync_follow_up_data;
    
} time_synchronisation_runtime_state_t;

///////////////////////////////////////////////////////////////////////////
// static function definitions
///////////////////////////////////////////////////////////////////////////
static network_error_t handle_ptp_message(ptp_msg_t *msg, time_stamp_t *msg_rx_ts);
static network_error_t handle_ptp_sync_message(ptp_sync_msg_t *msg, time_stamp_t *msg_rx_ts);
static network_error_t handle_ptp_pdelay_req_message(ptp_pdelay_req_msg_t *msg);
static network_error_t handle_ptp_pdelay_resp_message(ptp_pdelay_resp_msg_t *msg);
static network_error_t handle_ptp_follow_up_message(ptp_follow_up_msg_t *msg);
static network_error_t handle_ptp_pdelay_resp_follow_up_message(ptp_pdelay_resp_follow_up_msg_t *msg);
static network_error_t handle_ptp_announce_message(ptp_announce_msg_t *msg);

static network_error_t send_ptp_sync_message(void);
static network_error_t send_ptp_pdelay_req_message(void);
static network_error_t send_ptp_pdelay_resp_follow_up_message(void);
static network_error_t send_ptp_announce_message(void);
static network_error_t send_ptp_follow_up_message(void);

static network_error_t run_bmca(void);
static void correct_runtime_timestamps(time_interval_t *offset);

///////////////////////////////////////////////////////////////////////////
// static variable definitions
///////////////////////////////////////////////////////////////////////////
static uint8_t time_sync_rcv_buffer[TIME_SYNC_RCV_BUFFER_SPACE];
static socket_buffer_cfg_t time_synch_rcv_buffer_cfg;

static ptp_header_t default_header;

static socket_address_t ptp_mac_addr;
static time_synchronisation_runtime_config_t runtime_conf;
static time_synchronisation_runtime_state_t runtime_state;

static uint8_t gptp_multicast_address[6] = {0x01, 0x80, 0xC2, 0x00, 0x00, 0x0E};

///////////////////////////////////////////////////////////////////////////
// Implementation
///////////////////////////////////////////////////////////////////////////
network_error_t time_synchronisation_init(time_synchronisation_config_t *conf)
{
    mac_address_t if_mac;
    network_error_t ret = network_get_mac_address(conf->netw_if, &if_mac);
    if (ret != NET_ERR_OK)
    {
        return ret;
    }

    if (if_mac.length == 6)
    {
        runtime_conf.host_port_identity.clockIdentity.id[0] = if_mac.address[0];
        runtime_conf.host_port_identity.clockIdentity.id[1] = if_mac.address[1];
        runtime_conf.host_port_identity.clockIdentity.id[2] = if_mac.address[2];
        runtime_conf.host_port_identity.clockIdentity.id[3] = if_mac.address[3];
        runtime_conf.host_port_identity.clockIdentity.id[4] = 0xFF;
        runtime_conf.host_port_identity.clockIdentity.id[5] = 0xFE;
        runtime_conf.host_port_identity.clockIdentity.id[6] = if_mac.address[4];
        runtime_conf.host_port_identity.clockIdentity.id[7] = if_mac.address[5];
    }
    else
    {
        #pragma loopbound min 8 max 8
        for (uint8_t i = 0; i < 8; i++)
        {
            runtime_conf.host_port_identity.clockIdentity.id[i] = if_mac.address[i];
        }
    }

    runtime_conf.netw_if = conf->netw_if;

    runtime_conf.host_port_identity.portNumber = conf->netw_if;
    runtime_conf.domainNumber = conf->domainNumber;
    runtime_conf.logSyncInterval = conf->logSyncInterval;
    runtime_conf.logDelayReqInterval = conf->logDelayReqInterval;
    runtime_conf.logAnnounceInterval = conf->logAnnounceInterval;
    runtime_conf.logDelayRespTimeout = conf->logDelayRespTimeout;
    runtime_conf.forced_state = conf->forced_state;

    runtime_conf.priority1 = 248; // 248 is the default grandmaster capable endsystem that is not critical in gptp
    runtime_conf.priority2 = 248; // 248 is the default grandmaster capable endsystem in gptp
    runtime_conf.clock_accuracy = conf->clock_accuracy;
    runtime_conf.clockClass = 248;                                 // We assume to not be connected to a primary time source
    runtime_conf.timeSource = PTP_TIME_SOURCE_INTERNAL_OSCILLATOR; // As we are not connected to a primary time source we assume the internal oscillator is used
    runtime_conf.offsetScaledLogVariance = 0xFFFF;

    time_synch_rcv_buffer_cfg.buffer_entry_size = TIME_SYNC_RCV_BUFFER_SIZE;
    time_synch_rcv_buffer_cfg.buffer_size = TIME_SYNC_RCV_BUFFER_SIZE*TIME_SYNC_RCV_BUFFER_COUNT;
    time_synch_rcv_buffer_cfg.buffer_space = (uint8_t *)&time_sync_rcv_buffer;

    runtime_conf.socket_id = socket(SOCKET_DOMAIN_ETHERNET, SOCKET_TYPE_RAW, ETHERTYPE_PTP);
    if (runtime_conf.socket_id == 0xFFFF)
    {
        return NET_ERR_SOCKET;
    }

    ptp_mac_addr.hw_address.length = 6;
    ptp_mac_addr.hw_address.address[0] = gptp_multicast_address[0];
    ptp_mac_addr.hw_address.address[1] = gptp_multicast_address[1];
    ptp_mac_addr.hw_address.address[2] = gptp_multicast_address[2];
    ptp_mac_addr.hw_address.address[3] = gptp_multicast_address[3];
    ptp_mac_addr.hw_address.address[4] = gptp_multicast_address[4];
    ptp_mac_addr.hw_address.address[5] = gptp_multicast_address[5];
    ret = socket_bind(runtime_conf.socket_id, &ptp_mac_addr);

    if (ret != NET_ERR_OK)
    {
        return ret;
    }

    // init default header
    default_header.majorSdoId = 0x1;
    default_header.minorSdoId = 0x0;
    default_header.minorVersionPTP = 0x0;
    default_header.majorVersionPTP = 0x2;
    default_header.domainNumber = runtime_conf.domainNumber;
    default_header.sourcePortIdentity = runtime_conf.host_port_identity;

    // init runtime states
    runtime_state.sequenceId_announce = 0;
    runtime_state.sequenceId_delay_req = 0;
    runtime_state.sequenceId_sync = 0;
    runtime_state.current_ptp_state = PTP_PORT_STATE_SLAVE;

    // init intervals
    time_stamp_t current_ts;
    uint32_t ts_size;
    ret = network_get_option(runtime_conf.netw_if,
                             NETWORK_OPTION_GET_PTP_CLOCK, &current_ts, &ts_size);

    if (ret != NET_ERR_OK)
    {
        return ret;
    }

    if (runtime_conf.logAnnounceInterval >= 0)
    {
        runtime_state.next_announce.seconds = current_ts.seconds + (1 << runtime_conf.logAnnounceInterval);
        runtime_state.next_announce.nanoseconds = current_ts.nanoseconds;
    }
    else
    {
        runtime_state.next_announce.seconds = current_ts.seconds;
        runtime_state.next_announce.nanoseconds = current_ts.nanoseconds + (1000000000 >> -runtime_conf.logAnnounceInterval);
    }

    if (runtime_conf.logSyncInterval >= 0)
    {
        runtime_state.next_sync.seconds = current_ts.seconds + (1 << runtime_conf.logSyncInterval);
        runtime_state.next_sync.nanoseconds = current_ts.nanoseconds;
    }
    else
    {
        runtime_state.next_sync.seconds = current_ts.seconds;
        runtime_state.next_sync.nanoseconds = current_ts.nanoseconds + (1000000000 >> -runtime_conf.logSyncInterval);
    }

    if (runtime_conf.logDelayReqInterval >= 0)
    {
        runtime_state.next_delay_req.seconds = current_ts.seconds + (1 << runtime_conf.logDelayReqInterval);
        runtime_state.next_delay_req.nanoseconds = current_ts.nanoseconds;
    }
    else
    {
        runtime_state.next_delay_req.seconds = current_ts.seconds;
        runtime_state.next_delay_req.nanoseconds = current_ts.nanoseconds + (1000000000 >> -runtime_conf.logDelayReqInterval);
    }

    return NET_ERR_OK;
}

network_error_t time_synchronisation_update(void)
{
    socket_address_t src_addr;
    uint8_t data[TIME_SYNC_RCV_BUFFER_SIZE];
    uint16_t data_length;

    // Read out messages received by socket
    network_error_t ret = NET_ERR_OK;
    uint8_t cnt=0;
    time_stamp_t msg_rx_ts;
#pragma loopbound min 1 max TIME_SYNC_RCV_BUFFER_COUNT
    do
    {
        ret = socket_receive_w_ts(runtime_conf.socket_id, &src_addr, (uint8_t *)&data, &data_length, TIME_SYNC_RCV_BUFFER_SIZE, &msg_rx_ts);
        cnt++;
        if (ret == NET_ERR_OK)
        {
            network_error_t ret2 =  handle_ptp_message((ptp_msg_t *)&data, &msg_rx_ts);
            if (ret2 != NET_ERR_OK)
            {
                utils_log("handle_ptp_message type: 0x%02hhx error: 0x%04hx\n", ((ptp_msg_t *)&data)->header.msg_type, ret2);
            }            
        }
        else if(ret != NET_ERR_SOCKET_NO_MSG)
        {
            utils_log("socket_receive error: 0x%04hx\n", ret);
        }

    } while (ret != NET_ERR_SOCKET_NO_MSG);

    // Read out timestamps
    time_stamp_t current_ts, last_tx_ts;
    uint32_t ts_size;
    ret = network_get_option(runtime_conf.netw_if, NETWORK_OPTION_GET_PTP_CLOCK, &current_ts, &ts_size);

    if (ret != NET_ERR_OK)
    {
        utils_log("Error: NETWORK_OPTION_GET_PTP_CLOCK\n");
        return ret;
    }
    else if (ts_size != sizeof(time_stamp_t))
    {
        ret = NET_ERR_NOT_SUPPORTED;
    }

    network_error_t ret_tx_ts = network_get_option(runtime_conf.netw_if, NETWORK_OPTION_GET_PTP_LAST_TX_TIMESTAMP, &last_tx_ts, &ts_size);
    if (ts_size != sizeof(time_stamp_t))
    {
        ret = NET_ERR_NOT_SUPPORTED;
    }

    // Reset state if delay response timed out
    if ( (runtime_state.current_state == TIME_SYNC_STATE_PDELAY_MEASUREMENT_REQUESTER || runtime_state.current_state == TIME_SYNC_STATE_PDELAY_MEASUREMENT_REQUESTER_FOLLOW_UP) &&
        (current_ts.seconds > runtime_state.delayRespTimeout.seconds ||
            (current_ts.seconds == runtime_state.delayRespTimeout.seconds && current_ts.nanoseconds >= runtime_state.delayRespTimeout.nanoseconds))
        )
    {
        // Response Timeout
        runtime_state.current_state = TIME_SYNC_STATE_IDLE;
    }
    
    // check and run best master clock algorithm
    if ( runtime_conf.forced_state != TIME_SYNC_FORCED_STATE_SLAVE &&
        ((runtime_state.announce_receveid == 1 && runtime_state.current_ptp_state == PTP_PORT_STATE_MASTER) ||
        (runtime_state.announceTimeoutCnt >= runtime_conf.announceTimeout &&
         (runtime_state.current_ptp_state == PTP_PORT_STATE_SLAVE || runtime_state.current_ptp_state == PTP_PORT_STATE_PASSIVE))))
    {
        run_bmca();
    }
    
     
    if (runtime_state.current_state == TIME_SYNC_STATE_PDELAY_MEASUREMENT_RESPONDER)
    {
        //if (last_tx_ts.seconds > runtime_state.ts_delay_req_rcv.seconds ||
        //    (last_tx_ts.seconds == runtime_state.ts_delay_req_rcv.seconds && last_tx_ts.nanoseconds >= runtime_state.ts_delay_req_rcv.nanoseconds))
        if(ret_tx_ts == NET_ERR_OK)
        {

            ret = send_ptp_pdelay_resp_follow_up_message();
            runtime_state.current_state = TIME_SYNC_STATE_IDLE;
        }

    }
    
    else if(runtime_state.current_state == TIME_SYNC_STATE_SYNC)
    {
        if(last_tx_ts.seconds > runtime_state.ts_sync_send.seconds ||
            (last_tx_ts.seconds == runtime_state.ts_sync_send.seconds && last_tx_ts.nanoseconds >= runtime_state.ts_sync_send.nanoseconds))
        {
            ret = send_ptp_follow_up_message();
            runtime_state.current_state = TIME_SYNC_STATE_IDLE;
        }
    }
    else if (runtime_state.current_state == TIME_SYNC_STATE_IDLE)
    {
        
        if (current_ts.seconds > runtime_state.next_delay_req.seconds ||
                    (current_ts.seconds == runtime_state.next_delay_req.seconds && current_ts.nanoseconds >= runtime_state.next_delay_req.nanoseconds))
        {
            runtime_state.current_state = TIME_SYNC_STATE_PDELAY_MEASUREMENT_REQUESTER;
            ret = send_ptp_pdelay_req_message();

            if (runtime_conf.logDelayReqInterval >= 0)
            {
                runtime_state.next_delay_req.seconds = current_ts.seconds + (1 << runtime_conf.logDelayReqInterval);
                runtime_state.next_delay_req.nanoseconds = current_ts.nanoseconds;
            }
            else
            {
                runtime_state.next_delay_req.seconds = current_ts.seconds;
                runtime_state.next_delay_req.nanoseconds = current_ts.nanoseconds + (1000000000 >> -runtime_conf.logDelayReqInterval);
            }

            if (runtime_conf.logDelayRespTimeout >= 0)
            {
                runtime_state.delayRespTimeout.seconds = current_ts.seconds + (1 << runtime_conf.logDelayRespTimeout);
                runtime_state.delayRespTimeout.nanoseconds = current_ts.nanoseconds;
            }
            else
            {
                runtime_state.delayRespTimeout.seconds = current_ts.seconds;
                runtime_state.delayRespTimeout.nanoseconds = current_ts.nanoseconds + (1000000000 >> -runtime_conf.logDelayRespTimeout);
            }
        }
        // only execute one at a time to prevent mixing up of timestamps
        else if (
            (current_ts.seconds > runtime_state.next_announce.seconds ||
             (current_ts.seconds == runtime_state.next_announce.seconds && current_ts.nanoseconds >= runtime_state.next_announce.nanoseconds)))
        {
            // check whether a announce message was reiceved since the last check
            if (runtime_state.announce_receveid == 0)
            {
                runtime_state.announceTimeoutCnt++;
            }
            else
            {
                runtime_state.announce_receveid = 0;
                runtime_state.announceTimeoutCnt = 0;
            }
            // send only announce message if in master state
            if (runtime_state.current_ptp_state == PTP_PORT_STATE_MASTER)
            {
                ret = send_ptp_announce_message();
            }

            // calculate next announce time
            if (runtime_conf.logAnnounceInterval >= 0)
            {
                runtime_state.next_announce.seconds = current_ts.seconds + (1 << runtime_conf.logAnnounceInterval);
                runtime_state.next_announce.nanoseconds = current_ts.nanoseconds;
            }
            else
            {
                runtime_state.next_announce.seconds = current_ts.seconds;
                runtime_state.next_announce.nanoseconds = current_ts.nanoseconds + (1000000000 >> -runtime_conf.logAnnounceInterval);
            }
        }
        else if (runtime_state.current_ptp_state == PTP_PORT_STATE_MASTER &&
                (current_ts.seconds > runtime_state.next_sync.seconds ||
                 (current_ts.seconds == runtime_state.next_sync.seconds && current_ts.nanoseconds >= runtime_state.next_sync.nanoseconds)))
        {
            // initiate sending of sync message and trigger sending of follow up message when sync was transmitted
            ret = send_ptp_sync_message();
        
            // calculate next times for sync message
            if (runtime_conf.logSyncInterval >= 0)
            {
                runtime_state.next_sync.seconds = current_ts.seconds + (1 << runtime_conf.logSyncInterval);
                runtime_state.next_sync.nanoseconds = current_ts.nanoseconds;
            }
            else
            {
                runtime_state.next_sync.seconds = current_ts.seconds;
                runtime_state.next_sync.nanoseconds = current_ts.nanoseconds + (1000000000 >> -runtime_conf.logSyncInterval);
            }
        }
    }

    return ret;
}

static network_error_t handle_ptp_message(ptp_msg_t *msg, time_stamp_t *msg_rx_ts) 
{
    network_error_t ret = NET_ERR_OK;
    if (msg->header.domainNumber != runtime_conf.domainNumber)
    {
        return NET_ERR_OK;
    }
    //printf("msg->header.msg_type %20hx\n", msg->header.msg_type);
    switch (msg->header.msg_type)
    {
    case PTP_MSG_TYPE_SYNC:
        ret = handle_ptp_sync_message((ptp_sync_msg_t *)msg, msg_rx_ts);
        break;

    case PTP_MSG_TYPE_DELAY_REQ:
        ret = NET_ERR_NOT_SUPPORTED;
        break;

    case PTP_MSG_TYPE_PDELAY_REQ:
        ret = handle_ptp_pdelay_req_message((ptp_pdelay_req_msg_t *)msg);
        break;

    case PTP_MSG_TYPE_PDELAY_RESP:
        ret = handle_ptp_pdelay_resp_message((ptp_pdelay_resp_msg_t *)msg);
        break;

    case PTP_MSG_TYPE_FOLLOW_UP:
        ret = handle_ptp_follow_up_message((ptp_follow_up_msg_t *)msg);
        break;

    case PTP_MSG_TYPE_DELAY_RESP:
        ret = NET_ERR_NOT_SUPPORTED;
        break;

    case PTP_MSG_TYPE_PDELAY_RESP_FOLLOW_UP:
        ret = handle_ptp_pdelay_resp_follow_up_message((ptp_pdelay_resp_follow_up_msg_t *)msg);
        break;

    case PTP_MSG_TYPE_ANNOUNCE:
        ret = handle_ptp_announce_message((ptp_announce_msg_t *)msg);
        break;

    case PTP_MSG_TYPE_SIGNALLING:
    case PTP_MSG_TYPE_MANAGEMENT:
        ret = NET_ERR_NOT_SUPPORTED;
        break;
    default:
        break;
    }

    return ret;
}

static network_error_t handle_ptp_sync_message(ptp_sync_msg_t *msg, time_stamp_t *msg_rx_ts)
{
    uint32_t rx_size;   

    runtime_state.sync_follow_up_data.sequence_id = msg->header.sequenceId;

    if (msg->header.flagField.twoStepFlag == 1)
    {
        runtime_state.sync_follow_up_data.ts_sync_rcv.seconds = msg_rx_ts->seconds;
        runtime_state.sync_follow_up_data.ts_sync_rcv.nanoseconds = msg_rx_ts->nanoseconds;
        
        runtime_state.sync_follow_up_data.correctionField_sync_rcv = msg->header.correctionField;
    }
    else
    {
        // One Step is not supported by gPTP
        return NET_ERR_NOT_SUPPORTED;
    }
    return NET_ERR_OK;
}

static network_error_t handle_ptp_pdelay_req_message(ptp_pdelay_req_msg_t *msg)
{
    uint32_t rx_size;
    time_stamp_t rx_timestamp;
    network_error_t ret = network_get_option(runtime_conf.netw_if,
                                             NETWORK_OPTION_GET_PTP_LAST_RX_TIMESTAMP, &rx_timestamp, &rx_size);
    if (rx_size != sizeof(rx_timestamp))
    {
        ret = NET_ERR_NOT_SUPPORTED;
    }

    runtime_state.current_state = TIME_SYNC_STATE_PDELAY_MEASUREMENT_RESPONDER;
    
    ptp_pdelay_resp_msg_t response;
    response.header = default_header;
    response.header.sequenceId = msg->header.sequenceId;
    response.header.messageLength = sizeof(ptp_pdelay_resp_msg_t);
    response.header.msg_type = PTP_MSG_TYPE_PDELAY_RESP;
    response.header.logMessageInterval = 0x7f;
    response.header.flagField.raw = 0;
    response.header.flagField.twoStepFlag = 1;
    response.header.correctionField = 0;
    response.header.controlField = 0x05;

    response.requestReceiptTimestamp.nanoseconds = rx_timestamp.nanoseconds;
    response.requestReceiptTimestamp.seconds = rx_timestamp.seconds;
    runtime_state.ts_delay_req_rcv.nanoseconds = rx_timestamp.nanoseconds;
    runtime_state.ts_delay_req_rcv.seconds = rx_timestamp.seconds;

    response.requestingPortIdentity = msg->header.sourcePortIdentity;
    runtime_state.pdelay_response.requestingPortIdentity = msg->header.sourcePortIdentity;

    // save sequence id for delay_resp_follow_up
    runtime_state.pdelay_response.sequenceId_delay_req_rcv = msg->header.sequenceId;
    runtime_state.pdelay_response.correctionField_rcv = msg->header.correctionField;

    ret = socket_send(runtime_conf.socket_id, &ptp_mac_addr, (uint8_t *)&response, sizeof(response));
    
    return ret;
}

static network_error_t handle_ptp_pdelay_resp_message(ptp_pdelay_resp_msg_t *msg)
{

    if(runtime_state.sequenceId_delay_req != msg->header.sequenceId)
    {
        return NET_ERR_TIME_SYNC_WRONG_SEQUENCE_ID; 
    }

    uint32_t ts_size;
    time_stamp_t ts;
    // read out time stamp of request
    network_error_t ret = network_get_option(runtime_conf.netw_if,
                                             NETWORK_OPTION_GET_PTP_LAST_TX_TIMESTAMP, &ts, &ts_size);
    if (ts_size != sizeof(ts))
    {
        ret = NET_ERR_NOT_SUPPORTED;
    }
    // convert generic timestamp to ptp timestamp
    runtime_state.ts_t1.seconds = ts.seconds;
    runtime_state.ts_t1.nanoseconds = ts.nanoseconds;

    // Reception time stamp of request from responder
    runtime_state.ts_t2.seconds = msg->requestReceiptTimestamp.seconds;
    runtime_state.ts_t2.nanoseconds = msg->requestReceiptTimestamp.nanoseconds;

    // read out reception time stamp
    ret = network_get_option(runtime_conf.netw_if, NETWORK_OPTION_GET_PTP_LAST_RX_TIMESTAMP, &ts, &ts_size);
    if (ts_size != sizeof(ts))
    {
        return NET_ERR_NOT_SUPPORTED;
    }

    // convert generic timestamp to ptp timestamp
    runtime_state.ts_t4.seconds = ts.seconds;
    runtime_state.ts_t4.nanoseconds = ts.nanoseconds;

    // handle two step approach
    if (msg->header.flagField.twoStepFlag == 0)
    {
        runtime_state.current_state = TIME_SYNC_STATE_IDLE;
        return NET_ERR_NOT_SUPPORTED;
    }
    else
    {
        runtime_state.current_state = TIME_SYNC_STATE_PDELAY_MEASUREMENT_REQUESTER_FOLLOW_UP;
    }
    return NET_ERR_OK;
}

static network_error_t handle_ptp_follow_up_message(ptp_follow_up_msg_t *msg)
{
    network_error_t ret = NET_ERR_OK;

    if(runtime_state.sync_follow_up_data.sequence_id != msg->header.sequenceId)
    {
        return NET_ERR_TIME_SYNC_WRONG_SEQUENCE_ID;
    }

    // rx_ts - tx_ts - mean path delays
    time_interval_t offset;
    ptp_timestamp_t *preciseOriginTimestamp = &(msg->preciseOriginTimestamp);
    time_stamp_t tx_ts;
    tx_ts.seconds = preciseOriginTimestamp->seconds;
    tx_ts.nanoseconds = preciseOriginTimestamp->nanoseconds;
    
    utils_time_stamp_sub(&runtime_state.sync_follow_up_data.ts_sync_rcv, &tx_ts, &offset);
       
    utils_time_interval_sub(&offset, &runtime_state.mean_path_delay);

    offset.nanoseconds -= msg->header.correctionField;

    utils_time_interval_correct_of_uf(&offset);

    offset.nanoseconds -= runtime_state.sync_follow_up_data.correctionField_sync_rcv;

    utils_time_interval_correct_of_uf(&offset); 


    if(runtime_state.current_ptp_state != PTP_PORT_STATE_MASTER &&
        runtime_state.current_ptp_state != PTP_PORT_STATE_DISABLED)
    {

        if( (runtime_state.offset.seconds == 0 && runtime_state.offset.nanoseconds != 0) && 
            (runtime_state.offset.nanoseconds <= 100000 && runtime_state.offset.nanoseconds >= -100000) &&
            (offset.nanoseconds > 100000 || offset.nanoseconds < -100000))
        {
            //runtime_state.offset.seconds=0;
            //runtime_state.offset.nanoseconds=0;
            
            return ret;
        }

        if(offset.seconds !=0) 
        {
            time_stamp_t correction_ts;
            correction_ts.seconds = tx_ts.seconds;
            correction_ts.nanoseconds = tx_ts.nanoseconds;

            utils_time_stamp_add_interval(&correction_ts, &runtime_state.mean_path_delay);

            ret = network_set_option(runtime_conf.netw_if, NETWORK_OPTION_SET_PTP_CLOCK, &correction_ts, sizeof(correction_ts));

            time_interval_t correction;
            correction.nanoseconds =0;
            correction.seconds =0;
            ret = network_set_option(runtime_conf.netw_if, NETWORK_OPTION_SET_PTP_CLOCK_OFFSET, &correction, sizeof(correction));
            correct_runtime_timestamps(&correction);
            //runtime_state.offset = offset;
        }
        else if(offset.nanoseconds >= 100000 || offset.nanoseconds <= -100000)
        {
            time_stamp_t correction_ts;
            time_interval_t correction_offset;
            
            correction_offset.seconds = 0;
            correction_offset.nanoseconds = -offset.nanoseconds ;
            
            uint32_t ts_size;
            ret = network_get_option(runtime_conf.netw_if, NETWORK_OPTION_GET_PTP_CLOCK, &correction_ts, &ts_size);
            utils_time_stamp_add_interval(&correction_ts, &correction_offset);
            utils_time_stamp_add_interval(&correction_ts, &runtime_state.mean_path_delay);

            ret = network_set_option(runtime_conf.netw_if, NETWORK_OPTION_SET_PTP_CLOCK, &correction_ts, sizeof(correction_ts));
        }
        else
        {
            ret = network_set_option(runtime_conf.netw_if, NETWORK_OPTION_SET_PTP_CLOCK_OFFSET, &offset, sizeof(offset));
        }
    }

    runtime_state.offset = offset;

    return ret;
}

static network_error_t handle_ptp_pdelay_resp_follow_up_message(ptp_pdelay_resp_follow_up_msg_t *msg)
{
    if(runtime_state.sequenceId_delay_req != msg->header.sequenceId)
    {
        runtime_state.current_state = TIME_SYNC_STATE_IDLE;
        return NET_ERR_TIME_SYNC_WRONG_SEQUENCE_ID;
    }
    
    // Copy sending time stamp of request from responder
    runtime_state.ts_t3.seconds = msg->responseOriginTimestamp.seconds;
    runtime_state.ts_t3.nanoseconds = msg->responseOriginTimestamp.nanoseconds;
    time_interval_t delay_resp, delay_req, mean_path_delay;


    utils_time_stamp_sub(&runtime_state.ts_t3, & runtime_state.ts_t2, &delay_resp);

    utils_time_stamp_sub(&runtime_state.ts_t4, & runtime_state.ts_t1, &delay_req);

    mean_path_delay.seconds = delay_req.seconds;
    mean_path_delay.nanoseconds = delay_req.nanoseconds;

    utils_time_interval_sub(&mean_path_delay, &delay_resp);

    mean_path_delay.nanoseconds -= msg->header.correctionField;
    utils_time_interval_correct_of_uf(&mean_path_delay);

    mean_path_delay.nanoseconds >>= 1;
    if (mean_path_delay.seconds & 0x01)
    {
        mean_path_delay.nanoseconds += 500000000;
    }
    mean_path_delay.seconds >>= 1;
    
    if(mean_path_delay.seconds == 0 && mean_path_delay.nanoseconds >0 && mean_path_delay.nanoseconds < 1000000)
    {
        runtime_state.mean_path_delay = mean_path_delay;
    }

    runtime_state.current_state = TIME_SYNC_STATE_IDLE;

    return NET_ERR_OK;
}

static network_error_t handle_ptp_announce_message(ptp_announce_msg_t *msg)
{

    runtime_state.announce_data.grandmasterClockQuality = msg->grandmasterClockQuality;
    runtime_state.announce_data.grandmasterIdentity = msg->grandmasterIdentity;
    runtime_state.announce_data.grandmasterPriority1 = msg->grandmasterPriority1;
    runtime_state.announce_data.grandmasterPriority2 = msg->grandmasterPriority2;
    runtime_state.announce_data.stepsRemoved = msg->stepsRemoved;
    runtime_state.announce_data.timeSource = msg->timeSource;

    runtime_state.announce_receveid = 1;

    return NET_ERR_OK;
}

static network_error_t send_ptp_sync_message(void)
{
    runtime_state.sequenceId_sync++;

    ptp_sync_msg_t msg;
    msg.header = default_header;
    msg.header.sequenceId = runtime_state.sequenceId_sync;
    msg.header.messageLength = sizeof(ptp_sync_msg_t);
    msg.header.msg_type = PTP_MSG_TYPE_SYNC;
    msg.header.logMessageInterval = runtime_conf.logSyncInterval;
    msg.header.flagField.raw = 0;
    msg.header.flagField.twoStepFlag=1;
    msg.header.correctionField = 0;
    msg.header.controlField = 0x00;

    time_stamp_t sync_ts;
    uint32_t ts_size;
    network_error_t ret = network_get_option(runtime_conf.netw_if,
                                             NETWORK_OPTION_GET_PTP_CLOCK, &sync_ts, &ts_size);

    if (ret != NET_ERR_OK)
    {
        return ret;
    } else if (ts_size != sizeof(sync_ts))
    {
        ret = NET_ERR_NOT_SUPPORTED;
    }

    runtime_state.ts_sync_send.nanoseconds = sync_ts.nanoseconds;
    runtime_state.ts_sync_send.seconds =sync_ts.seconds;
    msg.originTimestamp.nanoseconds = 0; //sync_ts.nanoseconds;
    msg.originTimestamp.seconds =  0; //sync_ts.seconds;

    ret = socket_send(runtime_conf.socket_id, &ptp_mac_addr, (uint8_t *)&msg, sizeof(msg));
    
    runtime_state.current_state = TIME_SYNC_STATE_SYNC;

    return ret;
}

static network_error_t send_ptp_follow_up_message(void)
{
    ptp_follow_up_msg_t msg;
    msg.header = default_header;
    msg.header.sequenceId = runtime_state.sequenceId_sync;
    msg.header.messageLength = sizeof(ptp_follow_up_msg_t);
    msg.header.msg_type = PTP_MSG_TYPE_FOLLOW_UP;
    msg.header.logMessageInterval = runtime_conf.logDelayReqInterval;
    msg.header.flagField.raw = 0;
    //msg.header.flagField.twoStepFlag = 1;
    msg.header.correctionField = 0;
    msg.header.controlField = 0x02;

    time_stamp_t sync_tx_ts;
    uint32_t ts_size;
    network_error_t ret = network_get_option(runtime_conf.netw_if,
                                             NETWORK_OPTION_GET_PTP_LAST_TX_TIMESTAMP, &sync_tx_ts, &ts_size);

    if (ts_size != sizeof(sync_tx_ts))
    {
        ret = NET_ERR_NOT_SUPPORTED;
    }

    msg.preciseOriginTimestamp.nanoseconds = sync_tx_ts.nanoseconds;
    msg.preciseOriginTimestamp.seconds = sync_tx_ts.seconds;

    msg.tlv.tlvType = 0x03;
    msg.tlv.lengthField = 28;
    msg.tlv.organizationId[0] = 0;
    msg.tlv.organizationId[1] = 0x80;
    msg.tlv.organizationId[2] = 0xC2;
    msg.tlv.organizationSubType[0]=0;
    msg.tlv.organizationSubType[1]=0;
    msg.tlv.organizationSubType[2]=1;

    ret =  socket_send(runtime_conf.socket_id, &ptp_mac_addr, (uint8_t *)&msg, sizeof(msg));

    return ret;
}

static network_error_t send_ptp_pdelay_req_message(void)
{
    runtime_state.sequenceId_delay_req ++;

    ptp_pdelay_req_msg_t msg;
    msg.header = default_header;
    msg.header.sequenceId = runtime_state.sequenceId_delay_req;
    msg.header.messageLength = sizeof(ptp_pdelay_req_msg_t);
    msg.header.msg_type = PTP_MSG_TYPE_PDELAY_REQ;
    msg.header.logMessageInterval = 0x7F;
    msg.header.flagField.raw = 0;
    msg.header.flagField.twoStepFlag = 1;
    msg.header.correctionField = 0;
    msg.header.controlField = 0x05;

    msg.originTimestamp.nanoseconds = 0;
    msg.originTimestamp.seconds = 0;

    network_error_t ret =  socket_send(runtime_conf.socket_id, &ptp_mac_addr, (uint8_t *)&msg, sizeof(msg));

    return ret;
}

static network_error_t send_ptp_pdelay_resp_follow_up_message(void)
{
    ptp_pdelay_resp_follow_up_msg_t msg;
    msg.header = default_header;
    msg.header.sequenceId = runtime_state.pdelay_response.sequenceId_delay_req_rcv;
    msg.header.messageLength = sizeof(ptp_pdelay_resp_follow_up_msg_t);
    msg.header.msg_type = PTP_MSG_TYPE_PDELAY_RESP_FOLLOW_UP;
    msg.header.logMessageInterval = 0x7F;
    msg.header.flagField.raw = 0;
    msg.header.flagField.twoStepFlag = 1;
    msg.header.correctionField = runtime_state.pdelay_response.correctionField_rcv;
    msg.header.controlField = 0x05;

    time_stamp_t pdelay_resp_tx_ts;
    uint32_t ts_size;
    network_error_t ret = network_get_option(runtime_conf.netw_if,
                                             NETWORK_OPTION_GET_PTP_LAST_TX_TIMESTAMP, &pdelay_resp_tx_ts, &ts_size);

    if (ts_size != sizeof(pdelay_resp_tx_ts))
    {
        ret = NET_ERR_NOT_SUPPORTED;
    }

    msg.header.correctionField += pdelay_resp_tx_ts.nanoseconds;
    msg.responseOriginTimestamp.nanoseconds = pdelay_resp_tx_ts.nanoseconds;
    msg.responseOriginTimestamp.seconds = pdelay_resp_tx_ts.seconds;
    msg.requestingPortIdentity = runtime_state.pdelay_response.requestingPortIdentity;

    ret =  socket_send(runtime_conf.socket_id, &ptp_mac_addr, (uint8_t *)&msg, sizeof(msg));

    return ret;
}

static network_error_t send_ptp_announce_message(void)
{
    runtime_state.sequenceId_announce ++;

    ptp_announce_msg_t msg;
    msg.header = default_header;
    msg.header.sequenceId = runtime_state.sequenceId_announce;
    msg.header.messageLength = sizeof(ptp_announce_msg_t);
    msg.header.msg_type = PTP_MSG_TYPE_ANNOUNCE;
    msg.header.logMessageInterval = runtime_conf.logAnnounceInterval;
    msg.header.flagField.raw = 0;
    msg.header.correctionField = 0;
    msg.header.controlField = 0x05;

    msg.originTimestamp.nanoseconds = 0;
    msg.originTimestamp.seconds = 0;
    msg.currentUtcOffset = 0;
    msg.grandmasterClockQuality.clockAccuracy = runtime_conf.clock_accuracy;
    msg.grandmasterClockQuality.clockClass = runtime_conf.clockClass;
    msg.grandmasterClockQuality.offsetScaledLogVariance = runtime_conf.offsetScaledLogVariance;
    msg.timeSource = runtime_conf.timeSource;
    msg.grandmasterPriority1 = runtime_conf.priority1;
    msg.grandmasterPriority2 = runtime_conf.priority2;
    msg.stepsRemoved = 1;
    msg.grandmasterIdentity = runtime_conf.host_port_identity.clockIdentity;

    network_error_t ret =  socket_send(runtime_conf.socket_id, &ptp_mac_addr, (uint8_t *)&msg, sizeof(msg));

    return ret;
}

static network_error_t run_bmca(void)
{
    if(runtime_conf.forced_state == TIME_SYNC_FORCED_STATE_SLAVE)
    {
        runtime_state.current_ptp_state = PTP_PORT_STATE_SLAVE;
    }
    else if (runtime_state.announce_data.timeSource == PTP_TIME_SOURCE_NONE)
    {
        runtime_state.current_ptp_state = PTP_PORT_STATE_MASTER;
    }
    // simplied decision algorithm for single port devices
    else if (runtime_state.announce_data.grandmasterPriority1 < runtime_conf.priority1)
    {
        // foreign better
        runtime_state.current_ptp_state = PTP_PORT_STATE_SLAVE;
    }
    else if (runtime_state.announce_data.grandmasterPriority1 > runtime_conf.priority1)
    {
        // local better
        runtime_state.current_ptp_state = PTP_PORT_STATE_MASTER;
    }
    else if (runtime_state.announce_data.grandmasterClockQuality.clockClass < runtime_conf.clockClass)
    {
        // foreign better
        runtime_state.current_ptp_state = PTP_PORT_STATE_SLAVE;
    }
    else if (runtime_state.announce_data.grandmasterClockQuality.clockClass > runtime_conf.clockClass)
    {
        // local better
        runtime_state.current_ptp_state = PTP_PORT_STATE_MASTER;
    }
    else if (runtime_state.announce_data.grandmasterClockQuality.clockAccuracy < runtime_conf.clock_accuracy)
    {
        // foreign better
        runtime_state.current_ptp_state = PTP_PORT_STATE_SLAVE;
    }
    else if (runtime_state.announce_data.grandmasterClockQuality.clockAccuracy > runtime_conf.clock_accuracy)
    {
        // local better
        runtime_state.current_ptp_state = PTP_PORT_STATE_MASTER;
    }
    else if (runtime_state.announce_data.grandmasterClockQuality.offsetScaledLogVariance < runtime_conf.offsetScaledLogVariance)
    {
        // foreign better
        runtime_state.current_ptp_state = PTP_PORT_STATE_SLAVE;
    }
    else if (runtime_state.announce_data.grandmasterClockQuality.offsetScaledLogVariance > runtime_conf.offsetScaledLogVariance)
    {
        // local better
        runtime_state.current_ptp_state = PTP_PORT_STATE_MASTER;
    }
    else if (runtime_state.announce_data.grandmasterPriority2 < runtime_conf.priority2)
    {
        // foreign better
        runtime_state.current_ptp_state = PTP_PORT_STATE_SLAVE;
    }
    else if (runtime_state.announce_data.grandmasterPriority2 > runtime_conf.priority2)
    {
        // local better
        runtime_state.current_ptp_state = PTP_PORT_STATE_MASTER;
    }
    else
    {
#pragma loopbound min 1 max 8
        for (uint8_t i = 0; i < 8; i++)
        {
            if (runtime_state.announce_data.grandmasterIdentity.id[i] < runtime_conf.host_port_identity.clockIdentity.id[i])
            {
                // foreign better
                runtime_state.current_ptp_state = PTP_PORT_STATE_SLAVE;
            }
            else if (runtime_state.announce_data.grandmasterIdentity.id[i] > runtime_conf.host_port_identity.clockIdentity.id[i])
            {
                // local better
                runtime_state.current_ptp_state = PTP_PORT_STATE_MASTER;
            }
        }
    }
    return NET_ERR_OK;
}



static void correct_runtime_timestamps(time_interval_t *offset)
{
    // correct intervals
    time_stamp_t current_ts;
    uint32_t ts_size;
    network_error_t ret = network_get_option(runtime_conf.netw_if,
                             NETWORK_OPTION_GET_PTP_CLOCK, &current_ts, &ts_size);

    if (ret != NET_ERR_OK)
    {
        return;
    } 
    else if (ts_size != sizeof(current_ts))
    {
        ret = NET_ERR_NOT_SUPPORTED;
    }
    
    runtime_state.next_announce.seconds = current_ts.seconds;
    runtime_state.next_announce.nanoseconds = current_ts.nanoseconds ;

    runtime_state.next_sync.seconds = current_ts.seconds;
    runtime_state.next_sync.nanoseconds = current_ts.nanoseconds ;

    runtime_state.next_delay_req.seconds = current_ts.seconds;
    runtime_state.next_delay_req.nanoseconds = current_ts.nanoseconds;
}

void time_sync_get_runtime_data(time_interval_t *offfset, time_interval_t* mean_path_delay)
{
    *offfset = runtime_state.offset;
    *mean_path_delay = runtime_state.mean_path_delay;
}

