/*
 * Socket Module to send for network interaction 
 * 
 * Authors: Lehr Christoph
 */

#include <stdlib.h>
#include <stdio.h>

#include "../inc/socket.h"
#include "../inc/network.h"
#include "../inc/utils.h"
#include "protocols/udp.h"
#include "protocols/tcp.h"
#include "protocols/ipv4.h"
#include "protocols/arp.h"
#include "protocols/ethernet.h"

//////////////////////////////////////////////////////////////////

static socket_t config_array[MAX_SOCKETS];
static uint16_t config_wr_ptr=0;
static uint8_t global_logging=0;
static uint8_t accept_l2_broadcasts=0;
static uint8_t accept_ip_broadcasts=0;
static uint8_t missing_l2_no_fail=0;

network_error_t send_ethernet_socket(unsigned short socket_id, socket_address_t * dest_addr, unsigned char* data, unsigned short len);
network_error_t send_ip_socket(unsigned short socket_id, socket_address_t * dest_addr, unsigned char* data, unsigned short len);
network_error_t socket_handle_mac_layer(uint16_t net_id, uint8_t* data, uint16_t *len, ethertype_t *protocol, socket_address_t *src_addr,socket_address_t* dst_addr, socket_address_t* host_addr, uint16_t *buffer_offset, time_stamp_t* rx_ts);
network_error_t socket_handle_network_layer(uint16_t net_id, uint8_t* data, uint16_t len, ethertype_t l3_protocol, ip_protocol_t *l4_protocol, socket_address_t *src_addr,socket_address_t* dst_addr, socket_address_t* host_addr, uint16_t *buffer_offset, uint16_t *payload_length);
network_error_t socket_handle_transport_layer(uint8_t* data, uint16_t len, ip_protocol_t l4_protocol, socket_address_t *src_addr,socket_address_t* dst_addr, socket_address_t* host_addr, uint16_t *buffer_offset, time_stamp_t* rx_ts, uint16_t* payload_length);
static network_error_t system_send(uint16_t net_id, socket_address_t *src_addr, socket_address_t *dst_addr, ip_protocol_t l3_protocol, ethertype_t l2_protocol, uint8_t* data, uint16_t data_len, uint16_t header_len );

static void init_rcv_buffers(uint16_t socket_id);
static network_error_t socket_enque_msg( uint16_t socket_id, socket_address_t *src_address, uint8_t* frame_start, uint8_t* data, uint16_t length, time_stamp_t* rx_ts);

// request a socket
unsigned short socket(socket_domain_t domain, socket_type_t type, unsigned short protocol)
{
    // find next free entry
    char found =0;
    uint16_t socket_id; 
    #pragma loopbound min 1 max MAX_SOCKETS 
    for(unsigned short i=config_wr_ptr; i < MAX_SOCKETS; i++)
    {
        if(config_array[i].domain == SOCKET_DOMAIN_NONE)
        {
            config_array[i].domain = SOCKET_DOMAIN_CONFIG;
            config_wr_ptr = i;
            socket_id = i;
            found =1;

            break;
        }
    }

    // check if buffer was found
    if(found == 0 )
    {
        #pragma loopbound min 0 max MAX_NETWORK_IF_CONFIGS 
        for(unsigned short i=0; i < config_wr_ptr; i++)
        {
            if(config_array[i].domain == SOCKET_DOMAIN_NONE)
            {
                config_array[i].domain = SOCKET_DOMAIN_CONFIG;
                config_wr_ptr = i;
                socket_id = i;
                found =1;
                break;
            }
        }   
        // check if buffer was found
        if(found == 0)
        {
            return ~0;
        }
    }

    if(domain == SOCKET_DOMAIN_IP )
    {
        if(type == SOCKET_TYPE_DATAGRAM)
        {
            config_array[socket_id].ip.protocol = IP_PROTOCOL_UDP;
        }
        else if (type == SOCKET_TYPE_STREAM)
        {
            config_array[socket_id].ip.protocol = IP_PROTOCOL_TCP;
        }
        else if( type == SOCKET_TYPE_RAW)
        {
            config_array[socket_id].ip.protocol = protocol;
        }
        else
        {
            config_array[socket_id].domain = SOCKET_DOMAIN_NONE;
            return ~0;
        }
        // initialise source port with random value
        uint16_t port = rand() % (MAX_PORT_ID-MIN_PORT_ID);
        port += MIN_PORT_ID;
        config_array[socket_id].ip.src_port = port;
    }
    else if(domain == SOCKET_DOMAIN_ETHERNET )
    {
        if( type == SOCKET_TYPE_RAW)
        {
            config_array[socket_id].eth.protocol = protocol;
            
        }
        else
        {
            config_array[socket_id].domain = SOCKET_DOMAIN_NONE;
            return ~0;
        }
    }

    config_array[socket_id].domain = domain;
    config_array[socket_id].type = type;
    config_array[socket_id].bound_interface = GLOBAL_SOCKET;

    config_array[socket_id].options.priority_set = 0;
    return socket_id;
}

// network_error_t socket_connect(unsigned short socket_id, socket_address_t * dest_addr)
// {
//     if(socket_id >= MAX_SOCKETS || 
//         config_array[socket_id].domain == SOCKET_DOMAIN_NONE)
//     {
//         return NET_ERR_SOCKET_UNKNOWN;
//     }

//     if(config_array[socket_id].ip.protocol != IP_PROTOCOL_TCP)
//     {
//         return NET_ERR_NOT_SUPPORTED;
//     }

//     network_frame_t frame;
//     frame.user_data_length=0;

//     network_error_t ret;
//     src_addr.port = config_array[socket_id].ip.src_port;
//     if(config_array[socket_id].bound_interface < MAX_SOCKETS)
//     {
//         net_id = config_array[socket_id].bound_interface;
//         if(config_array[socket_id].options.debug_output || global_logging)
//         {
//             utils_log("Using bound interface %u\n", net_id);
//         }
//     }
//     else
//     {
//         if(config_array[socket_id].options.debug_output || global_logging)
//         {
//             utils_log("Searching for interface with IP ");
//             utils_print_ip(&dest_addr->ip_address);
//             utils_log("\n");
//         }
//         net_id = network_get_interface_by_ip_netmask(&dest_addr->ip_address);
//         if(config_array[socket_id].options.debug_output || global_logging)
//         {
//             utils_log("Found interface %u\n", net_id);
//         }
//     }
    
//     // l2 configuration is required in all cases
//     ret = network_get_l2_config(net_id, &l2_config);
//     if(ret != NET_ERR_OK)
//     {
//         if(config_array[socket_id].options.debug_output || global_logging)
//         {
//             utils_log("Error when retreiving L2 configuration of Interface %u: 0x%04hx\n", net_id, ret);
//         }

//         return ret;
//     }
    
//     // calculate required space for OSI layer 2 protocols
//     if(l2_config.l2_type == NETWORK_L2_ETHERNET)
//     {
//         header_length += sizeof(ethernet_header_t);
//     }
//     else if(l2_config.l2_type == NETWORK_L2_ETHERNET_VLAN)
//     {
//         header_length += sizeof(ethernet_header_vlan_t);
//     }

//     // Check for ARP Address
//     if(dest_addr->ip_address.version == IP_VERSION_4)
//     {
//         if(dest_addr->hw_address.length == 0)
//         {
//             if(config_array[socket_id].options.debug_output || global_logging)
//             {
//                 utils_log("Checking ARP-Table for IP ");
//                 utils_print_ip(&dest_addr->ip_address);
//                 utils_log("\n");
//             }
            
//             ret = arp_table_search(&dest_addr->ip_address, &dest_addr->hw_address);
//             if(ret != NET_ERR_OK)
//             {
//                 if(config_array[socket_id].options.debug_output || global_logging)
//                 {
//                     utils_log("No MAC Address available for IP ");
//                     utils_print_ip(&dest_addr->ip_address);
//                     utils_log("\n");
//                 }

//                 // send arp request
//                 arp_msg_t request;
//                 ret = arp_build_request(&dest_addr->ip_address, &src_addr.ip_address, &src_addr.hw_address, &request);
//                 if(ret != NET_ERR_OK)
//                 {
//                     if(config_array[socket_id].options.debug_output || global_logging)
//                     {
//                         utils_log("Error while building ARP-Request : 0x%04hx\n", ret);
//                     }

//                     return ret;
//                 }

//                 ret = system_send(net_id, &src_addr, dest_addr, IP_PROTOCOL_RESERVED, ETHERTYPE_ARP, (uint8_t*) &request, sizeof(arp_msg_t), header_length);

//                 if(ret != NET_ERR_OK)
//                 {
//                     if(config_array[socket_id].options.debug_output || global_logging)
//                     {
//                         utils_log("Error while sending ARP-Request : 0x%04hx\n", ret);
//                     }

//                     return ret;
//                 }

//                 if(missing_l2_no_fail == 0)
//                 {
//                     return NET_ERR_ARP_UNKNOWN_ADDR;
//                 }
//             }
//         }
//         header_length += sizeof(ipv4_header_t);
//     }
//     else if(dest_addr->ip_address.version == IP_VERSION_6)
//     {
//         // ndp check
//         // get ipv6 header + extension header length
//         // add to header length
//         return NET_ERR_NOT_SUPPORTED;
//     }
    
//     // create Ethernet Header
// }

// bind the socket to send and receive i.e. via TCP/UDP Port
// or bind socket to network interface for L2 communication
network_error_t socket_bind(unsigned short socket_id, socket_address_t * sock_addr)
{
    if(socket_id >= MAX_SOCKETS || 
        config_array[socket_id].domain == SOCKET_DOMAIN_NONE)
    {
        return NET_ERR_SOCKET_UNKNOWN;
    }

    if(config_array[socket_id].domain == SOCKET_DOMAIN_IP)
    {
        config_array[socket_id].ip.src_port = sock_addr->port;
        uint16_t net_id = network_get_interface_by_ip_netmask(&sock_addr->ip_address);
        if(net_id < MAX_NETWORK_IF_CONFIGS)
        {
            config_array[socket_id].bound_interface = net_id;
        }
    }
    else if (config_array[socket_id].domain == SOCKET_DOMAIN_ETHERNET)
    {
        config_array[socket_id].eth.hw_address = sock_addr->hw_address;
        config_array[socket_id].bound_interface = sock_addr->if_id;
    }
    else
    {
        return NET_ERR_SOCKET;
    }

    config_array[socket_id].recv_msg_queue.length = SOCKET_RECV_QUEUE_LENGTH;
    config_array[socket_id].recv_msg_queue.read_ptr=0;
    config_array[socket_id].recv_msg_queue.write_ptr=0;

    void *memory_pool = network_get_buffer(sizeof(socket_msg_queue_entry_t) * SOCKET_RECV_QUEUE_LENGTH);
    if(memory_pool == NULL)
    {
        return NET_ERR_MGMT_MEM_FULL;
    }
    config_array[socket_id].recv_msg_queue.queue=(socket_msg_queue_entry_t*) memory_pool;

    init_rcv_buffers(socket_id);

    if(config_array[socket_id].type == SOCKET_TYPE_STREAM)
    {
        tcp_init_server(&config_array[socket_id], socket_id);
    }

    return NET_ERR_OK;
}

network_error_t socket_bind_if(unsigned short socket_id, unsigned short interface_id)
{
    if(socket_id >= MAX_SOCKETS || 
        config_array[socket_id].domain == SOCKET_DOMAIN_NONE)
    {
        return NET_ERR_SOCKET_UNKNOWN;
    }

    if(config_array[socket_id].domain != SOCKET_DOMAIN_ETHERNET)
    {
        return NET_ERR_NOT_SUPPORTED;
    }

    network_error_t ret = network_interface_exists(interface_id);
    if(ret != NET_ERR_OK)
    {
        return ret;
    }
    config_array[socket_id].bound_interface = interface_id;
    
    config_array[socket_id].recv_msg_queue.length = SOCKET_RECV_QUEUE_LENGTH;
    config_array[socket_id].recv_msg_queue.read_ptr=0;
    config_array[socket_id].recv_msg_queue.write_ptr=0;

    void *memory_pool = network_get_buffer(sizeof(socket_msg_queue_entry_t) * SOCKET_RECV_QUEUE_LENGTH);
    if(memory_pool == NULL)
    {
        return NET_ERR_MGMT_MEM_FULL;
    }
    config_array[socket_id].recv_msg_queue.queue=(socket_msg_queue_entry_t*) memory_pool;
    
    init_rcv_buffers(socket_id);

    return NET_ERR_OK;
}


static void init_rcv_buffers(uint16_t socket_id)
{
    uint8_t *mem_pool = (uint8_t *) config_array[socket_id].recv_msg_queue.queue;

    // Not sure of this line, shouldn't be sizeof(socket_msg_queue_entry_t)
    for(uint8_t i=0; i < config_array[socket_id].recv_msg_queue.length*sizeof(socket_msg_queue_t) ; i++)
    {
        mem_pool[i] =0;
    }
}

// use socket id 0xFFFF for global options
network_error_t socket_set_option(unsigned short socket_id, socket_options_t option, void *value, uint32_t value_size)
{
    // Set globla options
    if( socket_id == GLOBAL_SOCKET )
    {
        if(option == SOCKET_OPTION_DEBUG_LOG)
        {
            uint8_t* val = (uint8_t*) value;
            global_logging = *val;
        }
        else if(option == SOCKET_OPTION_ACCEPT_L2_BRODCASTS)
        {
            uint8_t* val = (uint8_t*) value;
            accept_l2_broadcasts = *val;
        }
        else if(option == SOCKET_OPTION_ACCEPT_IP_BRODCASTS)
        {
            uint8_t* val = (uint8_t*) value;
            accept_ip_broadcasts = *val;
        }
        else if(option == SOCKET_OPTION_MISSING_L2_ADDRESS_NO_FAIL)
        {
            uint8_t* val = (uint8_t*) value;
            missing_l2_no_fail= *val;
        }
        else
        {
            return NET_ERR_INPUT_VALUE;
        }
        return NET_ERR_OK;
    }

    if(socket_id >= MAX_SOCKETS || 
        config_array[socket_id].domain == SOCKET_DOMAIN_NONE)
    {
        return NET_ERR_SOCKET_UNKNOWN;
    }

    if(option == SOCKET_OPTION_PRIORITY)
    {
        if(value_size > sizeof(uint8_t))
        {
            return NET_ERR_INPUT_VALUE;
        }
        uint8_t* option_value = (uint8_t * )value;
        *option_value &= 0xFF;
        config_array[socket_id].options.priority_set =1;
        config_array[socket_id].options.priority = *option_value;
    }
    else if(option == SOCKET_OPTION_BLOCKING)
    {
        config_array[socket_id].options.blocking =1;
    }
    else if(option == SOCKET_OPTION_NON_BLOCK)
    {
        config_array[socket_id].options.blocking =0;
    }
    else if(option == SOCKET_OPTION_DEBUG_LOG)
    {
        uint8_t* val = (uint8_t*) value;
        config_array[socket_id].options.debug_output = *val;
    }
    else
    {
        return NET_ERR_INPUT_VALUE;
    }

    return NET_ERR_OK;
}

// send data to a given destination address
network_error_t socket_send(unsigned short socket_id, socket_address_t * dest_addr, unsigned char* data, unsigned short len)
{
    if(socket_id >= MAX_SOCKETS || 
        config_array[socket_id].domain == SOCKET_DOMAIN_NONE)
    {
        return NET_ERR_SOCKET_UNKNOWN;
    }

    network_error_t ret;

    if(config_array[socket_id].domain == SOCKET_DOMAIN_IP)
    {
        ret =send_ip_socket(socket_id, dest_addr, data, len);
    }
    else if(config_array[socket_id].domain == SOCKET_DOMAIN_ETHERNET)
    {
        // handle Raw Ethernet Frame
        // determine mac layer
        // call driver
        ret =send_ethernet_socket(socket_id, dest_addr, data, len);
    }
    else
    {
        ret = NET_ERR_NOT_SUPPORTED;
    }

    return ret;
}

network_error_t send_ethernet_socket(unsigned short socket_id, socket_address_t * dest_addr, unsigned char* data, unsigned short len)
{
    uint16_t net_id;
    uint16_t header_length = 0;
    uint16_t buffer_offset = 0;
    network_layer2_config_t l2_config;
    socket_address_t src_addr;        

    network_frame_t frame;
    frame.user_data = data;
    frame.user_data_length=len;

    network_error_t ret;
    
    // if(config_array[socket_id].bound_interface < MAX_SOCKETS)
    if(config_array[socket_id].bound_interface < MAX_NETWORK_IF_CONFIGS)
    {
        net_id = config_array[socket_id].bound_interface;
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Using bound interface %u\n", net_id);
        }
    }
    else
    {
        return NET_ERR_SOCKET_NOT_BOUND;
    }
    
    // l2 configuration is required in all cases
    ret = network_get_l2_config(net_id, &l2_config);
    if(ret != NET_ERR_OK)
    {
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Error when retreiving L2 configuration of Interface %u: 0x%04hx\n", net_id, ret);
        }

        return ret;
    }
    
    // calculate required space for OSI layer 2 protocols
    if(l2_config.l2_type == NETWORK_L2_ETHERNET)
    {
        header_length = sizeof(ethernet_header_t);
    }
    else if(l2_config.l2_type == NETWORK_L2_ETHERNET_VLAN)
    {
        header_length = sizeof(ethernet_header_vlan_t);
    }
    
    ret = network_get_mac_address(net_id, &src_addr.hw_address);
    if(ret != NET_ERR_OK)
    {
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Error when retreiving MAC Address of Interface %u: 0x%04hx\n", net_id, ret);
        }

        return ret;
    }


    uint8_t buffer[header_length];
    buffer_offset = header_length;
    if(config_array[socket_id].options.debug_output || global_logging)
    {
        utils_log("Required Header Space: %u\n", header_length);
    }

    // create layer 2 protocol headers
    if(l2_config.l2_type == NETWORK_L2_ETHERNET || l2_config.l2_type == NETWORK_L2_ETHERNET_VLAN)
    {
        ethertype_t protocol = config_array[socket_id].eth.protocol;
      
        if(l2_config.l2_type == NETWORK_L2_ETHERNET)
        {
            buffer_offset -=  sizeof(ethernet_header_t);
            ret = ethernet_build_frame(&dest_addr->hw_address, &src_addr.hw_address, protocol, &buffer[buffer_offset]);
        }
        else if(l2_config.l2_type == NETWORK_L2_ETHERNET_VLAN)
        {
            buffer_offset -= sizeof(ethernet_header_vlan_t);
            if(config_array[socket_id].options.priority_set == 1)
            {
                l2_config.vlan.pcp = config_array[socket_id].options.priority;
            }
            ret = ethernet_build_vlan_frame(&dest_addr->hw_address, &src_addr.hw_address, protocol, &l2_config.vlan, &buffer[buffer_offset]);
        }

        if(ret != NET_ERR_OK)
        {
            if(config_array[socket_id].options.debug_output || global_logging)
            {
                utils_log("Error while building MAC Frame : 0x%04hx\n", ret);
            }
            return ret;
        }
    }
    
    frame.header_length = header_length;
    frame.header_data = &buffer;
    
    if(config_array[socket_id].options.priority_set == 1)
    {
        frame.priority = config_array[socket_id].options.priority;
    }
    else
    {
        frame.priority = 0;
    }

    if(config_array[socket_id].options.debug_output || global_logging)
    {
        utils_log("handover to network interfacex\n");
    }
    // handover to network interface
    return network_send(net_id, &frame, config_array[socket_id].options.blocking);
}


network_error_t send_ip_socket(unsigned short socket_id, socket_address_t * dest_addr, unsigned char* data, unsigned short len)
{
    uint16_t net_id;
    uint16_t header_length = 0;
    uint16_t buffer_offset = 0;
    network_layer2_config_t l2_config;
    socket_address_t src_addr;        

    network_frame_t frame;
    frame.user_data = data;
    frame.user_data_length=len;

    network_error_t ret;
    src_addr.port = config_array[socket_id].ip.src_port;

    // Look for the network interface ID
    // if(config_array[socket_id].bound_interface < MAX_SOCKETS)
    if(config_array[socket_id].bound_interface < MAX_NETWORK_IF_CONFIGS)
    {
        net_id = config_array[socket_id].bound_interface;
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Using bound interface %u\n", net_id);
        }
    }
    else
    {
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Searching for interface with IP ");
            utils_print_ip(&dest_addr->ip_address);
            utils_log("\n");
        }
        net_id = network_get_interface_by_ip_netmask(&dest_addr->ip_address);
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Found interface %u\n", net_id);
        }
        if(net_id > MAX_NETWORK_IF_CONFIGS) {
            return NET_ERR_SOCKET_NOT_BOUND;
        }
    }
    
    // l2 configuration is required in all cases
    ret = network_get_l2_config(net_id, &l2_config);
    if(ret != NET_ERR_OK)
    {
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Error when retreiving L2 configuration of Interface %u: 0x%04hx\n", net_id, ret);
        }

        return ret;
    }
    
    // calculate required space for OSI layer 2 protocols
    if(l2_config.l2_type == NETWORK_L2_ETHERNET)
    {
        header_length += sizeof(ethernet_header_t);
    }
    else if(l2_config.l2_type == NETWORK_L2_ETHERNET_VLAN)
    {
        header_length += sizeof(ethernet_header_vlan_t);
    }
    
    ret = network_get_interface_ip_address(net_id, &src_addr.ip_address);
    if(ret != NET_ERR_OK)
    {
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Error when retreiving IP Address of Interface %u: 0x%04hx\n", net_id, ret);
        }

        return ret;
    }

    ret = network_get_mac_address(net_id, &src_addr.hw_address);
    if(ret != NET_ERR_OK)
    {
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Error when retreiving MAC Address of Interface %u: 0x%04hx\n", net_id, ret);
        }

        return ret;
    }

    if(dest_addr->ip_address.version == IP_VERSION_4)
    {
        if(dest_addr->hw_address.length == 0)
        {
            if(config_array[socket_id].options.debug_output || global_logging)
            {
                utils_log("Checking ARP-Table for IP ");
                utils_print_ip(&dest_addr->ip_address);
                utils_log("\n");
            }
            
            ret = arp_table_search(&dest_addr->ip_address, &dest_addr->hw_address);
            if(ret != NET_ERR_OK)
            {
                if(config_array[socket_id].options.debug_output || global_logging)
                {
                    utils_log("No MAC Address available for IP ");
                    utils_print_ip(&dest_addr->ip_address);
                    utils_log("\n");
                }

                // send arp request
                arp_msg_t request;
                ret = arp_build_request(&dest_addr->ip_address, &src_addr.ip_address, &src_addr.hw_address, &request);
                if(ret != NET_ERR_OK)
                {
                    if(config_array[socket_id].options.debug_output || global_logging)
                    {
                        utils_log("Error while building ARP-Request : 0x%04hx\n", ret);
                    }

                    return ret;
                }

                ret = system_send(net_id, &src_addr, dest_addr, IP_PROTOCOL_RESERVED, ETHERTYPE_ARP, (uint8_t*) &request, sizeof(arp_msg_t), header_length);

                if(ret != NET_ERR_OK)
                {
                    if(config_array[socket_id].options.debug_output || global_logging)
                    {
                        utils_log("Error while sending ARP-Request : 0x%04hx\n", ret);
                    }

                    return ret;
                }

                if(missing_l2_no_fail == 0)
                {
                    return NET_ERR_ARP_UNKNOWN_ADDR;
                }
            }
        }
        header_length += sizeof(ipv4_header_t);
    }
    else if(dest_addr->ip_address.version == IP_VERSION_6)
    {
        // ndp check
        // get ipv6 header + extension header length
        // add to header length
        return NET_ERR_NOT_SUPPORTED;
    }
    else {
        return NET_ERR_NOT_SUPPORTED;
    }

    if (config_array[socket_id].type == SOCKET_TYPE_DATAGRAM)
    {
        header_length += sizeof(udp_header_t);

        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Building UDP Packet:\n");
            utils_log("\tSrc IP: ");
            utils_print_ip(&src_addr.ip_address);
            utils_log(" MAC:");
            utils_print_mac(&src_addr.hw_address);
            utils_log(" Port: %u \n\tDst IP: ", src_addr.port);
            utils_print_ip(&dest_addr->ip_address);
            utils_log(" MAC:");
            utils_print_mac(&dest_addr->hw_address);
            utils_log(" Port: %u \n", dest_addr->port);
        }
    }
    else if (config_array[socket_id].type == SOCKET_TYPE_STREAM)
    {
        header_length += sizeof(tcp_header_t);
    }

    uint8_t buffer[header_length];
    buffer_offset = header_length;
    if(config_array[socket_id].options.debug_output || global_logging)
    {
        utils_log("Required Header Space: %u\n", header_length);
    }

    // handle in IP encapsulated protocol
    if (config_array[socket_id].type == SOCKET_TYPE_DATAGRAM)
    {
        // handle UDP Header
        buffer_offset -= sizeof(udp_header_t);
        ret = udp_build_datagram(
            &dest_addr->ip_address, dest_addr->port, 
            &src_addr.ip_address,  src_addr.port,
            data, len, &buffer[buffer_offset]);

        if(ret != NET_ERR_OK)
        {
            if(config_array[socket_id].options.debug_output || global_logging)
            {
                utils_log("Error while building UDP Datagram : 0x%04hx\n", ret);
            }
            return ret;
        }
    }
    else if(config_array[socket_id].type == SOCKET_TYPE_STREAM)
    {
        uint8_t buffer_tcp_hdr[sizeof(tcp_header_t)];
        buffer_offset -= sizeof(tcp_header_t);

        ret = tcp_build_datagram(&dest_addr->ip_address, 
            dest_addr->port, &src_addr.ip_address, src_addr.port, 
            &data, &len, buffer_tcp_hdr);
        if(ret != NET_ERR_OK)
        {
            if(config_array[socket_id].options.debug_output || global_logging)
            {
                utils_log("Error while building TCP Datagram : 0x%04hx\n", ret);
            }
            return ret;
        }
        memcpy(&buffer[buffer_offset], buffer_tcp_hdr, sizeof(tcp_header_t));
        frame.user_data_length = len;
        frame.user_data = data;
    }
    else
    {
        return NET_ERR_SOCKET_UNKNOWN;
    }

    // create IP header
    if(dest_addr->ip_address.version == IP_VERSION_4)
    {
        uint8_t buffer_ipv4_hdr[sizeof(ipv4_header_t)];
        uint16_t ip_encapsulated_length = (header_length-buffer_offset)+ len;
        buffer_offset -= sizeof(ipv4_header_t);
        ret = ipv4_build_datagram(&dest_addr->ip_address, &src_addr.ip_address, 
                        config_array[socket_id].ip.protocol, buffer_ipv4_hdr, ip_encapsulated_length);
        
        if(ret != NET_ERR_OK)
        {
            if(config_array[socket_id].options.debug_output || global_logging)
            {
                utils_log("Error while building IP Datagram : 0x%04hx\n", ret);
            }
            return ret;
        }
        memcpy(&buffer[buffer_offset], buffer_ipv4_hdr, sizeof(ipv4_header_t));
    }
    
    // create layer 2 protocol headers
    if(l2_config.l2_type == NETWORK_L2_ETHERNET || l2_config.l2_type == NETWORK_L2_ETHERNET_VLAN)
    {
        ethertype_t protocol;
        if(dest_addr->ip_address.version == IP_VERSION_4)
        {
            protocol = ETHERTYPE_IPv4;
        }
        else if(dest_addr->ip_address.version == IP_VERSION_6)
        {
            protocol = ETHERTYPE_IPv6;
        }
        
        if(l2_config.l2_type == NETWORK_L2_ETHERNET)
        {
            buffer_offset -=  sizeof(ethernet_header_t);
            ret = ethernet_build_frame(&dest_addr->hw_address, &src_addr.hw_address, protocol, &buffer[buffer_offset]);
        }
        else if(l2_config.l2_type == NETWORK_L2_ETHERNET_VLAN)
        {
            buffer_offset -= sizeof(ethernet_header_vlan_t);
            if(config_array[socket_id].options.priority_set == 1)
            {
                l2_config.vlan.pcp = config_array[socket_id].options.priority;
            }
            ret = ethernet_build_vlan_frame(&dest_addr->hw_address, &src_addr.hw_address, protocol, &l2_config.vlan, &buffer[buffer_offset]);
        }

        if(ret != NET_ERR_OK)
        {
            if(config_array[socket_id].options.debug_output || global_logging)
            {
                utils_log("Error while building MAC Frame : 0x%04hx\n", ret);
            }
            return ret;
        }
    }
    
    frame.header_length = header_length;
    frame.header_data = &buffer[buffer_offset];
    
    if(config_array[socket_id].options.priority_set == 1)
    {
        frame.priority = config_array[socket_id].options.priority;
    }
    else
    {
        frame.priority = 0;
    }
    
    if(config_array[socket_id].options.debug_output || global_logging)
    {
        utils_log("handover to network interface\n");
    }

    // handover to network interface
    return network_send(net_id, &frame, config_array[socket_id].options.blocking);
}

// send data to a given destination address
network_error_t socket_handle_frame(uint16_t net_id, uint8_t* data, uint16_t len, time_stamp_t *rx_ts)
{
    network_error_t ret;
    socket_address_t src_addr, dst_addr, host_addr;
    uint16_t buffer_offset = 0;
    ethertype_t l3_protocol;
    ip_protocol_t l4_protocol;
    uint16_t payload_lenght;

    if(global_logging)
    {
        utils_log("Received Frame from Interface : %u\n", net_id);
    }

    ret = network_get_interface_ip_address(net_id, &host_addr.ip_address);
    if(ret != NET_ERR_OK)
    {
        if(global_logging)
        {
            utils_log("Error when retreiving IP Address of Interface %u: 0x%04hx\n", net_id, ret);
        }

        return ret;
    }

    ret = network_get_mac_address(net_id, &host_addr.hw_address);
    if(ret != NET_ERR_OK)
    {
        if( global_logging)
        {
            utils_log("Error when retreiving MAC Address of Interface %u: 0x%04hx\n", net_id, ret);
        }

        return ret;
    }

    ret = socket_handle_mac_layer(net_id, data, &len, &l3_protocol, &src_addr, &dst_addr, &host_addr, &buffer_offset, rx_ts);
    // buffer_offset == 0 -> packet has been processed in this layer
    if(ret != NET_ERR_OK || buffer_offset == 0)
    {
        if(global_logging && ret != NET_ERR_OK)
        {
            utils_log("MAC Layer error when parsing frame : 0x%04hx\n", ret);
        }
        return ret;
    }

    ret = socket_handle_network_layer(net_id, data, len, l3_protocol, &l4_protocol, &src_addr, &dst_addr, &host_addr, &buffer_offset, &payload_lenght);
    // buffer_offset == 0 -> packet has been processed in this layer
    if(ret != NET_ERR_OK || buffer_offset == 0)
    {
        if(global_logging && ret != NET_ERR_OK)
        {
            utils_log("Network Layer error when parsing frame : 0x%04hx\n", ret);
        }
        return ret;
    }

    ret= socket_handle_transport_layer(data, len, l4_protocol, &src_addr, &dst_addr, &host_addr, &buffer_offset, rx_ts, &payload_lenght);

    if(global_logging && ret != NET_ERR_OK)
    {
        utils_log("Transport Layer error when parsing frame : 0x%04hx\n", ret);
    }
    return ret;
}

network_error_t socket_handle_mac_layer(
    uint16_t net_id, uint8_t* data, uint16_t *len, ethertype_t *protocol, 
    socket_address_t *src_addr, socket_address_t* dst_addr, socket_address_t* host_addr, 
    uint16_t *buffer_offset, time_stamp_t *rx_ts)
{
    // l2 configuration is required in all cases
    network_layer2_config_t l2_config;
    network_error_t ret = network_get_l2_config(net_id, &l2_config);
    if(ret != NET_ERR_OK)
    {
        if(global_logging)
        {
            utils_log("Error when retreiving L2 configuration of Interface %u: 0x%04hx\n", net_id, ret);
        }

        return ret;
    }
    
    if(l2_config.l2_type == NETWORK_L2_ETHERNET || l2_config.l2_type == NETWORK_L2_ETHERNET_VLAN)
    {
        ret = ethernet_handle_frame(&data[*buffer_offset], *len - *buffer_offset, protocol,  &src_addr->hw_address, &dst_addr->hw_address);
        if(ret != NET_ERR_OK)
        {
            if(global_logging)
            {
                utils_log("Error while parsing Ethernet Frame: 0x%04hx\n", ret);
            }
            return ret;
        }

        if(*protocol == ETHERTYPE_VLAN)
        {
            ret = ethernet_handle_vlan_frame(&data[*buffer_offset], *len - *buffer_offset, protocol,  &src_addr->hw_address, &dst_addr->hw_address, &l2_config);
            if(ret != NET_ERR_OK)
            {
                if(global_logging)
                {
                    utils_log("Error while parsing VLAN Tag of Ethernet Frame: 0x%04hx\n", ret);
                }
                return ret;
            }
            *buffer_offset += sizeof(ethernet_header_vlan_t);
        }
        else
        {
            *buffer_offset += sizeof(ethernet_header_t);
        }

        // cut off FCS from Ethernet, no longer needed
        *len -= 4;

        if(global_logging)
        {
            utils_log("\tAccept Broadcasts: %i\n", accept_l2_broadcasts);
            utils_log("\tSrc MAC:  ");
            utils_print_mac(&src_addr->hw_address);
            utils_log("\n\tDest MAC: ");
            utils_print_mac(&dst_addr->hw_address);
            utils_log("\n\tHost MAC: ");
            utils_print_mac(&host_addr->hw_address);
            utils_log("\n");
            utils_log("\tProtocol: %04x\n", *protocol);
        }

        // handle encapsulated protocol
        if(*protocol == ETHERTYPE_IPv4 || *protocol == ETHERTYPE_IPv6)
        {
            // Check if frame was meant for this interface
            uint8_t addresses_equal =1;
            if(accept_l2_broadcasts != 0)
            {
                addresses_equal  = utils_compare_mac_broadcast(&dst_addr->hw_address);
            }

            addresses_equal &= utils_compare_mac(&dst_addr->hw_address, &host_addr->hw_address);
        
            if(addresses_equal != 0) 
            {

                if(global_logging)
                {
                    utils_log("Received frame with MAC Address which did not match accept criteria, discarding\n");
                }
                return NET_ERR_ETH_WRONG_ADDR;
            }

        }
        else if(*protocol == ETHERTYPE_ARP)
        {
            arp_msg_t response;
            ret = arp_handle_message(&data[*buffer_offset], *len - *buffer_offset, &host_addr->ip_address, &host_addr->hw_address, &response);
            
            if(global_logging)
            {
                utils_log("Received ARP frame\n");
                utils_log("\tSrc MAC:");
                utils_print_mac(&src_addr->hw_address);
                utils_log("\n\tDest MAC: ");
                utils_print_mac(&dst_addr->hw_address);
                utils_log("\tHost MAC: ");
                utils_print_mac(&host_addr->hw_address);
                utils_log("\n");
            }

            if(ret == NET_ERR_ARP_RESPONSE)
            {
                if(global_logging)
                {
                    utils_log("Send ARP Reply to MAC ");
                    utils_print_mac(&src_addr->hw_address);
                    utils_log("\n");
                }


                if(l2_config.l2_type == NETWORK_L2_ETHERNET)
                {
                    *buffer_offset = sizeof(ethernet_header_t);
                }
                else if(l2_config.l2_type == NETWORK_L2_ETHERNET_VLAN)
                {
                    *buffer_offset = sizeof(ethernet_header_vlan_t);
                }

                ret = system_send(net_id, dst_addr, src_addr, IP_PROTOCOL_RESERVED, ETHERTYPE_ARP, (uint8_t *) &response, sizeof(arp_msg_t), *buffer_offset);
                
                if(ret != NET_ERR_OK && global_logging)
                {
                    utils_log("Error while sending ARP Reply to MAC ");
                    utils_print_mac(&src_addr->hw_address);
                    utils_log("\n");
                }
            }
            *buffer_offset = 0;
            return ret;
        }
        else
        {
            #pragma loopbound min 1 max MAX_SOCKETS 
            for(unsigned short i=0; i < MAX_SOCKETS; i++)
            {
                if(config_array[i].domain == SOCKET_DOMAIN_ETHERNET && config_array[i].eth.protocol == *protocol)
                {
                    // handle all ethernet frames for this protocol received by this interface for all mac addresses
                    if((config_array[i].bound_interface == net_id && config_array[i].eth.hw_address.length == 0) || 
                       // handle all ethernet frames for this protocol received by this interface with a specific mac address
                       (config_array[i].bound_interface == net_id && utils_compare_mac(&dst_addr->hw_address, &config_array[i].eth.hw_address) == 0) ||
                       // handle all ethernet received by any interface with a specific mac address i.e. multicast
                       (utils_compare_mac(&dst_addr->hw_address, &config_array[i].eth.hw_address) == 0)
                    )
                    {
                        if(global_logging)
                        {
                            utils_log("Found Socket Configuration with ID %u \n", i);
                        }
                        
                        if(config_array[i].recv_msg_queue.length > 0)
                        {
                            ret = socket_enque_msg(i,src_addr, data, &data[*buffer_offset], *len - *buffer_offset, rx_ts);
                            *buffer_offset = 0;
                            return ret;
                        }
                        else if(global_logging)
                        {
                            utils_log("Socket has no receive buffer configured\n");
                        }
                        return NET_ERR_SOCKET_NO_RCV_BUFFER;
                    }
                }
            }
            return NET_ERR_SOCKET_PROTOCOL_NOT_CONFIG;
        }
    }
    else
    {
        return NET_ERR_NOT_SUPPORTED;
    }

    return NET_ERR_OK;
}

network_error_t socket_handle_network_layer(
    uint16_t net_id, uint8_t* data, uint16_t len, ethertype_t l3_protocol, ip_protocol_t *l4_protocol, 
    socket_address_t *src_addr,socket_address_t* dst_addr, socket_address_t* host_addr, 
    uint16_t *buffer_offset, uint16_t *payload_length)
{
    // only IPv4 Frames currently supported 
    network_error_t ret;
    if(l3_protocol == ETHERTYPE_IPv4)
    {
        ret = ipv4_handle_packet(&data[*buffer_offset], len - *buffer_offset, l4_protocol, &src_addr->ip_address, &dst_addr->ip_address, payload_length);
        if(ret != NET_ERR_OK)
        {
            if(global_logging)
            {
                utils_log("Error while parsing IP Datagram: 0x%04hx\n", ret);
            }
            return ret;
        }
        
        *buffer_offset += sizeof(ipv4_header_t);
    }

    // Check if frame was meant for this interface
    uint8_t addresses_equal =1;
    // TODO: Add handling for Multicast addresses
    if(accept_ip_broadcasts)
    {
        uint16_t netmask = network_get_interface_ip_netmask(net_id);
        addresses_equal = utils_compare_ip_broadcast(&dst_addr->ip_address, netmask);
    }

    addresses_equal &= utils_compare_ip(&dst_addr->ip_address, &host_addr->ip_address);

    //if(utils_compare_ip(&dst_addr->ip_address, &if_addr.ip.ip_address) != 0)
    if(addresses_equal != 0) 
    {
        if(global_logging)
        {
            utils_log("Datagram was addressed to device with differen IP Address: ");
            utils_print_ip(&dst_addr->ip_address);
            utils_log("\n");
        }
        return NET_ERR_IP_WRONG_ADDR;
    }
    
    if(*l4_protocol == IP_PROTOCOL_ICMP)
    {
        *buffer_offset = 0;
        return NET_ERR_SOCKET_PROTOCOL_NOT_CONFIG;
    }
    
    return NET_ERR_OK;
}

// handle received layer 4 protocol
network_error_t socket_handle_transport_layer(
    uint8_t* data, uint16_t len, ip_protocol_t l4_protocol, 
    socket_address_t *src_addr,socket_address_t* dst_addr, socket_address_t* host_addr, 
    uint16_t *buffer_offset, time_stamp_t *rx_ts, uint16_t* payload_length)
{
    if(l4_protocol == IP_PROTOCOL_UDP)
    {
        
        network_error_t ret = udp_handle_datagram(
            &data[*buffer_offset], len - *buffer_offset, 
            &dst_addr->ip_address, &dst_addr->port, 
            &src_addr->ip_address, &src_addr->port
        );
        if(ret != NET_ERR_OK)
        {
            if(global_logging)
            {
                utils_log("Error while handling UDP Datagram : 0x%04hx \n", ret);
            }
            return ret;
        }
        
        udp_header_t *udp_header = (udp_header_t*) &data[*buffer_offset];
        
        *buffer_offset += sizeof(udp_header_t);
        
        #pragma loopbound min 1 max MAX_SOCKETS 
        for(unsigned short i=0; i < MAX_SOCKETS; i++)
        {
            if(config_array[i].domain == SOCKET_DOMAIN_IP && 
                config_array[i].ip.protocol == IP_PROTOCOL_UDP &&
                config_array[i].ip.src_port == dst_addr->port && 
                config_array[i].bound_interface != GLOBAL_SOCKET)
            {
                if(global_logging)
                {
                    utils_log("Found Socket Configuration with ID %u \n", i);
                }
               
                if(config_array[i].recv_msg_queue.length > 0)
                {
                    ret = socket_enque_msg(i,src_addr, data, &data[*buffer_offset], udp_header->length, rx_ts);
                    *buffer_offset = 0;
                    return ret;
                }
                else if(global_logging)
                {
                    utils_log("Socket has no receive buffer configured\n");
                }

                return NET_ERR_SOCKET_NO_RCV_BUFFER;
            }
        }
    }
    else if(l4_protocol == IP_PROTOCOL_TCP)
    {
        uint16_t socket_tcp_serv_id;
        network_error_t ret;
        uint8_t buffer_tcp[*payload_length];
        memcpy(buffer_tcp, (data + (*buffer_offset)), *payload_length);

        ret = tcp_handle_datagram(  buffer_tcp, &dst_addr->ip_address, &dst_addr->port, src_addr, 
                                    &socket_tcp_serv_id, payload_length, rx_ts);

        if(ret != NET_ERR_OK)
        {
            if(global_logging)
            {
                utils_log("Error while handling TCP Datagram : 0x%04hx \n", ret);
            }
            return ret;
        }

        tcp_header_t* tcp_header = (tcp_header_t*) (data + (*buffer_offset));
        // Access to the data_offset value of the TCP header
        *buffer_offset += (((tcp_header->flags) >> DATA_OFFSET_SHT) * 4);

        if( socket_tcp_serv_id <= MAX_SOCKETS && 
            config_array[socket_tcp_serv_id].domain == SOCKET_DOMAIN_IP && 
            config_array[socket_tcp_serv_id].ip.protocol == IP_PROTOCOL_TCP &&
            config_array[socket_tcp_serv_id].ip.src_port == dst_addr->port && 
            config_array[socket_tcp_serv_id].bound_interface != GLOBAL_SOCKET)
        {
            if(global_logging)
            {
                utils_log("Found Socket server TCP with ID %u \n", socket_tcp_serv_id);
            }

            if(config_array[socket_tcp_serv_id].recv_msg_queue.length > 0)
            {
                if(*payload_length > 0)
                {
                    ret = socket_enque_msg(socket_tcp_serv_id, src_addr, data, &data[*buffer_offset], *payload_length, rx_ts);
                }
                else
                {
                    ret = NET_ERR_OK;
                    network_return_buffer(data);
                }
                *buffer_offset = 0;
                return ret;
            }
            else if(global_logging)
            {
                utils_log("Socket has no receive buffer configured\n");
            }
                return NET_ERR_SOCKET_NO_RCV_BUFFER;
        }
    }
    else
    {

        #pragma loopbound min 1 max MAX_SOCKETS 
        for(unsigned short i=0; i < MAX_SOCKETS; i++)
        {
            if(config_array[i].domain == SOCKET_DOMAIN_IP && config_array[i].ip.protocol == l4_protocol)
            {
                if(global_logging)
                {
                    utils_log("Found Socket Configuration with ID %u \n", i);
                }

                if(config_array[i].recv_msg_queue.length > 0)
                {
                    network_error_t ret = socket_enque_msg(i,src_addr, data, &data[*buffer_offset], len - *buffer_offset, rx_ts);
                    *buffer_offset = 0;
                    return ret;
                }
                else if(global_logging)
                {
                    utils_log("Socket has no receive buffer configured\n");
                }
                
                return NET_ERR_SOCKET_NO_RCV_BUFFER;
            }
        }
    }

    return NET_ERR_SOCKET_PROTOCOL_NOT_CONFIG;
}


static network_error_t system_send(uint16_t net_id, socket_address_t *src_addr, socket_address_t *dst_addr, ip_protocol_t l3_protocol, ethertype_t l2_protocol, uint8_t* data, uint16_t data_len, uint16_t header_len )
{
    uint8_t header_buffer[header_len];

    network_frame_t frame;
    frame.user_data = data;
    frame.user_data_length=data_len;
    frame.header_length = header_len;
    frame.priority = 0;

    uint16_t buffer_offset = header_len;
    
    network_error_t ret;
    if(l3_protocol != IP_PROTOCOL_RESERVED)
    {
        if(src_addr->ip_address.version == IP_VERSION_4)
        {
            buffer_offset -= sizeof(ipv4_header_t);
            ret = ipv4_build_datagram(&dst_addr->ip_address, &src_addr->ip_address, 
                        l3_protocol, &header_buffer[buffer_offset], data_len);

            if(ret != NET_ERR_OK)
            {
                if( global_logging)
                {
                    utils_log("system_send: Error while building IP Datagram : 0x%04hx\n", ret);
                }
                return ret;
            }
        }
    }
    // l2 configuration is required in all cases
    network_layer2_config_t l2_config;
    ret = network_get_l2_config(net_id, &l2_config);
    if(ret != NET_ERR_OK)
    {
        if( global_logging)
        {
            utils_log("system_send: Error when retreiving L2 configuration of Interface %u: 0x%04hx\n", net_id, ret);
        }
        return ret;
    }
    
    // calculate required space for OSI layer 2 protocols
    if(l2_config.l2_type == NETWORK_L2_ETHERNET)
    {
        buffer_offset -= sizeof(ethernet_header_t);
        ret = ethernet_build_frame(&dst_addr->hw_address, &src_addr->hw_address, l2_protocol, &header_buffer[buffer_offset]);
    }
    else if(l2_config.l2_type == NETWORK_L2_ETHERNET_VLAN)
    {
        buffer_offset -= sizeof(ethernet_header_vlan_t);
        ret = ethernet_build_vlan_frame(&dst_addr->hw_address, &src_addr->hw_address, l2_protocol, &l2_config.vlan, &header_buffer[buffer_offset]);
    }
    
    if(ret != NET_ERR_OK)
    {
        if( global_logging)
        {
            utils_log("system_send: Error while building MAC Frame : 0x%04hx\n", ret);
        }
        return ret;
    }

    frame.header_data = &header_buffer[buffer_offset];
    return network_send(net_id, &frame, 0);
}

static network_error_t socket_enque_msg(uint16_t socket_id, socket_address_t *src_address, uint8_t* frame_start, uint8_t* msg, uint16_t length, time_stamp_t * rx_ts)
{
    // utils_log("start socket_enque_msg()\n");
    socket_msg_queue_t *queue = & config_array[socket_id].recv_msg_queue;
    uint8_t write_ptr = queue->write_ptr;

    // check whether an entry is left
    if(queue->queue[queue->write_ptr].frame_length != 0)
    {
        return NET_ERR_SOCKET_RCV_BUFFER_FULL;
    }
 
    // calculate next write pointer
    queue->write_ptr = (queue->write_ptr+1>= queue->length ) ? 0 : queue->write_ptr+1;
    
    queue->queue[write_ptr].frame_length = length;
    // utils_log("addr during writing frame_length = %p\n", &queue->queue[write_ptr].frame_length);
    queue->queue[write_ptr].buffer = (void *)frame_start;
    queue->queue[write_ptr].msg = (void *) msg;
    queue->queue[write_ptr].ts = *rx_ts;
    queue->queue[write_ptr].addr = *src_address;
    
    if(config_array[socket_id].options.debug_output || global_logging)
    {
        utils_log("Message stored in buffer %i of socket %i\n", write_ptr, socket_id);
    }

    return NET_ERR_OK;
}


network_error_t socket_receive(uint16_t socket_id, socket_address_t *src_addr, uint8_t* data, uint16_t *data_length, uint16_t buffer_length)
{    
    socket_msg_queue_t *queue = & config_array[socket_id].recv_msg_queue;
    uint8_t read_ptr = queue->read_ptr;

    if(queue->queue[queue->read_ptr].frame_length >= buffer_length)
    {
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Message stored in buffer of socket %i to large for given buffer: message size %i, buffer size %i\n", socket_id, queue->queue[queue->read_ptr].frame_length, buffer_length);
        }
        return NET_ERR_SOCKET_MSMG_LEN_EXCEEDED;
    }

    // check whether an entry is left
    if(queue->queue[queue->read_ptr].frame_length == 0)
    {
        // utils_log("frame lenght = 0\n");
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("No message received at socket %i\n", socket_id);
        }
        return NET_ERR_SOCKET_NO_MSG;
    }

    queue->read_ptr = (queue->read_ptr+1>= queue->length) ? 0 : queue->read_ptr+1;
    
    *data_length = queue->queue[read_ptr].frame_length;
    *src_addr = queue->queue[read_ptr].addr;
   
    // copy data
    // TODO: Fix copy process
    /*
    #pragma loopbound min 1 max 375
    for(uint16_t i=0; i<(*rcv_data_length>>2); i++)
    {
        data_ptr[i] = buf_ptr[i];
    }
    */
    
    uint8_t *msg = (uint8_t *) queue->queue[read_ptr].msg; 
    #pragma loopbound min 1 max SOCKET_MAX_FRAME_SIZE
    for(uint16_t i=0; i< queue->queue[read_ptr].frame_length; i++)
    {
        data[i] = msg[i];
    }
    
    network_return_buffer(queue->queue[read_ptr].buffer);

    // Mark queue entry as free again
    queue->queue[read_ptr].frame_length = 0;

    if(config_array[socket_id].options.debug_output || global_logging)
    {
        utils_log("Message removed from buffer %i of socket %i\n", read_ptr, socket_id);
    }

    return NET_ERR_OK;
}


network_error_t socket_receive_w_ts(uint16_t socket_id, socket_address_t *src_addr, uint8_t* data, uint16_t *data_length, uint16_t buffer_length, time_stamp_t* rx_ts)
{
    socket_msg_queue_t *queue = & config_array[socket_id].recv_msg_queue;
    uint8_t read_ptr = queue->read_ptr;

    if(queue->queue[queue->read_ptr].frame_length >= buffer_length)
    {
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("Message stored in buffer of socket %i to large for given buffer: message size %i, buffer size %i\n", socket_id, queue->queue[queue->read_ptr].frame_length, buffer_length);
        }
        return NET_ERR_SOCKET_MSMG_LEN_EXCEEDED;
    }

    // check whether an entry is left
    if(queue->queue[queue->read_ptr].frame_length == 0)
    {
        if(config_array[socket_id].options.debug_output || global_logging)
        {
            utils_log("No message received at socket %i\n", socket_id);
        }
        return NET_ERR_SOCKET_NO_MSG;
    }

    queue->read_ptr = (queue->read_ptr+1>= queue->length) ? 0 : queue->read_ptr+1;
    
    *data_length = queue->queue[read_ptr].frame_length;
    *src_addr = queue->queue[read_ptr].addr;
    *rx_ts = queue->queue[read_ptr].ts;

    // copy data
    // TODO: Fix copy process
    /*
    #pragma loopbound min 1 max 375
    for(uint16_t i=0; i<(*rcv_data_length>>2); i++)
    {
        data_ptr[i] = buf_ptr[i];
    }
    */

    uint8_t *msg = (uint8_t *) queue->queue[read_ptr].msg; 
    #pragma loopbound min 1 max SOCKET_MAX_FRAME_SIZE
    for(uint16_t i=0; i< queue->queue[read_ptr].frame_length; i++)
    {
        data[i] = msg[i];
    }
    
    network_return_buffer(queue->queue[read_ptr].buffer);
    
    // Mark queue entry as free again
    queue->queue[read_ptr].frame_length = 0;

    if(config_array[socket_id].options.debug_output || global_logging)
    {
        utils_log("Message removed from buffer %i of socket %i\n", read_ptr, socket_id);
    }

    return NET_ERR_OK;
}

// return pointer to a specific socket
socket_t* socket_get(uint16_t socket_id)
{
    // socket_id can't be > MAX_SOCKET, verify it in the calling function
    return &config_array[socket_id];
}

// copy all the contenent of addr1 in addr2 (deep copy)
void socket_deep_copy_addr(socket_address_t* addr1, socket_address_t* addr2)
{
    addr2->hw_address.length = addr1->hw_address.length;
    for(uint8_t i = 0; i < 8; i++)
    {
        addr2->hw_address.address[i] = addr1->hw_address.address[i];
    }
    
    addr2->ip_address.version = addr1->ip_address.version;
    for(uint8_t i = 0; i < 4; i++)
    {
        addr2->ip_address.ipv4_addr[i] = addr1->ip_address.ipv4_addr[i];
    }
    //to do copy addr ipv6
    addr2->port = addr1->port;
    addr2->if_id = addr1->if_id;
}